#include <bitvector.h>
#include <string.h>

BitVector *bv_create_size(int size)
{
  BitVector *bv = ALLOC(BitVector);

  bv->capa = (size >> 3) + 1;
  bv->bits = ALLOC_N(uchar, bv->capa);
  memset(bv->bits, 0, bv->capa);

  bv->size = 0;
  bv->count = 0;
  bv->curr_bit = -1;
  return bv;
}

BitVector *bv_create()
{
  return bv_create_size(BV_INIT_CAPA);
}

void bv_destroy(BitVector *bv)
{
  free(bv->bits);
  free(bv);
}

void bv_set(BitVector *bv, int bit)
{
  uchar *byte_p;
  int byte = bit>>3;
  uchar bitmask = 1<<(bit&7);
  if (bv->size <= byte) {
    bv->size = byte + 1;
    if (bv->size >= bv->capa) {
      int capa = bv->capa * 2;
      while (capa < bv->size) capa *= 2;
      REALLOC_N(bv->bits, uchar, capa);
      memset(bv->bits + bv->capa, 0, capa - bv->capa);
      bv->capa = capa;
    }
  }
  byte_p = &(bv->bits[byte]);
  if ((bitmask & *byte_p) == 0) {
    bv->count++;
    *byte_p |= bitmask;
  }
}

int bv_get(BitVector *bv, int bit)
{
  int byte = bit>>3;
  if (byte >= bv->size) return 0;
  return (bv->bits[byte]>>(bit&7))&1;
}

void bv_clear(BitVector *bv)
{
  memset(bv->bits, 0, bv->size);
  bv->count = 0;
}

void bv_unset(BitVector *bv, int bit)
{
  uchar *byte_p;
  uchar bitmask;
  int byte = bit>>3;
  
  if (byte >= bv->size) return;
  
  byte_p = &(bv->bits[byte]);
  bitmask = 1<<(bit&7);
  if ((bitmask & *byte_p) > 0) {
    bv->count--;
    *byte_p &= ~bitmask;
  }
}

void bv_write(BitVector *bv, Store *store, char *name)
{
  OutStream *os = store->create_output(store, name);
  os_write_vint(os, bv->size);
  os_write_bytes(os, bv->bits, bv->size);
  os_close(os);
}

const uchar BYTE_COUNTS[] = { // table of bits/char
  0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};

int bv_count(BitVector *bv) 
{
  // if the vector has been modified
  int i, c = 0;
  uchar *bytes = bv->bits;
  for (i = 0; i < bv->size; i++)
    c += BYTE_COUNTS[bytes[i]]; // sum bits per char
  bv->count = c;
  return c;
}

BitVector *bv_read(Store *store, char *name)
{
  BitVector *bv = ALLOC(BitVector);
  InStream *is = store->open_input(store, name);
  bv->capa = bv->size = (int)is_read_vint(is);
  bv->bits = ALLOC_N(uchar, bv->capa);
  is_read_bytes(is, bv->bits, 0, bv->size);
  is_close(is);
  bv_count(bv);
  return bv;
}

void bv_scan_reset(BitVector *bv)
{
  bv->curr_bit = -1;
}

inline int bv_scan_next_from(BitVector *bv, register const int from)
{
  register const uchar *const bits = bv->bits;
  register const int size = bv->size;
  register int byte_pos = (from) >> 3;
  register int inc = ((from) & 7);
  register int bit = 1 << inc;
  register int mask = 0xff << inc;
  register int byte;

  if (byte_pos >= size) return -1;
  if ((bits[byte_pos]&mask) == 0) {
    inc = 0;
    bit = 1;
    do {
      byte_pos++;
      if (byte_pos >= size) return -1;
    } while (bits[byte_pos] == 0);
  }

  byte = bits[byte_pos];
  while ((byte & bit) == 0) {
    bit <<= 1;
    inc++;
  }

  return bv->curr_bit = ((byte_pos << 3) + inc);
}

inline int bv_scan_next(BitVector *bv)
{
  return bv_scan_next_from(bv, bv->curr_bit+1);
}

