#include "index.h" 

static char * const ALREADY_CLOSED_MSG = "Already closed";
static char * const STREAM_CLOSED_MSG = "Stream closed";
static char * const MISSING_FILE_MSG = "No sub-file found";
static char * const ALREADY_MERGED_MSG = "Already merged";
static char * const REMAINDER_ERROR_MSG = "Non-zero remainder length after copying";
static char * const FILE_OFFSET_MSG =  "Difference in the output file offsets"
        " does not match the original file length";
static char * const NO_FILES_TO_MERGE_MSG = "No Files to merge into the compound file";

extern void store_destroy(Store *store);

/****************************************************************************
 *
 * CompoundStore
 *
 ****************************************************************************/

typedef struct FileEntry {
  int offset;
  int length;
} FileEntry;

void cmpd_touch(Store *store, char *filename)
{
  store->dir.cmpd->store->touch(store->dir.cmpd->store, filename);
}

int cmpd_exists(Store *store, char *filename)
{
  if (h_get(store->dir.cmpd->entries, filename) != NULL) {
    return true;
  } else {
    return false;
  }
}

/**
 * @throws UNSUPPORTED_ERROR
 */
int cmpd_remove(Store *store, char *filename)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return 0;
}

/**
 * @throws UNSUPPORTED_ERROR
 */
int cmpd_rename(Store *store, char *from, char *to)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return 0;
}

int cmpd_count(Store *store)
{
  return store->dir.cmpd->entries->used;
}

/**
 * @throws UNSUPPORTED_ERROR
 */
void cmpd_clear(Store *store)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
}

void cmpd_close_i(Store *store)
{
  CompoundStore *cmpd = store->dir.cmpd;
  if (cmpd->stream == NULL)
    RAISE(IO_ERROR, ALREADY_CLOSED_MSG);

  h_destroy(cmpd->entries);

  is_close(cmpd->stream);
  cmpd->stream = NULL;
  free(store->dir.cmpd);
  store_destroy(store);
}

int cmpd_length(Store *store, char *filename)
{
  FileEntry *fe = (FileEntry *)h_get(store->dir.cmpd->entries, filename);
  if (fe != NULL)
    return fe->length;
  else
    return 0;
}

void cmpdi_seek_internal(InStream *is, int pos) {}
void cmpdi_close_internal(InStream *is)
{
  //is_close(is->d.cis->sub);
  free(is->d.cis);
}

void cmpdi_clone_internal(InStream *is, InStream *new_is)
{
  CompoundInStream *cis = ALLOC(CompoundInStream);
  //cis->sub = is_clone(is->d.cis->sub);
  cis->sub = is->d.cis->sub;
  cis->offset = is->d.cis->offset;
  cis->length = is->d.cis->length;
  new_is->d.cis = cis;
}

int cmpdi_length_internal(InStream *is)
{
  return (is->d.cis->length);
}

/*
 * raises: EOF_ERROR
 */
void cmpdi_read_internal(InStream *is, uchar *b, int offset, int len)
{
  CompoundInStream *cis = is->d.cis;
  int start = is_pos(is);
  if ((start + len) > cis->length)
    RAISE(EOF_ERROR, EOF_ERROR_MSG);
  is_seek(cis->sub, cis->offset + start);
  is_read_bytes(cis->sub, b, offset, len);
}

InStream *cmpd_create_input(InStream *sub_is, int offset, int length)
{
  InStream *is = is_create();
  CompoundInStream *cis = ALLOC(CompoundInStream);
  //cis->sub = is_clone(sub_is);
  cis->sub = sub_is;
  cis->offset = offset;
  cis->length = length;
  is->d.cis = cis;
  is->file = NULL;

  is->read_internal = &cmpdi_read_internal;
  is->seek_internal = &cmpdi_seek_internal;
  is->close_internal = &cmpdi_close_internal;
  is->clone_internal = &cmpdi_clone_internal;
  is->length_internal = &cmpdi_length_internal;
  return is;
}

InStream *cmpd_open_input(Store *store, const char *filename)
{
  FileEntry *entry;
  CompoundStore *cmpd = store->dir.cmpd;
  InStream *is;

  mutex_lock(&store->mutex);
  if (cmpd->stream == NULL) {
    mutex_unlock(&store->mutex);
    RAISE(IO_ERROR, STREAM_CLOSED_MSG);
  }

  entry = (FileEntry *)h_get(cmpd->entries, filename);
  if (entry == NULL) {
    mutex_unlock(&store->mutex);
    RAISE(IO_ERROR, MISSING_FILE_MSG);
  }

  is = cmpd_create_input(cmpd->stream, entry->offset, entry->length);
  mutex_unlock(&store->mutex);

  return is;
}

OutStream *cmpd_create_output(Store *store, const char *filename)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return NULL;
}

Lock *cmpd_open_lock(Store *store, char *lockname)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return NULL;
}

void cmpd_close_lock(Lock *lock)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
}

Store *open_cmpd_store(Store *store, const char *name)
{
  int count, i, offset;
  char *fname;
  FileEntry *entry;
  Store * volatile new_store = NULL;
  CompoundStore * volatile cmpd = NULL;
  InStream * volatile is = NULL;

  TRY
    new_store = store_create();
    cmpd = ALLOC(CompoundStore);

    cmpd->store   = store;
    cmpd->name    = name;
    cmpd->entries = h_new_str(&free, &free);
    is = cmpd->stream = store->open_input(store, cmpd->name);

    // read the directory and init files
    count = (int)is_read_vint(is);
    entry = NULL;
    for (i = 0; i < count; i++) {
      offset = (int)is_read_long(is);
      fname = is_read_string(is);

      if (entry != NULL) {
        // set length of the previous entry
        entry->length = offset - entry->offset;
      }

      entry = ALLOC(FileEntry);
      entry->offset = offset;
      h_set(cmpd->entries, fname, entry);
    }

    // set the length of the final entry
    if (entry != NULL)
      entry->length = is_length(is) - entry->offset;
  XCATCHALL
    free(new_store);
    free(cmpd);
    if (is) is_close(is);
  XENDTRY


  new_store->dir.cmpd = cmpd;
  new_store->touch    = &cmpd_touch;
  new_store->exists   = &cmpd_exists;
  new_store->remove   = &cmpd_remove;
  new_store->rename   = &cmpd_rename;
  new_store->count    = &cmpd_count;
  new_store->clear    = &cmpd_clear;
  new_store->length   = &cmpd_length;
  new_store->close_i  = &cmpd_close_i;
  new_store->create_output = &cmpd_create_output;
  new_store->open_input    = &cmpd_open_input;
  new_store->open_lock     = &cmpd_open_lock;
  new_store->close_lock    = &cmpd_close_lock;
  return new_store;
}

/****************************************************************************
 *
 * CompoundWriter
 *
 ****************************************************************************/

typedef struct WFileEntry {
  char *name;
  int dir_offset;
  int data_offset;
} WFileEntry;

WFileEntry *wfe_create(char *name)
{
  WFileEntry *wfe = ALLOC(WFileEntry);
  wfe->name = name;
  return wfe;
}

CompoundWriter *open_cw(Store *store, char *name)
{
  CompoundWriter *cw = ALLOC(CompoundWriter);
  cw->store = store;
  cw->name = name;
  cw->ids = hs_str_create(NULL);
  cw->file_entries = ary_create(1, &free);
  cw->merged = false;
  return cw;
}

void cw_add_file(CompoundWriter *cw, char *id)
{
  if (cw->merged) RAISE(STATE_ERROR, ALREADY_MERGED_MSG);
  if (hs_add(cw->ids, id) != HASH_KEY_DOES_NOT_EXIST)
    RAISE(STATE_ERROR, ALREADY_MERGED_MSG);

  hs_add(cw->ids, id);
  ary_append(cw->file_entries, wfe_create(id));
}

void cw_copy_file(CompoundWriter *cw, WFileEntry *src, OutStream *os)
{
  int start_ptr = os_pos(os);
  int end_ptr;
  int remainder, length, len;
  uchar buffer[BUFFER_SIZE];
  
  InStream *is = cw->store->open_input(cw->store, src->name);
  
  TRY
    remainder = length = is_length(is);


    while (remainder > 0) {
      len = MIN(remainder, BUFFER_SIZE);
      is_read_bytes(is, buffer, 0, len);
      os_write_bytes(os, buffer, len);
      remainder -= len;
    }

    // Verify that remainder is 0
    if (remainder != 0)
      RAISE(IO_ERROR, REMAINDER_ERROR_MSG);

    // Verify that the output length diff is equal to original file
    end_ptr = os_pos(os);
    len = end_ptr - start_ptr;
    if (len != length)
      RAISE(IO_ERROR, FILE_OFFSET_MSG);

  XFINALLY
    is_close(is);
  XENDTRY
}

void cw_close(CompoundWriter *cw)
{
  OutStream * volatile os = NULL;
  int i;
  WFileEntry *wfe;

  if (cw->merged) RAISE(STATE_ERROR, ALREADY_MERGED_MSG);
  if (cw->ids->size <= 0)
    RAISE(STATE_ERROR, NO_FILES_TO_MERGE_MSG);

  cw->merged = true;

  TRY
    os = cw->store->create_output(cw->store, cw->name);
    os_write_vint(os, cw->file_entries->size);

    /* Write the directory with all offsets at 0.
     * Remember the positions of directory entries so that we can adjust the
     * offsets later */

    for (i = 0; i < cw->file_entries->size; i++) {
      wfe = (WFileEntry *)cw->file_entries->elems[i];
      wfe->dir_offset = os_pos(os);
      os_write_long(os, 0);  // for now
      os_write_string(os, wfe->name);
    }

    /* Open the files and copy their data into the stream.  Remember the
     * locations of each file's data section. */
    for (i = 0; i < cw->file_entries->size; i++) {
      wfe = (WFileEntry *)cw->file_entries->elems[i];
      wfe->data_offset = os_pos(os);
      cw_copy_file(cw, wfe, os);
    }

    /* Write the data offsets into the directory of the compound stream */
    for (i = 0; i < cw->file_entries->size; i++) {
      wfe = (WFileEntry *)cw->file_entries->elems[i];
      os_seek(os, wfe->dir_offset);
      os_write_long(os, wfe->data_offset);
    }

  XFINALLY
    if (os) os_close(os);
    hs_destroy(cw->ids);
    ary_destroy(cw->file_entries);
    free(cw);
    break;
  XENDTRY
}

