#include "document.h"
#include <string.h>

/****************************************************************************
 *
 * DocField
 *
 ****************************************************************************/

static char * const INDEX_NO_STORE_NO_MSG = "it doesn't make sense to have a field that is neither indexed nor stored";
static char * const INDEX_NO_TV_YES_MSG = "cannot store term vector information for a field that is not indexed";
static char * const INVALID_STORE_VAL_MSG = "Invalid value for store in document field";
static char * const INVALID_INDEX_VAL_MSG = "Invalid value for index in document field";
static char * const INVALID_TV_VAL_MSG = "Invalid value for term_vector in document field";
static char * const BIN_FIELD_STORE_NO_MSG = "It doesn't make sense not to store binary data";

/**
 * @throws ARG_ERROR
 */
inline void df_set(DocField *df, const char *name,
    char *data, int store, int index, int tv)
{
  if ((index == DF_INDEX_NO) && (store == DF_STORE_NO))
    RAISE(ARG_ERROR, INDEX_NO_STORE_NO_MSG);
  if ((index == DF_INDEX_NO) && (tv != DF_TERM_VECTOR_NO))
    RAISE(ARG_ERROR, INDEX_NO_TV_YES_MSG);
  df->name = estrdup(name);
  df->data = data;
  df->blen = (int)strlen(data);
  df_set_store(df, store);
  df_set_index(df, index);
  df_set_term_vector(df, tv);
  df->is_binary = false;
  df->boost = 1.0;
}

/*
 * @throws ARG_ERROR
 */
DocField *df_create(const char *name, char *data, int store, int index, int tv)
{
  DocField *df = ALLOC(DocField);
  df_set(df, name, data, store, index, tv);
  return df;
}

DocField *df_clone(DocField *self)
{
  DocField *clone = ALLOC(DocField);
  memcpy(clone, self, sizeof(DocField));
  clone->name = estrdup(self->name);
  clone->data = estrdup(self->data);
  return clone;
}

void df_destroy(DocField *df)
{
  free(df->name);
  free(df);
}

void df_destroy_data(DocField *df)
{
  free(df->data);
  free(df->name);
  free(df);
}

/*
 * @throws ARG_ERROR
 */
void df_set_store(DocField *df, int store)
{
  switch (store) {
    case DF_STORE_YES:
      df->is_stored = true;
      df->is_compressed = false;
      break;
    case DF_STORE_NO:
      df->is_stored = false;
      df->is_compressed = false;
      break;
    case DF_STORE_COMPRESS:
      df->is_stored = true;
      df->is_compressed = true;
      break;
    default:
      RAISE(ARG_ERROR, INVALID_STORE_VAL_MSG);
  }
}

/*
 * @throws ARG_ERROR
 */
void df_set_index(DocField *df, int index)
{
  df->omit_norms = false;
  switch (index) {
    case DF_INDEX_NO:
      df->is_indexed = false;
      df->is_tokenized = false;
      break;
    case DF_INDEX_TOKENIZED:
      df->is_indexed = true;
      df->is_tokenized = true;
      break;
    case DF_INDEX_UNTOKENIZED:
      df->is_indexed = true;
      df->is_tokenized = false;
      break;
    case DF_INDEX_NO_NORMS:
      df->is_indexed = true;
      df->is_tokenized = false;
      df->omit_norms = true;
      break;
    default:
      RAISE(ARG_ERROR, INVALID_INDEX_VAL_MSG);
  }
}

/*
 * @throws ARG_ERROR
 */
void df_set_term_vector(DocField *df, int tv)
{
  switch (tv) {
    case DF_TERM_VECTOR_NO:
      df->store_tv = false;
      df->store_offset = false;
      df->store_pos = false;
      break;
    case DF_TERM_VECTOR_YES:
      df->store_tv = true;
      df->store_offset = false;
      df->store_pos = false;
      break;
    case DF_TERM_VECTOR_WITH_OFFSETS:
      df->store_tv = true;
      df->store_offset = true;
      df->store_pos = false;
      break;
    case DF_TERM_VECTOR_WITH_POSITIONS:
      df->store_tv = true;
      df->store_offset = false;
      df->store_pos = true;
      break;
    case DF_TERM_VECTOR_WITH_POSITIONS_OFFSETS:
      df->store_tv = true;
      df->store_offset = true;
      df->store_pos = true;
      break;
    default:
      RAISE(ARG_ERROR, INVALID_TV_VAL_MSG);
  }
}

/*
 * @throws ARG_ERROR
 */
DocField *df_create_binary(char *name, char *data, int blen, int store)
{
  DocField *df;
  
  if (store == DF_STORE_NO) {
    RAISE(ARG_ERROR, BIN_FIELD_STORE_NO_MSG);
  }
  
  df = df_create(name, data, store, DF_INDEX_NO, DF_TERM_VECTOR_NO);
  df->is_binary = true;
  df->blen = blen;
  return df;
}

char *df_to_s(DocField *self)
{
  /* the length of the str is name.len + data.len + 119, add safety 10 */
  char *str = ALLOC_N(char, strlen(self->name) + strlen(self->data) + 129);
  char *str_ptr = str;

  if (self->is_stored) {
    sprintf(str, "stored/%s,", self->is_compressed ?  "compressed" : "uncompressed");
    str_ptr = str + strlen(str);
  }
  sprintf(str_ptr, "%s%s%s%s%s%s%s<%s:%s>",
      self->is_indexed ? "indexed," : "",
      self->is_tokenized ? "tokenized," : "",
      self->store_tv ? "store_term_vector," : "",
      self->store_offset ? "store_offsets," : "",
      self->store_pos ? "store_positions," : "",
      self->omit_norms ? "omit_norms," : "",
      self->is_binary ? "binary," : "",
      self->name,
      self->is_binary ? "=bin_data=" : self->data);

  return str;
}

/****************************************************************************
 *
 * Document
 *
 ****************************************************************************/

Document *doc_create()
{
  Document *doc = ALLOC(Document);
  doc->fields = h_new_str(&free, (free_ft)&ary_destroy);
  doc->fcnt = 0;
  doc->dfcnt = 0;
  doc->field_arr = NULL;
  doc->df_arr = NULL;
  doc->boost = 1.0;
  doc->free_data = (free_ft)&df_destroy_data;
  return doc;
}

Document *doc_create_keep_data()
{
  Document *doc = doc_create();
  doc->free_data = (free_ft)&df_destroy;
  return doc;
}

void doc_destroy(Document *doc)
{
  int i;
  if (doc->free_data) {
    for (i = 0; i < doc->dfcnt; i++) {
      doc->free_data(doc->df_arr[i]);
    }
  }
  free(doc->field_arr);
  free(doc->df_arr);
  h_destroy(doc->fields);
  free(doc);
}

void doc_add_field(Document *doc, DocField *df)
{
  Array *fields = (Array *)h_get(doc->fields, df->name);
  if (fields == NULL) {
    fields = ary_create(1, NULL);
    h_set(doc->fields, estrdup(df->name), fields);
    doc->fcnt++;
    REALLOC_N(doc->field_arr, Array *, doc->fcnt);
    doc->field_arr[doc->fcnt-1] = fields;
  }
  ary_append(fields, df);
  doc->dfcnt++;
  REALLOC_N(doc->df_arr, DocField *, doc->dfcnt);
  doc->df_arr[doc->dfcnt-1] = df;
}

DocField *doc_get_field(Document *doc, const char *fname)
{
  Array *fields = (Array *)h_get(doc->fields, fname);
  if (fields) {
    return fields->elems[0];
  } else {
    return NULL;
  }
}

Array *doc_get_fields(Document *doc, const char *fname)
{
  return (Array *)h_get(doc->fields, fname);
}

/**
 * TODO:
 * This is not exactly elegant or efficient but it works and is not going to
 * be a performance problem. Still, it would be nice to make the code a little
 * clearer.
 */
Array *doc_remove_fields(Document *doc, const char *fname)
{
  Array *fields = (Array *)h_rem(doc->fields, fname, true);
  if (fields) {
    int i, j;
    doc->fcnt--;
    for (i = 0; i < doc->fcnt; i++) {
      if (fields == doc->field_arr[i]) {
        memmove(&doc->field_arr[i],
                &doc->field_arr[i+1],
                sizeof(void *) * (doc->fcnt - i));
        break;
      }
    }
    for (i = 0, j = 0; i < doc->dfcnt && j < fields->size;) {
      if (fields->elems[j] == doc->df_arr[i]) {
        memmove(&doc->df_arr[i],
                &doc->df_arr[i+1],
                sizeof(void *) * (doc->dfcnt - i - 1));
        j++;
        doc->dfcnt--;
      } else {
        i++;
      }
    }
    fields->free_elem = doc->free_data;
    return fields;
  } else {
    return NULL;
  }
}

DocField *doc_remove_field(Document *doc, const char *fname)
{
  DocField *df = NULL;
  Array *dfs = (Array *)h_get(doc->fields, fname);
  if (dfs) {
    df = ary_remove(dfs, 0);
    if (dfs->size == 0) {
      Array *fields = doc_remove_fields(doc, fname);
      fields->free_elem = doc->free_data;
      ary_destroy(fields);
    } else {
      int i;
      for (i = 0; i < doc->dfcnt; i++) {
        if (df == doc->df_arr[i]) {
          memmove(&doc->df_arr[i],
                  &doc->df_arr[i+1],
                  sizeof(void *) * (doc->dfcnt - i - 1));
          doc->dfcnt--;
          break;
        }
      }
    }
  }
  return df;
}

bool doc_delete_fields(Document *doc, const char *fname)
{
  Array *fields = doc_remove_fields(doc, fname);
  if (fields) {
    fields->free_elem = doc->free_data;
    ary_destroy(fields);
    return true;
  } else {
    return false;
  }
}

char *doc_to_s(Document *doc)
{
  int i, len = 20;
  char *str, *str_ptr;
  char **df_strs = ALLOC_N(char *, doc->dfcnt);
  for (i = 0; i < doc->dfcnt; i++) {
    df_strs[i] = df_to_s(doc->df_arr[i]);
    len += (int)strlen(df_strs[i]) + 3;
  }
  str_ptr = str = ALLOC_N(char, len);
  sprintf(str_ptr, "Document {\n");
  str_ptr += strlen(str_ptr);
  for (i = 0; i < doc->dfcnt; i++) {
    sprintf(str_ptr, "  %s\n", df_strs[i]);
    free(df_strs[i]);
    str_ptr += strlen(str_ptr);
  }
  sprintf(str_ptr, "}");
  free(df_strs);

  return str;
}
