#include "store.h"
#include "frtio.h"
#ifdef WIN32
# ifndef S_IRUSR
#  define S_IRUSR _S_IREAD
# endif
# ifndef S_IWUSR
# define S_IWUSR _S_IWRITE
# endif
# include <io.h>
#else
#include <unistd.h>
# define _close close
# define _open open
# define _creat creat
# define _lseek lseek
# define _read read
# define _stat stat
# define _fstat fstat
#endif
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>


static char * const FILE_OPEN_ERROR_MSG = "Couldn't open the file to read";
static char * const SEEK_ERROR_MSG = "Seek error message";
static char * const WRITE_ERROR_MSG = "Write error message";

extern void store_destroy(Store *store);

/**
 * Create the file +filename+ in the +store+.
 *
 * @param store the store to create the file in
 * @param filename the name of the file to create
 * @throws IO_ERROR if the file cannot be created
 */
static void fs_touch(Store *store, char *filename)
{
  int f;
  char path[MAX_FILE_PATH];
  join_path(path, store->dir.path, filename);
  if ((f = _creat(path, S_IRUSR|S_IWUSR)) == 0) {
    RAISE(IO_ERROR, strerror(errno));
  }
  _close(f);
}

/**
 * Return true if a file of name +filename+ exists in +store+.
 *
 * @param store the store to check in
 * @param filename the name of the file to check for
 * @returns true if the file exists
 * @throws IO_ERROR if there is an error checking for the files existance
 */
int fs_exists(Store *store, char *filename)
{
  char buf[MAX_FILE_PATH];
  return exists(join_path(buf, store->dir.path, filename));
}

/**
 * Remove the file +filename+ from the +store+
 *
 * @param store the store to remove from
 * @param filename the name of the file to remove
 * @returns On success, zero is returned.  On error, -1 is returned, and errno
 *          is set appropriately.
 */
int fs_remove(Store *store, char *filename)
{
  char buf[MAX_FILE_PATH];
  return remove(join_path(buf, store->dir.path, filename));
}

/**
 * Rename the file in the +store+ from the name +from+ to the name +to+.
 *
 * @param store the store to rename the file in
 * @param from the name of the file to rename
 * @param to the new name of the file
 * @returns On success, zero is returned.  On error, -1 is returned, and errno
 *          is set appropriately.
 */
int fs_rename(Store *store, char *from, char *to)
{
  char buf1[MAX_FILE_PATH], buf2[MAX_FILE_PATH];

#ifdef WIN32
  remove(join_path(buf1, store->dir.path, to));
#endif

  if (rename(join_path(buf1, store->dir.path, from),
    join_path(buf2, store->dir.path, to)) < 0) {
    RAISE(IO_ERROR, strerror(errno));
  }
  return true;
}

/**
 * Returns the number of files in the store.
 *
 * @param store the store to count the files in
 * @return the number of files in the store
 * @throws IO_ERROR if there is an error opening the directory
 */
int fs_count(Store *store)
{
  return fcount(store->dir.path);
}

/**
 * Call the function +func+ with each filename in the store and the arg that
 * you passed. If you need to open the file you should pass the store as the
 * argument. If you need to pass more than one argument, you should pass a
 * struct.
 *
 * @param store the store to grab the files from
 * @param func the function to call with each files name and the +arg+ passed
 * @param arg the argument to pass to the function
 * @throws IO_ERROR if there is an error opening the directory
 */
void fs_each(Store *store, void (*func)(char *fname, void *arg), void *arg)
{
  dir_each(store->dir.path, func, arg);
}

/**
 * Destroy the store.
 *
 * @param p the store to destroy
 * @throws IO_ERROR if there is an error deleting the locks
 */
void fs_destroy(Store *store)
{
  fs_clear_locks(store);
  free(store->dir.path);
  store_destroy(store);
}

/**
 * Return the length of the file +filename+ in +store+
 *
 * @param store the store to use
 * @param the name of the file to check the length of
 * @return the length of the file in bytes
 * @throws IO_ERROR if there is an error checking the file length
 */
int fs_length(Store *store, char *filename)
{
  char buf[MAX_FILE_PATH];
  struct stat stt;

  if (stat(join_path(buf, store->dir.path, filename), &stt))
    RAISE(IO_ERROR, strerror(errno));

  return stt.st_size;
}

/**
 * Flush +len+ characters from +src+ to the output stream +os+
 *
 * @param os the output stream to write to
 * @param src the characters to write to the output stream
 * @param len the number of characters to write
 * @throws IO_ERROR if there is an error writing the characters
 */
void fso_flush_internal(OutStream *os, uchar *src, int len)
{
  if (len != fwrite(src, sizeof(uchar), len, (FILE *)os->file))
    RAISE(IO_ERROR, WRITE_ERROR_MSG);
}

/**
 * Seek +pos+ in the output stream
 *
 * @param os the output stream to seek in
 * @param pos the position to seek in the stream
 * @throws IO_ERROR if there is an error seeking in the output stream
 */
void fso_seek_internal(OutStream *os, int pos)
{
  if (fseek((FILE *)os->file, pos, SEEK_SET))
    RAISE(IO_ERROR, strerror(errno));
}

/**
 * Close any resources used by the output stream +os+
 *
 * @param os the output stream to seek in
 * @throws IO_ERROR if there is an closing the file
 */
void fso_close_internal(OutStream *os)
{
  if (fclose((FILE *)os->file))
    RAISE(IO_ERROR, strerror(errno));
}

/**
 * Allocate the resources needed for the output stream in the +store+ with the
 * name +filename+
 *
 * @param store the store to create the output stream in.
 * @param filename the name of the output stream
 * @return a newly allocated filestream
 * @throws IO_ERROR if there is an error opening the output stream resources
 */
OutStream *fs_create_output(Store *store, const char *filename)
{
  char buf[MAX_FILE_PATH];
  FILE *f = fopen(join_path(buf, store->dir.path, filename), "wb");
  OutStream *os;
  if (!f) RAISE(IO_ERROR, strerror(errno));

  os = os_create();
  os->file = f;
  os->flush_internal = &fso_flush_internal;
  os->seek_internal  = &fso_seek_internal;
  os->close_internal = &fso_close_internal;
  return os;
}

/**
 * Read +len+ characters from the input stream into the +offset+ position in
 * +b+, an array of unsigned characters.
 *
 * @param is the input stream to read from
 * @param b an array of characters which must be allocated with  at least
 *          +offset+ + +len+ bytes
 * @param offset the position in +b+ to start writing the bytes read
 * @param len the number of bytes to read
 * @throws IO_ERROR if there is an error reading from the input stream
 */
void fsi_read_internal(InStream *is, uchar *b, int offset, int len)
{
  int fd = (int)is->file;
  int pos = is_pos(is);
  if (pos != _lseek(fd, 0, SEEK_CUR)) {
    _lseek(fd, pos, SEEK_SET);
  }
  if (_read(fd, b, len) != len) {
    /* win: the wrong value can be returned for some reason so double check */
    if (_lseek(fd, 0, SEEK_CUR) != (pos + len)) {
      RAISE(EOF_ERROR, strerror(errno));
    }
  }
}

/**
 * Seek position +pos+ in input stream +is+
 *
 * @param is the input stream to seek in
 * @param pos the position to seek
 * @throws IO_ERROR if the seek fails
 */
void fsi_seek_internal(InStream *is, int pos)
{
  if (_lseek((int)is->file, pos, SEEK_SET) < 0) {
    RAISE(IO_ERROR, strerror(errno));
  }
}

/**
 * Close the resources allocated to the inputstream +is+
 *
 * @param is the input stream to close
 * @throws IO_ERROR if the close fails
 */
void fsi_close_internal(InStream *is)
{
  if (!is->is_clone) {
    if (_close((int)is->file)) {
      RAISE(IO_ERROR, strerror(errno));
    }
    free(is->d.path);
  }
}

/**
 * Returns the length of the input stream +is+
 *
 * @param is the input stream to get the length of
 * @throws IO_ERROR if there is an error getting the file length
 */
int fsi_length(InStream *is)
{
  struct _stat stt;
  if (_fstat((int)is->file, &stt)) {
    RAISE(IO_ERROR, strerror(errno));
  }
  return stt.st_size;
}

/**
 * Clone the input stream. Nothing to do for a file system input stream
 */
void fsi_clone_internal(InStream *is, InStream *new_is)
{ }

/**
 * Open an input stream in the +store+ with the name +filename+
 *
 * @param store the store to create the input stream in
 * @param filename the name of the input stream
 * @throws IO_ERROR if the input stream cannot be opened
 */
InStream *fs_open_input(Store *store, const char *filename)
{
  InStream *is;
  char buf[MAX_FILE_PATH];
  int fd = _open(join_path(buf, store->dir.path, filename), O_RDONLY);
  if (fd < 0) {
    RAISE(IO_ERROR, FILE_OPEN_ERROR_MSG);
  }
  
  is = is_create();
  is->file = (void *)fd;
  is->d.path = estrdup(buf);
  is->is_clone = false;
  is->read_internal  = &fsi_read_internal;
  is->seek_internal  = &fsi_seek_internal;
  is->close_internal = &fsi_close_internal;
  is->clone_internal = &fsi_clone_internal;
  is->length_internal= &fsi_length;
  return is;
}

#define LOCK_OBTAIN_TIMEOUT 5

/**
 * Obtain a lock on the lock +lock+
 *
 * @param lock the lock to obtain
 */
int fs_lock_obtain(Lock *lock)
{
  int f;
  int trys = LOCK_OBTAIN_TIMEOUT;
  while (((f = _open(lock->name, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR|S_IWUSR)) < 0) &&
    (trys > 0)) {
    trys--;
  }
  if (f >= 0) {
    _close(f);
    return true;
  } else {
    return false;
  }
}

/**
 * Returns true if +lock+ is locked. To test if the file is locked:wq
 *
 * @param lock the lock to test
 * @throws 
 */
int fs_lock_is_locked(Lock *lock)
{
  int f = _open(lock->name, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR|S_IWUSR);
  if (f >= 0) {
    if (_close(f) || remove(lock->name)) {
      RAISE(IO_ERROR, strerror(errno));
    }
    return false;
  } else {
    return true;
  }
}

void fs_lock_release(Lock *lock)
{
  remove(lock->name);
}

Lock *fs_open_lock(Store *store, char *lockname)
{
  Lock *lock = ALLOC(Lock);
  char lname[100];
  char buf[MAX_FILE_PATH];
  sprintf(lname, "%s%s.lck", LOCK_PREFIX, lockname);
  lock->name = estrdup(join_path(buf, store->dir.path, lname));
  lock->obtain = &fs_lock_obtain;
  lock->release = &fs_lock_release;
  lock->is_locked = &fs_lock_is_locked;
  return lock;
}

void fs_close_lock(Lock *lock)
{
  remove(lock->name);
  free(lock->name);
  free(lock);
}

#ifdef WIN32
static HshTable stores = {
  /* fill */0,
  /* used */0,
  /* mask */Hsh_MINSIZE - 1,
  /* table */stores.smalltable,
  /* smalltable */{0},
  /* lookup */&h_lookup_str,
  /* hash */NULL,
  /* eq */NULL,
  /* free_key */(free_ft)&dummy_free,
  /* free_value */(free_ft)&fs_destroy
};
#else
static HshTable stores = {
  fill:0,
  used:0,
  mask:Hsh_MINSIZE - 1,
  table:stores.smalltable,
  lookup:&h_lookup_str,
  hash:NULL,
  eq:NULL,
  free_key:(free_ft)&dummy_free,
  free_value:(free_ft)&fs_destroy
};
#endif

#ifndef FERRET_EXT
static mutex_t stores_mutex = MUTEX_INITIALIZER;
#endif

void fs_close_i(Store *store)
{
  mutex_lock(&stores_mutex);
  h_del(&stores, store->dir.path);
  mutex_unlock(&stores_mutex);
}

Store *fs_store_create(const char *pathname)
{
  Store *new_store = store_create();
  
  new_store->dir.path      = estrdup(pathname);
  new_store->touch         = &fs_touch;
  new_store->exists        = &fs_exists;
  new_store->remove        = &fs_remove;
  new_store->rename        = &fs_rename;
  new_store->count         = &fs_count;
  new_store->close_i       = &fs_close_i;
  new_store->clear         = &fs_clear;
  new_store->clear_all     = &fs_clear_all;
  new_store->clear_locks   = &fs_clear_locks;
  new_store->length        = &fs_length;
  new_store->each          = &fs_each;
  new_store->create_output = &fs_create_output;
  new_store->open_input    = &fs_open_input;
  new_store->open_lock     = &fs_open_lock;
  new_store->close_lock    = &fs_close_lock;
  return new_store;
}

Store *open_fs_store(const char *pathname)
{
  Store *store = NULL;
  
  mutex_lock(&stores_mutex);
  store = h_get(&stores, pathname);
  if (store) {
    mutex_lock(&store->mutex);
    store->ref_cnt++;
    mutex_unlock(&store->mutex);
  } else {
    store = fs_store_create(pathname);
    h_set(&stores, store->dir.path, store);
  }
  mutex_unlock(&stores_mutex);

  return store;
}
