#include <hash.h>
#include <global.h>
#include <string.h>

unsigned int hash(const char *str)
{
  register unsigned int h;
  register unsigned char *p;

  h = 0;
  for (p = (unsigned char *) str; *p != '\0'; p++)
    h = MULTIPLIER * h + *p;

  return h % NUM_ENTRIES;
}

HashEntry **ht_create()
{
  int i;
  HashEntry **new_ht = ALLOC_N(HashEntry *, NUM_ENTRIES);

  for (i = 0; i < NUM_ENTRIES; i++)
    new_ht[i] = NULL;

  return new_ht;
}

int ht_count(HashEntry **ht)
{
  int i, cnt=0;
  HashEntry *he;
  for (i = 0; i < NUM_ENTRIES; i++) {
    for (he = ht[i]; he != NULL; he = he->next) {
      cnt++;
    }
  }
  return cnt;
}

void ht_destroy_all(HashEntry **ht, void (*fn)(void *))
{
  int i;
  HashEntry *he, *he_tmp;
  for (i = 0; i < NUM_ENTRIES; i++) {
    for (he = ht[i]; he != NULL;) {
      he_tmp = he;
      he = he->next;
      fn(he_tmp->value);
      free(he_tmp->name);
      free(he_tmp);
    }
  }
  free(ht);
}

void ht_destroy(HashEntry **ht)
{
  int i;
  HashEntry *he, *he_tmp;
  for (i = 0; i < NUM_ENTRIES; i++) {
    for (he = ht[i]; he != NULL;) {
      he_tmp = he;
      he = he->next;
      free(he_tmp->name);
      free(he_tmp);
    }
  }
  free(ht);
}

void *ht_get(HashEntry **ht, char *name)
{
  HashEntry *he;
  int h = hash(name);

  for (he = ht[h]; he != NULL; he = he->next) {
    if (strcmp(name, he->name) == 0) {
      return he->value;
    }
  }

  return NULL;
}

void ht_set(HashEntry **ht, char *name, void *value)
{
  HashEntry *he;
  int h = hash(name);

  for (he = ht[h]; he != NULL; he = he->next) {
    if (strcmp(name, he->name) == 0) {
      he->value = value;
      return;
    }
  }

  he = ALLOC(HashEntry);
  he->name = estrdup(name);
  he->value = value;
  he->next = ht[h];
  ht[h] = he;
  return;
}

void *ht_delete(HashEntry **ht, char *name)
{
  HashEntry *he, *tmp;
  void *rval;

  int h = hash(name);

  if (ht[h] == NULL)
    return NULL;

  if (strcmp(ht[h]->name, name) == 0) {
    tmp = ht[h];
    ht[h] = tmp->next;
    rval = tmp->value;
    free(tmp->name);
    free(tmp);
    return rval;
  }

  for (he = ht[h]; he != NULL; he = he->next) {
    if (he->next != NULL && strcmp(name, he->next->name) == 0) {
      tmp = he->next;
      he->next = tmp->next;
      rval = tmp->value;
      free(tmp->name);
      free(tmp);
      return rval;
    }
  }
  return NULL;
}

/****************************************************************************
 *
 * HshTable
 *
 ****************************************************************************/

char *dummy_key = "";

#define PERTURB_SHIFT 5

#ifdef VALGRIND
#define MAX_FREE_HASH_TABLES 0
#else
#define MAX_FREE_HASH_TABLES 80
#endif

static HshTable *free_hts[MAX_FREE_HASH_TABLES];
static int num_free_hts = 0;

inline unsigned int str_hash(const char *const str)
{
  register unsigned int h=0;
  register unsigned char *p = (unsigned char *)str;

  for (; *p; p++)
    h = 37 * h + *p;

  return h;
}

HshEntry *h_lookup_str(HshTable *ht, register const void *key_p)
{
  register char *key = (char *)key_p;
  register unsigned int hash = str_hash(key);
  register unsigned int perturb;
  register int mask = ht->mask;
  register HshEntry *he0 = ht->table;
  register int i = hash & mask;
  register HshEntry *he = &he0[i];
  register HshEntry *freeslot;

  if (he->key == NULL || he->key == key) {
    he->hash = hash;
    return he;
  }
  if (he->key == dummy_key) {
    freeslot = he;
  } else {
    if ((he->hash == hash) && (strcmp(he->key, key) == 0)) {
      return he;
    }
    freeslot = NULL;
  }

  for (perturb = hash; ; perturb >>= PERTURB_SHIFT) {
    i = (i << 2) + i + perturb + 1;
    he = &he0[i & mask];
    if (he->key == NULL) {
      if (freeslot != NULL) he = freeslot;
      he->hash = hash;
      return he;
    }
    if (he->key == key
        || (he->hash == hash
          && he->key != dummy_key
          && strcmp(he->key, key) == 0))
      return he;
    if (he->key == dummy_key && freeslot == NULL)
      freeslot = he;
  }
}

typedef int (*eq_func)(const void *key1, const void *key2);

HshEntry *h_lookup(HshTable *ht, register const void *key)
{
  register unsigned int hash = ht->hash(key);
  register unsigned int perturb;
  register int mask = ht->mask;
  register HshEntry *he0 = ht->table;
  register int i = hash & mask;
  register HshEntry *he = &he0[i];
  register HshEntry *freeslot;
  eq_func eq = ht->eq;

  if (he->key == NULL || he->key == key) {
    he->hash = hash;
    return he;
  }
  if (he->key == dummy_key) {
    freeslot = he;
  } else {
    if ((he->hash == hash) && eq(he->key, key))
      return he;
    freeslot = NULL;
  }

  for (perturb = hash; ; perturb >>= PERTURB_SHIFT) {
    i = (i << 2) + i + perturb + 1;
    he = &he0[i & mask];
    if (he->key == NULL) {
      if (freeslot != NULL) he = freeslot;
      he->hash = hash;
      return he;
    }
    if (he->key == key
        || (he->hash == hash
          && he->key != dummy_key
          && eq(he->key, key)))
      return he;
    if (he->key == dummy_key && freeslot == NULL)
      freeslot = he;
  }
}

typedef void (*free_func)(void *);
void dummy_free(void *p)
{ }

HshTable *h_new_str(free_ft free_key, free_ft free_value)
{
  HshTable *ht;
  if (num_free_hts > 0) {
    ht = free_hts[--num_free_hts];
  } else {
    ht = ALLOC(HshTable);
  }
  ht->fill = 0;
  ht->used = 0;
  ht->mask = Hsh_MINSIZE - 1;
  ht->table = ht->smalltable;
  memset(ht->smalltable, 0, sizeof(ht->smalltable));
  ht->lookup = &h_lookup_str;
  ht->free_key = free_key != NULL ? free_key : &dummy_free;
  ht->free_value = free_value != NULL ? free_value : &dummy_free;
  return ht;
}

HshTable *h_new(hash_ft hash, eq_ft eq, free_ft free_key, free_ft free_value)
{
  HshTable *ht = h_new_str(free_key, free_value);

  ht->lookup = &h_lookup;
  ht->eq = eq;
  ht->hash = hash;
  return ht;
}

void h_clear(HshTable *ht)
{
  int i;
  HshEntry *he;
  free_func free_key = ht->free_key;
  free_func free_value = ht->free_value;

  /* Clear all the hash values and keys as necessary */
  for (i = 0; i <= ht->mask; i++) {
    he = &ht->table[i];
    if (he->key != NULL && he->key != dummy_key) {
      free_value(he->value);
      free_key(he->key);
    }
    he->key = NULL;
  }
  ht->used = 0;
  ht->fill = 0;
  /* for some reason the following line doesn't seem to help */
  /* ZEROSET(ht->table, HshEntry, ht->mask + 1); */
}

void h_destroy(HshTable *ht)
{
  h_clear(ht);

  /* if a new table was created, be sure to free it */
  if (ht->table != ht->smalltable) free(ht->table);

  if (num_free_hts < MAX_FREE_HASH_TABLES) {
    free_hts[num_free_hts++] = ht;
  } else {
    free(ht);
  }
}

void *h_get(HshTable *ht, const void *key)
{
  return ht->lookup(ht, key)->value;
}

int h_del(HshTable *ht, const void *key)
{
  HshEntry *he = ht->lookup(ht, key);

  if (he->key != NULL && he->key != dummy_key) {
    ht->free_key(he->key);
    ht->free_value(he->value);
    he->key = dummy_key;
    he->value = NULL;
    ht->used--;
    return true;
  } else {
    return false;
  }
}

void *h_rem(HshTable *ht, const void *key, bool destory_key)
{
  void *val;
  HshEntry *he = ht->lookup(ht, key);

  if (he->key != NULL && he->key != dummy_key) {
    if (destory_key)
      ht->free_key(he->key);

    he->key = dummy_key;
    val = he->value;
    he->value = NULL;
    ht->used--;
    return val;
  } else {
    return NULL;
  }
}

int h_resize(HshTable *ht, int min_newsize) {
  HshEntry smallcopy[Hsh_MINSIZE];
  HshEntry *oldtable;
  HshEntry *he_old, *he_new;
  int newsize, i;

  for (newsize = Hsh_MINSIZE; newsize < min_newsize; newsize <<= 1)
    ;
  
  oldtable = ht->table;
  if (newsize == Hsh_MINSIZE) {
    if (ht->table == ht->smalltable) {
      /* need to copy the d*(int *)ata out so we can rebuild the table into
       * the same space */
      memcpy(smallcopy, ht->smalltable, sizeof(smallcopy));
      oldtable = smallcopy;
    } else {
      ht->table = ht->smalltable;
    } 
  } else {
    ht->table = ALLOC_N(HshEntry, newsize);
  }
  memset(ht->table, 0, sizeof(HshEntry) * newsize);
  i = ht->used;
  ht->fill = ht->used;
  ht->mask = newsize - 1;

  for (he_old = oldtable; i > 0; he_old++) {
    if (he_old->value != NULL) {/* active entry */
      he_new = ht->lookup(ht, he_old->key);
      he_new->key = he_old->key;
      he_new->value = he_old->value;
      i--;
    } /* else empty entry so nothing to do */
  }
  if (oldtable != smallcopy && oldtable != ht->smalltable)
    free(oldtable);
  return 0;
}

int h_set(HshTable *ht, const void *key, void *value)
{
  int ret_val = HASH_KEY_DOES_NOT_EXIST;
  HshEntry *he = ht->lookup(ht, key);
  int fill = ht->fill;
  if (he->key == NULL) {
    ht->fill++;
    ht->used++;
  } else if (he->key == dummy_key) {
    ht->used++;
  } else if (he->key != key) {
    ht->free_key(he->key);
    if (he->value != value) ht->free_value(he->value);
    ret_val = HASH_KEY_EQUAL;
  } else {
    if (he->value != value) ht->free_value(he->value);
    ret_val = HASH_KEY_SAME;
  }
  he->key = (void *)key;
  he->value = value;
  if ((ht->fill > fill) && (ht->fill * 3 > ht->mask * 2))
    h_resize(ht, ht->used * ((ht->used > SLOW_DOWN) ? 4 : 2));
  return ret_val;
}

int h_has_key(HshTable *ht, const void *key)
{
  HshEntry *he = ht->lookup(ht, key);
  if (he->key == NULL || he->key == dummy_key) {
    return HASH_KEY_DOES_NOT_EXIST;
  } else if (he->key == key) {
    return HASH_KEY_SAME;
  } else {
    return HASH_KEY_EQUAL;
  }
}

void h_each(HshTable *ht,
    void (*each_kv)(void *key, void *value, void *arg),
    void *arg)
{
  HshEntry *he;
  int i = ht->used;
  for (he = ht->table; i > 0; he++) {
    if (he->value != NULL) {/* active entry */
      each_kv(he->key, he->value, arg);
      i--;
    }
  }
}

HshTable *h_clone(HshTable *ht,
    h_clone_func_t clone_key,
    h_clone_func_t clone_value)
{
  void *key, *value;
  HshEntry *he;
  int i = ht->used;
  HshTable *ht_clone;

  if (ht->lookup == &h_lookup_str) {
    ht_clone = h_new_str(ht->free_key, ht->free_value);
  } else {
    ht_clone = h_new(ht->hash, ht->eq, ht->free_key, ht->free_value);
  }

  for (he = ht->table; i > 0; he++) {
    if (he->value != NULL) {/* active entry */
      key = clone_key ? clone_key(he->key) : he->key;
      value = clone_value ? clone_value(he->value) : he->value;
      h_set(ht_clone, key, value);
      i--;
    }
  }
  return ht_clone;
}

