#include <store.h>
#include <string.h>

static char * const STORE_EOF_ERROR_MSG = "EOF Error when trying to refill";

Buffer *buf_create()
{
  Buffer *buf = ALLOC(Buffer);
  buf->start = 0;
  buf->pos = 0;
  buf->len = 0;
  return buf;
}

void buf_destroy(Buffer *buf)
{
  free(buf);
}

OutStream *os_create()
{
  OutStream *os = ALLOC(OutStream);
  os->buf.start = 0;
  os->buf.pos = 0;
  os->buf.len = 0;
  return os;
}

inline void os_flush(OutStream *os)
{
  os->flush_internal(os, os->buf.buf, os->buf.pos);
  os->buf.start += os->buf.pos;
  os->buf.pos = 0;
}

void os_close(OutStream *os)
{
  os_flush(os);
  os->close_internal(os);
  free(os);
}

int os_pos(OutStream *os)
{
  return os->buf.start + os->buf.pos;
}

void os_seek(OutStream *os, int new_pos)
{
  os_flush(os);
  os->buf.start = new_pos;
  os->seek_internal(os, new_pos);
}

#define write_byte(os, b) os->buf.buf[os->buf.pos++] = b

inline void os_write_byte(OutStream *os, uchar b)
{
  if (os->buf.pos >= BUFFER_SIZE) {
    os_flush(os);
  }
  write_byte(os, b);
}

void os_write_bytes(OutStream *os, uchar *b, int len)
{
  if (os->buf.pos > 0) { /* flush buffer */
    os_flush(os);
  }

  if (len < BUFFER_SIZE) {
    os->flush_internal(os, b, len);
    os->buf.start += len;
  } else {
    int pos = 0;
    int size;
    while (pos < len) {
      if (len - pos < BUFFER_SIZE) {
        size = len - pos;
      } else {
        size = BUFFER_SIZE;
      }
      os->flush_internal(os, b + pos, size);
      pos += size;
      os->buf.start += size;
    }
  }
}

InStream *is_create()
{
  InStream *is = ALLOC(InStream);
  is->buf.start = 0;
  is->buf.pos = 0;
  is->buf.len = 0;
  return is;
}

void is_refill(InStream *is)
{
  int start = is->buf.start + is->buf.pos;
  int last = start + BUFFER_SIZE;
  int flen = is->length_internal(is);
  if (last > flen) { /* don't read past EOF */
    last = flen;
  }

  is->buf.len = last - start;
  if (is->buf.len <= 0) {
    RAISE(EOF_ERROR, STORE_EOF_ERROR_MSG);
  }

  is->read_internal(is, is->buf.buf, 0, is->buf.len);

  is->buf.start = start;
  is->buf.pos = 0;
}

#define read_byte(is) is->buf.buf[is->buf.pos++]
inline uchar is_read_byte(InStream *is)
{
  if (is->buf.pos >= is->buf.len) {
    is_refill(is);
  }

  return read_byte(is);
}

int is_pos(InStream *is)
{
  return is->buf.start + is->buf.pos;
}

uchar *is_read_bytes(InStream *is, uchar *b, int offset, int len)
{
  int i, start;
  if ((offset + len) < BUFFER_SIZE) {
    for (i = offset; i < offset + len; i++) {
      b[i] = is_read_byte(is);
    }
  } else {                           // read all-at-once
    start = is_pos(is);
    is->seek_internal(is, start);
    is->read_internal(is, b, offset, len);

    is->buf.start = start + len;        // adjust stream variables
    is->buf.pos = 0;
    is->buf.len = 0;                    // trigger refill on read
  }
  return b;
}

void is_seek(InStream *is, int pos)
{
  if (pos >= is->buf.start && pos < (is->buf.start + is->buf.len)) {
    is->buf.pos = pos - is->buf.start;  // seek within buffer
  } else {
    is->buf.start = pos;
    is->buf.pos = 0;
    is->buf.len = 0;               // trigger refill() on read()
    is->seek_internal(is, pos);
  }
}

void is_close(InStream *is)
{
  is->close_internal(is);
  free(is);
}

InStream *is_clone(InStream *is) 
{
  InStream *new_index_i = ALLOC(InStream);
  memcpy(new_index_i, is, sizeof(InStream));
  new_index_i->is_clone = true;
  is->clone_internal(is, new_index_i);
  return new_index_i;
}

int
is_read_int(InStream *is)
{
  return ((int)is_read_byte(is) << 24) |
         ((int)is_read_byte(is) << 16) |
         ((int)is_read_byte(is) <<  8) |
          (int)is_read_byte(is);
}

llong
is_read_long(InStream *is)
{
  return ((llong)is_read_byte(is) << 56) |
         ((llong)is_read_byte(is) << 48) |
         ((llong)is_read_byte(is) << 40) |
         ((llong)is_read_byte(is) << 32) |
         ((llong)is_read_byte(is) << 24) |
         ((llong)is_read_byte(is) << 16) |
         ((llong)is_read_byte(is) <<  8) |
          (llong)is_read_byte(is);
}

unsigned int
is_read_uint(InStream *is)
{
  return ((unsigned int)is_read_byte(is) << 24) |
         ((unsigned int)is_read_byte(is) << 16) |
         ((unsigned int)is_read_byte(is) <<  8) |
          (unsigned int)is_read_byte(is);
}

ullong
is_read_ulong(InStream *is)
{
  return ((ullong)is_read_byte(is) << 56) |
         ((ullong)is_read_byte(is) << 48) |
         ((ullong)is_read_byte(is) << 40) |
         ((ullong)is_read_byte(is) << 32) |
         ((ullong)is_read_byte(is) << 24) |
         ((ullong)is_read_byte(is) << 16) |
         ((ullong)is_read_byte(is) <<  8) |
          (ullong)is_read_byte(is);
}

/* optimized to use unchecked read_byte if there is definitely space */
inline ullong
is_read_vint(InStream *is)
{
  register ullong res, b;
  register int shift = 7;

  if (is->buf.pos > (is->buf.len - VINT_MAX_LEN)) {
    b = is_read_byte(is);
    res = b & 0x7F; // 0x7F = 0b01111111
    
    while ((b & 0x80) != 0) {// 0x80 = 0b10000000
      b = is_read_byte(is);
      res |= (b & 0x7F) << shift;
      shift += 7;
    }
  } else { // unchecked
    b = read_byte(is);
    res = b & 0x7F; // 0x7F = 0b01111111
    
    while ((b & 0x80) != 0) {// 0x80 = 0b10000000
      b = read_byte(is);
      res |= (b & 0x7F) << shift;
      shift += 7;
    }
  }
  
  return res;
}

inline void
is_skip_vints(InStream *is, register int cnt)
{
  for (; cnt > 0; cnt--) {
    while ((is_read_byte(is) & 0x80) != 0) {
    }
  }
}

inline void
is_read_chars(InStream *is, char* buffer, int off, int len)
{
  int end, i;
  
  end = off + len;
  
  for(i = off; i < end; i++) {
	buffer[i] = is_read_byte(is);
  }
}

char *
is_read_string(InStream *is)
{
  register int length = (int)is_read_vint(is);
  char *str = ALLOC_N(char, length + 1);
  str[length] = '\0';

  if (is->buf.pos > (is->buf.len - length)) {
    register int i;
    for(i = 0; i < length; i++) {
      str[i] = is_read_byte(is);
    }
  } else { // unchecked
    memcpy(str, is->buf.buf + is->buf.pos, length);
    is->buf.pos += length;
  }
  //is_read_chars(is, str, 0, length);

  return str;
}

void
os_write_int(OutStream *os, int l)
{
  os_write_byte(os, (uchar)((l >> 24) & 0xFF));
  os_write_byte(os, (uchar)((l >> 16) & 0xFF));
  os_write_byte(os, (uchar)((l >>  8) & 0xFF));
  os_write_byte(os, (uchar)(l & 0xFF));
}

void
os_write_long(OutStream *os, llong l)
{
  os_write_byte(os, (uchar)((l >> 56) & 0xFF));
  os_write_byte(os, (uchar)((l >> 48) & 0xFF));
  os_write_byte(os, (uchar)((l >> 40) & 0xFF));
  os_write_byte(os, (uchar)((l >> 32) & 0xFF));
  os_write_byte(os, (uchar)((l >> 24) & 0xFF));
  os_write_byte(os, (uchar)((l >> 16) & 0xFF));
  os_write_byte(os, (uchar)((l >>  8) & 0xFF));
  os_write_byte(os, (uchar)(l & 0xFF));
}

void
os_write_uint(OutStream *os, unsigned int l)
{
  os_write_byte(os, (uchar)((l >> 24) & 0xFF));
  os_write_byte(os, (uchar)((l >> 16) & 0xFF));
  os_write_byte(os, (uchar)((l >>  8) & 0xFF));
  os_write_byte(os, (uchar)(l & 0xFF));
}

void
os_write_ulong(OutStream *os, ullong l)
{
  os_write_byte(os, (uchar)((l >> 56) & 0xFF));
  os_write_byte(os, (uchar)((l >> 48) & 0xFF));
  os_write_byte(os, (uchar)((l >> 40) & 0xFF));
  os_write_byte(os, (uchar)((l >> 32) & 0xFF));
  os_write_byte(os, (uchar)((l >> 24) & 0xFF));
  os_write_byte(os, (uchar)((l >> 16) & 0xFF));
  os_write_byte(os, (uchar)((l >>  8) & 0xFF));
  os_write_byte(os, (uchar)(l & 0xFF));
}

/* optimized to use an unchecked write if there is space */
inline void
os_write_vint(OutStream *os, register ullong i)
{
  if (os->buf.pos > VINT_END) {
    while (i > 127) {
      os_write_byte(os, (uchar)((i & 0x7f) | 0x80));
      i >>= 7;
    }
    os_write_byte(os, (uchar)(i));
  } else {
    while (i > 127) {
      write_byte(os, (uchar)((i & 0x7f) | 0x80));
      i >>= 7;
    }
    write_byte(os, (uchar)(i));
  }
}

void
os_write_chars(OutStream *os, char *buf, int start, int length)
{
  int i;

  for (i = start; i < start + length; i++) {
    os_write_byte(os, buf[i]);
  }
}

void
os_write_string(OutStream *os, char *str)
{
  int len = (int)strlen(str);
  os_write_vint(os, len);

  os_write_chars(os, str, 0, len);
}

int file_is_lock(char *filename)
{
  int start = (int)strlen(filename) - 4;
  return ((start > 0) && (strcmp(".lck", &filename[start]) == 0));
}
