#ifndef WIN32

#include "global.h"
#include "store.h"
#include <dirent.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

/**
 * Create a filepath for a file in the store using the operating systems
 * default file seperator.
 */
char *join_path(char *buf, const char *base, const char *filename)
{
  sprintf(buf, "%s/%s", base, filename);
  return buf;
}

bool exists(char *path)
{
  int fd = open(path, 0);
  if (fd < 0) {
    if (errno != ENOENT) {
      RAISE(IO_ERROR, strerror(errno));
    }
    return false;
  }
  close(fd);
  return true;
}

int fcount(char *path)
{
  int cnt = 0;
  struct dirent *de;
  DIR *d = opendir(path);

  if (!d) RAISE(IO_ERROR, strerror(errno));

  while ((de = readdir(d)) != NULL) {
    if (de->d_name[0] != '.') {
      cnt++;
    }
  }
  closedir(d);

  return cnt;
}

void dir_each(char *path, void (*func)(char *fname, void *arg), void *arg)
{
  struct dirent *de;
  DIR *d = opendir(path);

  if (!d) RAISE(IO_ERROR, strerror(errno));

  while ((de = readdir(d)) != NULL) {
    if (de->d_name[0] != '.' && !file_is_lock(de->d_name)) {
      func(de->d_name, arg);
    }
  }
  closedir(d);
}

/**
 * Clear all the locks in the store.
 *
 * @param store the store to clear the locks from
 * @throws IO_ERROR if there is an error opening the directory
 */
void fs_clear_locks(Store *store)
{
  struct dirent *de;
  DIR *d = opendir(store->dir.path);

  if (!d) RAISE(IO_ERROR, strerror(errno));

  while ((de = readdir(d)) != NULL) {
    if (file_is_lock(de->d_name)) {
      char buf[MAX_FILE_PATH];
      remove(join_path(buf, store->dir.path, de->d_name));
    }
  }
  closedir(d);
}

/**
 * Clear all files from the store except the lock files.
 *
 * @param store the store to clear all the files from
 * @throws IO_ERROR if there is an error deleting the files
 */
void fs_clear(Store *store)
{
  struct dirent *de;
  DIR *d = opendir(store->dir.path);

  if (!d) RAISE(IO_ERROR, strerror(errno));

  while ((de = readdir(d)) != NULL) {
    if (de->d_name[0] != '.' && !file_is_lock(de->d_name)) {
      char buf[MAX_FILE_PATH];
      remove(join_path(buf, store->dir.path, de->d_name));
    }
  }
  closedir(d);
}

/**
 * Clear all files from the store including the lock files.
 *
 * @param store the store to clear all the files from
 * @throws IO_ERROR if there is an error deleting the files
 */
void fs_clear_all(Store *store)
{
  struct dirent *de;
  DIR *d = opendir(store->dir.path);

  if (!d) RAISE(IO_ERROR, strerror(errno));

  while ((de = readdir(d)) != NULL) {
    if (de->d_name[0] != '.') {
      char buf[MAX_FILE_PATH];
      remove(join_path(buf, store->dir.path, de->d_name));
    }
  }
  closedir(d);
}

#endif

