#include <priorityqueue.h>
  
PriorityQueue *pq_create(int max_size, bool (*less_than)(void *p1, void *p2))
{
  PriorityQueue *pq = ALLOC(PriorityQueue);
  pq->count = 0;
  pq->size = max_size;
  pq->heap = ALLOC_N(void *, (max_size + 1));
  pq->lt = less_than;
  pq->free_elem = &free;
  return pq;
}

void pq_destroy(PriorityQueue *pq)
{
  free(pq->heap);
  free(pq);
}

void pq_up(PriorityQueue *pq)
{
	void **heap = pq->heap;
  void *node;
	int i = pq->count;
	int j = i >> 1;

  node = heap[i];

	while ((j > 0) && pq->lt(node, heap[j])) {
		heap[i] = heap[j];
		i = j;
		j = j >> 1;
	}
  heap[i] = node;
}

void pq_down(PriorityQueue *pq)
{
  register int i = 1;
  register int j = 2; //i << 1;
  register int k = 3; //j + 1;
  register int count = pq->count;
  void **heap = pq->heap;
  void *node = heap[i]; // save top node

  if ((k <= count) && (pq->lt(heap[k], heap[j])))
    j = k;

  while ((j <= count) && pq->lt(heap[j], node)) {
    heap[i] = heap[j];      // shift up child
    i = j;
    j = i << 1;
    k = j + 1;
    if ((k <= count) && pq->lt(heap[k], heap[j]))
      j = k;
  }
  heap[i] = node;
}

void pq_push(PriorityQueue *pq, void *elem) 
{
  pq->count++;
  pq->heap[pq->count] = elem;
  pq_up(pq);
}

void *pq_top(PriorityQueue *pq)
{
  return pq->heap[1];
}

void *pq_pop(PriorityQueue *pq)
{
  if (pq->count > 0) {
    void *result = pq->heap[1];        // save first value
    pq->heap[1] = pq->heap[pq->count]; // move last to first
    pq->heap[pq->count] = NULL;
    pq->count--;
    pq_down(pq);                       // adjust heap
    return result;
  } else {
    return NULL;
  }
}

void pq_clear(PriorityQueue *pq)
{
  int i;
  for (i = 1; i <= pq->count; i++) {
    pq->free_elem(pq->heap[i]);
    pq->heap[i] = NULL;
  }
  pq->count = 0;
}

int pq_insert(PriorityQueue *pq, void *elem) 
{
  if (pq->count < pq->size) {
    pq_push(pq, elem);
    return true;
  } else if (pq->count > 0 && pq->lt(pq_top(pq), elem)) {
    pq->free_elem(pq->heap[1]);
    pq->heap[1] = elem;
    pq_down(pq);
    return true;
  } else {
    pq->free_elem(elem);
    return false;
  }
}

/*****************************************************************************
 *
 * PriorityQueue2
 *
 *****************************************************************************/

PriorityQueue2 *pq2_create(int max_size,
    bool (*less_than)(PriorityQueue2 *pq, void *p1, void *p2),
    free_ft destroy)
{
  PriorityQueue2 *pq = ALLOC(PriorityQueue2);
  pq->count = 0;
  pq->size = max_size;
  pq->heap = ALLOC_N(void *, (max_size + 1));
  pq->lt = less_than;
  pq->free_elem = &free;
  pq->destroy = destroy;
  return pq;
}

void pq2_destroy(PriorityQueue2 *pq)
{
  free(pq->heap);
  free(pq);
}

void pq2_up(PriorityQueue2 *pq)
{
  void **heap = pq->heap;
  void *node;
  int i = pq->count;
	int j = i >> 1;
  
  node = heap[i];

	while ((j > 0) && pq->lt(pq, node, heap[j])) {
		heap[i] = heap[j];
		i = j;
		j = j >> 1;
	}
  heap[i] = node;
}

void pq2_down(PriorityQueue2 *pq)
{
  register int i = 1;
  register int j = 2; //i << 1;
  register int k = 3; //j + 1;
  register int count = pq->count;
  void **heap = pq->heap;
  void *node = heap[i]; // save top node

  if ((k <= count) && (pq->lt(pq, heap[k], heap[j])))
    j = k;

  while ((j <= count) && pq->lt(pq, heap[j], node)) {
    heap[i] = heap[j];      // shift up child
    i = j;
    j = i << 1;
    k = j + 1;
    if ((k <= count) && pq->lt(pq, heap[k], heap[j]))
      j = k;
  }
  heap[i] = node;
}

void pq2_push(PriorityQueue2 *pq, void *elem) 
{
  pq->count++;
  pq->heap[pq->count] = elem;
  pq2_up(pq);
}

void *pq2_top(PriorityQueue2 *pq)
{
  return pq->heap[1];
}

void *pq2_pop(PriorityQueue2 *pq)
{
  if (pq->count > 0) {
    void *result = pq->heap[1];        // save first value
    pq->heap[1] = pq->heap[pq->count]; // move last to first
    pq->heap[pq->count] = NULL;
    pq->count--;
    pq2_down(pq);                       // adjust heap
    return result;
  } else {
    return NULL;
  }
}

void pq2_clear(PriorityQueue2 *pq)
{
  int i;
  for (i = 1; i <= pq->count; i++) {
    pq->free_elem(pq->heap[i]);
    pq->heap[i] = NULL;
  }
  pq->count = 0;
}

int pq2_insert(PriorityQueue2 *pq, void *elem) 
{
  if (pq->count < pq->size) {
    pq2_push(pq, elem);
    return true;
  } else if (pq->count > 0 && pq->lt(pq, pq2_top(pq), elem)) {
    pq->free_elem(pq->heap[1]);
    pq->heap[1] = elem;
    pq2_down(pq);
    return true;
  } else {
    pq->free_elem(elem);
    return false;
  }
}
