#include "search.h"
#include <string.h>

/***************************************************************************
 *
 * Weight
 *
 ***************************************************************************/

char *csw_to_s(Weight *self)
{
  return strfmt("ConstantScoreWeight(%f)", self->value);
}

Explanation *csw_explain(Weight *self, IndexReader *ir, int doc_num)
{
  Filter *filter = (Filter *)self->query->data;
  Explanation *expl;
  char *filter_str = filter->to_s(filter);
  BitVector *bv = filt_get_bv(filter, ir);

  if (bv_get(bv, doc_num)) {
    expl = expl_create(self->value,
        strfmt("ConstantScoreQuery(%s), product of:", filter_str));
    expl_add_detail(expl, expl_create(self->query->boost, estrdup("boost")));
    expl_add_detail(expl, expl_create(self->qnorm, estrdup("query_norm")));
  } else {
    expl = expl_create(self->value,
        strfmt("ConstantScoreQuery(%s), does not match id %d",
          filter_str, doc_num));
  }
  free(filter_str);
  return expl;
}

Weight *csw_create(Query *query, Searcher *searcher)
{
  Weight *self = w_create(query);
  self->scorer    = &cssc_create;
  self->explain   = &csw_explain;
  self->to_s      = &csw_to_s;
  self->sum_of_squared_weights = &w_sum_of_squared_weights;

  self->similarity = query->get_similarity(query, searcher);
  self->idf = 1.0;

  return self;
}

/***************************************************************************
 *
 * ConstantScoreQuery
 *
 ***************************************************************************/

char *csq_to_s(Query *self, char *field)
{
  Filter *filter = (Filter *)self->data;
  char *filter_str = filter->to_s(filter);
  char *buffer;
  if (self->boost == 1.0) {
    buffer = strfmt("ConstantScore(%s)", filter_str);
  } else {
    buffer = strfmt("ConstantScore(%s)^%f", filter_str, self->boost);
  }
  free(filter_str);
  return buffer;;
}

void csq_destroy(Query *self)
{
  if (self->destroy_all) {
    Filter *filter = (Filter *)self->data;
    filter->destroy(filter);
  }
  q_destroy_i(self);
}

static uint csq_hash(Query *self)
{
  return filt_hash((Filter *)self->data);
}

static int csq_eq(Query *self, Query *o)
{
  return filt_eq((Filter *)self->data, (Filter *)o->data);
}

Query *csq_create(Filter *filter)
{
  Query *self = q_create();
  self->data = filter;

  self->type = CONSTANT_QUERY;
  self->to_s = &csq_to_s;
  self->hash = &csq_hash;
  self->eq = &csq_eq;
  self->destroy_i = &csq_destroy;
  self->create_weight_i = &csw_create;

  return self;
}

/***************************************************************************
 *
 * ConstantScoreScorer
 *
 ***************************************************************************/

float cssc_score(Scorer *self)
{
  return ((ConstantScoreScorer *)self->data)->score;
}

bool cssc_next(Scorer *self)
{
  BitVector *bv = ((ConstantScoreScorer *)self->data)->bv;
  return ((self->doc = bv_scan_next(bv)) >= 0);
}

bool cssc_skip_to(Scorer *self, int doc_num)
{
  BitVector *bv = ((ConstantScoreScorer *)self->data)->bv;
  return ((self->doc = bv_scan_next_from(bv, doc_num)) >= 0);
}

Explanation *cssc_explain(Scorer *self, int doc_num)
{
  return expl_create(1.0, estrdup("ConstantScoreScorer"));
}

Scorer *cssc_create(Weight *weight, IndexReader *ir)
{
  Scorer *self = scorer_create(weight->similarity);
  Filter *filter = (Filter *)weight->query->data;
  ConstantScoreScorer *cssc = ALLOC(ConstantScoreScorer);
  ZEROSET(cssc, ConstantScoreScorer, 1);
  self->data = cssc;
  cssc->score = weight->value;
  cssc->bv = filt_get_bv(filter, ir);

  self->score = &cssc_score;
  self->next = &cssc_next;
  self->skip_to = &cssc_skip_to;
  self->explain = &cssc_explain;
  self->destroy = &scorer_destroy_i;
  return self;
}
