#include "search.h"
#include <string.h>

Scorer *fqsc_create(Scorer *scorer, BitVector *bv, Similarity *sim);

/***************************************************************************
 *
 * Weight
 *
 ***************************************************************************/

char *fqw_to_s(Weight *self)
{
  return strfmt("FilteredQueryWeight(%f)", self->value);
}

float fqw_sum_of_squared_weights(Weight *self)
{
  Weight *sw = (Weight *)self->data;
  return sw->sum_of_squared_weights(sw);
}

void fqw_normalize(Weight *self, float normalization_factor)
{
  Weight *sw = (Weight *)self->data;
  sw->normalize(sw, normalization_factor);
}

float fqw_get_value(Weight *self)
{
  Weight *sw = (Weight *)self->data;
  return sw->get_value(sw);
}

Explanation *fqw_explain(Weight *self, IndexReader *ir, int doc_num)
{
  Weight *sw = (Weight *)self->data;
  return sw->explain(sw, ir, doc_num);
}

Scorer *fqw_scorer(Weight *self, IndexReader *ir)
{
  Weight *sw = (Weight *)self->data;
  Scorer *scorer = sw->scorer(sw, ir);
  Filter *filter = ((FilteredQuery *)self->query->data)->filter;

  return fqsc_create(scorer, filter->get_bv(filter, ir), self->similarity);
}

void fqw_destroy(Weight *self)
{
  Weight *sw = (Weight *)self->data;
  sw->destroy(sw);
  w_destroy(self);
}

Weight *fqw_create(Query *query, Weight *sub_weight, Similarity *sim)
{
  Weight *self = w_create(query);

  self->data = sub_weight;

  self->get_value = &fqw_get_value;
  self->normalize = &fqw_normalize;
  self->scorer    = &fqw_scorer;
  self->explain   = &fqw_explain;
  self->to_s      = &fqw_to_s;
  self->destroy   = &fqw_destroy;
  self->sum_of_squared_weights = &fqw_sum_of_squared_weights;

  self->similarity = sim;
  self->idf = 1.0;
  self->value = sub_weight->value;

  return self;
}

/***************************************************************************
 *
 * FilteredQueryQuery
 *
 ***************************************************************************/

char *fq_to_s(Query *self, char *field)
{
  FilteredQuery *fq = (FilteredQuery *)self->data;
  char *filter_str = fq->filter->to_s(fq->filter);
  char *query_str = fq->query->to_s(fq->query, field);
  char *buffer;
  if (self->boost == 1.0) {
    buffer = strfmt("FilteredQuery(query:%s, filter:%s)",
        query_str, filter_str);
  } else {
    buffer = strfmt("FilteredQuery(query:%s, filter:%s)^%f",
        query_str, filter_str, self->boost);
  }
  free(filter_str);
  free(query_str);
  return buffer;;
}

void fq_destroy(Query *self)
{
  if (self->destroy_all) {
    FilteredQuery *fq = (FilteredQuery *)self->data;
    fq->filter->destroy(fq->filter);
    q_deref(fq->query);
  }
  free(self->data);
  q_destroy_i(self);
}

Weight *fq_create_weight(Query *self, Searcher *searcher)
{
  Query *sub_query = ((FilteredQuery *)self->data)->query;
  return fqw_create(self, q_weight(sub_query, searcher),
      searcher->similarity);
}

Query *fq_create(Query *query, Filter *filter)
{
  Query *self = q_create();

  FilteredQuery *fq = ALLOC(FilteredQuery);
  fq->query = query;
  fq->filter = filter;
  self->data = fq;

  self->type = FILTERED_QUERY;
  self->to_s = &fq_to_s;
  self->destroy_i = &fq_destroy;
  self->create_weight_i = &fq_create_weight;

  return self;
}

/***************************************************************************
 *
 * FilteredQueryScorer
 *
 ***************************************************************************/

typedef struct FilteredQueryScorer {
  Scorer *sub_scorer;
  BitVector *bv;
} FilteredQueryScorer;

float fqsc_score(Scorer *self)
{
  Scorer *sub_sc = ((FilteredQueryScorer *)self->data)->sub_scorer;
  return sub_sc->score(sub_sc);
}

bool fqsc_next(Scorer *self)
{
  Scorer *sub_sc = ((FilteredQueryScorer *)self->data)->sub_scorer;
  BitVector *bv = ((FilteredQueryScorer *)self->data)->bv;
  while (sub_sc->next(sub_sc)) {
    self->doc = sub_sc->doc;
    if (bv_get(bv, self->doc)) return true;
  }
  return false;
}

bool fqsc_skip_to(Scorer *self, int doc_num)
{
  Scorer *sub_sc = ((FilteredQueryScorer *)self->data)->sub_scorer;
  BitVector *bv = ((FilteredQueryScorer *)self->data)->bv;
  if (sub_sc->skip_to(sub_sc, doc_num)) {
    self->doc = sub_sc->doc;
    do {
      if (bv_get(bv, self->doc)) return true;
    } while (sub_sc->next(sub_sc));
  }
  return false;
}

Explanation *fqsc_explain(Scorer *self, int doc_num)
{
  Scorer *sub_sc = ((FilteredQueryScorer *)self->data)->sub_scorer;
  return sub_sc->explain(sub_sc, doc_num);
}

void fqsc_destroy(Scorer *self)
{
  FilteredQueryScorer *fqsc = (FilteredQueryScorer *)self->data;
  bv_destroy(fqsc->bv);
  fqsc->sub_scorer->destroy(fqsc->sub_scorer);
  scorer_destroy_i(self);
}

Scorer *fqsc_create(Scorer *scorer, BitVector *bv, Similarity *sim)
{
  Scorer *self = scorer_create(sim);
  FilteredQueryScorer *fqsc = ALLOC(FilteredQueryScorer);
  fqsc->sub_scorer = scorer;
  fqsc->bv = bv;
  self->data = fqsc;

  self->score = &fqsc_score;
  self->next = &fqsc_next;
  self->skip_to = &fqsc_skip_to;
  self->explain = &fqsc_explain;
  self->destroy = &fqsc_destroy;
  return self;
}
