#include "search.h"
#include <string.h>

static char * const FIELD_CHANGE_ERROR_MSG = "All phrase terms must be in the same field.";

/***************************************************************************
 *
 * MultiPhraseWeight
 *
 ***************************************************************************/

char *mphw_to_s(Weight *self)
{
  return strfmt("MultiPhraseWeight(%f)", self->value);
}

Scorer *mphw_scorer(Weight *self, IndexReader *ir)
{
  Scorer *phsc;
  MultiPhraseQuery *mphq = (MultiPhraseQuery *)self->query->data;
  int i;
  TermDocEnum **tps;

  if (mphq->t_cnt == 0) {
    return NULL; /* optimize zero-term case */
  }

  tps = ALLOC_N(TermDocEnum *, mphq->t_cnt);

  for (i = 0; i < mphq->t_cnt; i++) {
    if (mphq->pt_cnt[i] == 1) {
      tps[i] = ir_term_positions_for(ir, mphq->terms[i][0]);
    } else {
      tps[i] = mtdpe_create(ir, mphq->terms[i], mphq->pt_cnt[i]);
    }
    if (tps[i] == NULL) {
      /* free everything we just created and return NULL */
      int j;
      for (j = 0; j < i; j++) {
        tps[i]->close(tps[i]);
      }
      free(tps);
      return NULL;
    }
  }

  if (mphq->slop == 0) {       /* optimize exact case */
    phsc = exact_phrase_scorer_create(self, tps, mphq->positions, mphq->t_cnt,
                                      self->similarity,
                                      ir->get_norms(ir, mphq->field));
  } else {
    phsc = sloppy_phrase_scorer_create(self, tps, mphq->positions, mphq->t_cnt,
                                       self->similarity,
                                       mphq->slop,
                                       ir->get_norms(ir, mphq->field));
  }
  free(tps);
  return phsc;
}

Explanation *mphw_explain(Weight *self, IndexReader *ir, int doc_num)
{
  Explanation *idf_expl1;
  Explanation *idf_expl2;
  Explanation *query_expl;
  Explanation *qnorm_expl;
  Explanation *field_expl;
  Explanation *tf_expl;
  Scorer *scorer;
  uchar *field_norms;
  float field_norm;
  Explanation *field_norm_expl;

  char *query_str = self->query->to_s(self->query, "");
  MultiPhraseQuery *mphq = (MultiPhraseQuery *)self->query->data;
  int i, j;
  char *doc_freqs = NULL;
  int len = 0, pos = 0;

  Explanation *expl = expl_create(0.0,
      strfmt("weight(%s in %d), product of:", query_str, doc_num));

  for (i = 0; i < mphq->t_cnt; i++) {
    for (j = 0; j < mphq->pt_cnt[i]; j++) {
      len += (int)strlen(mphq->terms[i][j]->text) + 30;
    }
  }
  doc_freqs = ALLOC_N(char, len);
  for (i = 0; i < mphq->t_cnt; i++) {
    for (j = 0; j < mphq->pt_cnt[i]; j++) {
      Term *term = mphq->terms[i][j];
      sprintf(doc_freqs + pos, "%s=%d, ", term->text, ir->doc_freq(ir, term));
      pos += (int)strlen(doc_freqs + pos);
    }
  }
  pos -= 2; // remove ", " from the end
  doc_freqs[pos] = 0;

  idf_expl1 = expl_create(self->idf,
      strfmt("idf(%s:<%s>)", mphq->field, doc_freqs));
  idf_expl2 = expl_create(self->idf,
      strfmt("idf(%s:<%s>)", mphq->field, doc_freqs));
  free(doc_freqs);
  
  /* explain query weight */
  query_expl = expl_create(0.0, 
      strfmt("query_weight(%s), product of:", query_str));

  if (self->query->boost != 1.0) {
    expl_add_detail(query_expl, expl_create(self->query->boost, estrdup("boost")));
  }
  expl_add_detail(query_expl, idf_expl1);
  
  qnorm_expl = expl_create(self->qnorm, estrdup("query_norm"));
  expl_add_detail(query_expl, qnorm_expl);
  
  query_expl->value = self->query->boost * self->idf * self->qnorm;

  expl_add_detail(expl, query_expl);
 
  /* explain field weight */
  field_expl = expl_create(0.0, 
      strfmt("field_weight(%s in %d), product of:", query_str, doc_num));
  free(query_str);

  scorer = self->scorer(self, ir);
  tf_expl = scorer->explain(scorer, doc_num);
  scorer->destroy(scorer);
  expl_add_detail(field_expl, tf_expl);
  expl_add_detail(field_expl, idf_expl2);

  field_norms = ir->get_norms(ir, mphq->field);
  field_norm = (field_norms != NULL)
    ? sim_decode_norm(self->similarity, field_norms[doc_num])
    : (float)0.0;
  field_norm_expl = expl_create(field_norm,
    strfmt("field_norm(field=%s, doc=%d)", mphq->field, doc_num));

  expl_add_detail(field_expl, field_norm_expl);

  field_expl->value = tf_expl->value * self->idf * field_norm;
  
  /* combine them */
  if (query_expl->value == 1.0) {
    expl_destoy(expl);
    return field_expl;
  } else {
    expl->value = (query_expl->value * field_expl->value);
    expl_add_detail(expl, field_expl);
    return expl;
  }
}

Weight *mphw_create(Query *query, Searcher *searcher)
{
  Weight *self = w_create(query);

  MultiPhraseQuery *mphq = (MultiPhraseQuery *)query->data;
  int i, j;

  self->scorer    = &mphw_scorer;
  self->explain   = &mphw_explain;
  self->to_s      = &mphw_to_s;
  self->sum_of_squared_weights = &w_sum_of_squared_weights;

  self->similarity = query->get_similarity(query, searcher);
  self->query = query;
  self->value = query->boost;
  self->idf = 0.0;

  for (i = 0; i < mphq->t_cnt; i++) {
    for (j = 0; j < mphq->pt_cnt[i]; j++) {
      self->idf += sim_idf_term(self->similarity, mphq->terms[i][j], searcher);
    }
  }

  return self;
}

/***************************************************************************
 *
 * MultiPhraseQuery
 *
 ***************************************************************************/

#define GET_MPHQ MultiPhraseQuery *mphq = (MultiPhraseQuery *)self->data

/**
 * NOTE: terms must be allocated and it will be freed when the query is
 * destroyed.
 */
void mphq_add_terms(Query *self, Term **terms, int t_cnt, int pos_inc)
{
  GET_MPHQ;
  int position, index = mphq->t_cnt;
  if (index >= mphq->t_capa) {
    mphq->t_capa <<= 1;
    REALLOC_N(mphq->terms, Term **, mphq->t_capa);
    REALLOC_N(mphq->positions, int, mphq->t_capa);
    REALLOC_N(mphq->pt_cnt, int, mphq->t_capa);
  }
  if (index == 0) {
    position = 0;
    mphq->field = terms[0]->field;
  } else {
    int i;
    position = mphq->positions[index - 1] + pos_inc;
    for (i = 0; i < t_cnt; i++) {
      if (strcmp(terms[i]->field, mphq->field) != 0) {
        RAISE(ARG_ERROR, FIELD_CHANGE_ERROR_MSG);
      }
    }
  }

  mphq->terms[index] = terms;
  mphq->pt_cnt[index] = t_cnt;
  mphq->positions[index] = position;
  mphq->t_cnt++;
}

void mphq_destroy(Query *self)
{
  GET_MPHQ;
  int i, j;
  if (self->destroy_all) {
    for (i = 0; i < mphq->t_cnt; i++) {
      for (j = 0; j < mphq->pt_cnt[i]; j++) {
        term_destroy(mphq->terms[i][j]);
      }
      free(mphq->terms[i]);
    }
  }
  free(mphq->terms);
  free(mphq->positions);
  free(mphq->pt_cnt);
  free(mphq);

  q_destroy_i(self);
}

void mphq_extract_terms(Query *self, HashSet *terms)
{
  GET_MPHQ;
  int i, j;
  for (i = 0; i < mphq->t_cnt; i++) {
    for (j = 0; j < mphq->pt_cnt[i]; j++) {
      hs_add(terms, term_clone(mphq->terms[i][j]));
    }
  }
}

char *mphq_to_s(Query *self, char *field)
{
  GET_MPHQ;
  int i, j, buf_index = 0, len = 0, pos, last_pos = -1;
  char *buffer;
  if (!mphq->t_cnt) return NULL;
  len = (int)strlen(mphq->field) + 1;
  for (i = 0; i < mphq->t_cnt; i++) {
    for (j = 0; j < mphq->pt_cnt[i]; j++) {
      len += (int)strlen(mphq->terms[i][j]->text) + 1;
    }
  }

  // add space for extra characters and boost and slop
  len += 100 + 3 * mphq->positions[mphq->t_cnt - 1];

  buffer = ALLOC_N(char, len);

  if (strcmp(field, mphq->field) != 0) {
    len = (int)strlen(mphq->field);
    memcpy(buffer, mphq->field, len);
    buffer[len] = ':';
    buf_index += len + 1;
  }
  buffer[buf_index++] = '"';

  for (i = 0; i < mphq->t_cnt; i++) {
    pos = mphq->positions[i];
    for (j = last_pos; j < pos - 1; j++) {
      memcpy(buffer + buf_index, "<> ", 3);
      buf_index += 3;
    }
    last_pos = pos;

    for (j = 0; j < mphq->pt_cnt[i]; j++) {
      Term *term = mphq->terms[i][j];
      len = (int)strlen(term->text);
      memcpy(buffer + buf_index, term->text, len);
      buf_index += len;
      buffer[buf_index++] = '|';
    }
    buffer[buf_index-1] = ' '; /* change last '|' to ' ' */
  }

  if (buffer[buf_index-1] == ' ') buf_index--;
  buffer[buf_index++] = '"';
  buffer[buf_index] = 0;
  if (mphq->slop != 0) {
    sprintf(buffer + buf_index, "~%d", mphq->slop);
    buf_index += (int)strlen(buffer + buf_index);
  }
  if (self->boost != 1.0) {
    buffer[buf_index] = '^';
    dbl_to_s(buffer + buf_index + 1, self->boost);
  }
  return buffer;
}

Query *mphq_rewrite(Query *self, IndexReader *ir) 
{
  GET_MPHQ;
  if (mphq->t_cnt == 1) { // optimize one-term case
    Term **terms = mphq->terms[0];
    Query *bq = bq_create(true);
    int i;
    for (i = 0; i < mphq->pt_cnt[0]; i++) {
      bq_add_query(bq, tq_create(term_clone(terms[i])), BC_SHOULD);
    }
    bq->boost = self->boost;
    return bq;
  } else {
    self->ref_cnt++;
    return self;
  }
}
    
static uint mphq_hash(Query *self)
{
  int i, j;
  uint hash = 0;
  MultiPhraseQuery *mphq = (MultiPhraseQuery *)self->data;
  for (i = 0; i < mphq->t_cnt; i++) {
    for (j = 0; j < mphq->pt_cnt[i]; j++) {
      hash ^= (term_hash(mphq->terms[i][j]) ^ mphq->positions[i]);
    }
    hash <<= 1;
  }
  return (hash ^ mphq->slop);
}

static int mphq_eq(Query *self, Query *o)
{
  int i, j;
  MultiPhraseQuery *mphq1 = (MultiPhraseQuery *)self->data;
  MultiPhraseQuery *mphq2 = (MultiPhraseQuery *)o->data;
  if (mphq1->slop != mphq2->slop) return false;
  for (i = 0; i < mphq1->t_cnt; i++) {
    if ((mphq1->pt_cnt[i] != mphq2->pt_cnt[i]) ||
        (mphq1->positions[i] != mphq2->positions[i])) return false;

    for (j = 0; j < mphq1->pt_cnt[i]; j++) {
      if (!term_eq(mphq1->terms[i][j], mphq2->terms[i][j])) return false;
    }
  }
  return true;
}

Query *mphq_create()
{
  Query *self = q_create();

  MultiPhraseQuery *mphq = ALLOC(MultiPhraseQuery);
  ZEROSET(mphq, MultiPhraseQuery, 1);
  mphq->t_capa = PHQ_INIT_CAPA;
  mphq->terms = ALLOC_N(Term **, PHQ_INIT_CAPA);
  mphq->positions = ALLOC_N(int, PHQ_INIT_CAPA);
  mphq->pt_cnt = ALLOC_N(int, PHQ_INIT_CAPA);
  self->data = mphq;

  self->type = MULTI_PHRASE_QUERY;
  self->rewrite = &mphq_rewrite;
  self->extract_terms = &mphq_extract_terms;
  self->to_s = &mphq_to_s;
  self->hash = &mphq_hash;
  self->eq = &mphq_eq;
  self->destroy_i = &mphq_destroy;
  self->create_weight_i = &mphw_create;
  return self;
}
