#include <string.h>
#include "search.h"

static char * const NIL_BOUNDS_ERROR_MSG = "At least one value must be non-nil";
static char * const LOWER_BOUND_ERROR_MSG = "The lower bound must be non-nil to be inclusive";
static char * const UPPER_BOUND_ERROR_MSG = "The upper bound must be non-nil to be inclusive";
static char * const BOUND_ORDER_ERROR_MSG = "The lower bound must less than the upper bound";
        
/*****************************************************************************
 *
 * Range
 *
 *****************************************************************************/

char *range_to_s(Range *range, char *field, float boost)
{
  char *buffer, *b;
  size_t flen, llen, ulen;

  flen = strlen(range->field);
  llen = range->lower_term ? strlen(range->lower_term) : 0;
  ulen = range->upper_term ? strlen(range->upper_term) : 0;
  buffer = ALLOC_N(char, flen + llen + ulen + 40);
  b = buffer;

  if (strcmp(field, range->field)) {
    memcpy(buffer, range->field, flen * sizeof(char));
    b += flen;
    *b = ':';
    b++;
  }

  if (range->lower_term) {
    *b = range->include_lower ? '[' : '{';
    b++;
    memcpy(b, range->lower_term, llen);
    b += llen;
  } else {
    *b = '<';
    b++;
  }

  if (range->upper_term && range->lower_term) {
    *b = ' '; b++;
  }

  if (range->upper_term) {
    memcpy(b, range->upper_term, ulen);
    b += ulen;
    *b = range->include_upper ? ']' : '}';
    b++;
  } else {
    *b = '>';
    b++;
  }

  *b = 0;
  if (boost != 1.0) {
    *b = '^';
    dbl_to_s(b + 1, boost);
  }
  return buffer;
}

void range_destroy(void *p)
{
  Range *range = (Range *)p;
  free(range->field);
  if (range->lower_term) free(range->lower_term);
  if (range->upper_term) free(range->upper_term);
  free(range);
}

static inline uint range_hash(Range *self)
{
  return self->include_lower | (self->include_upper << 1) |
      ((str_hash(self->field) ^
        (self->lower_term ? str_hash(self->lower_term) : 0) ^
        (self->upper_term ? str_hash(self->upper_term) : 0)) << 2);
}

static inline int str_eq(char *s1, char *s2)
{
  return (s1 && s2 && (strcmp(s1, s2) == 0)) || (s1 == s2);
}

static inline int range_eq(Range *self, Range *o)
{
  return (str_eq(self->field, o->field) &&
          str_eq(self->lower_term, o->lower_term) &&
          str_eq(self->upper_term, o->upper_term) &&
          (self->include_lower == o->include_lower) &&
          (self->include_upper == o->include_upper));
}

Range *range_create(const char *field, char *lower_term, char *upper_term,
    bool include_lower, bool include_upper)
{
  Range *range; 

  if (!lower_term && !upper_term) 
    RAISE(ARG_ERROR, NIL_BOUNDS_ERROR_MSG);
  if (include_lower && !lower_term) 
    RAISE(ARG_ERROR, LOWER_BOUND_ERROR_MSG);
  if (include_upper && !upper_term) 
    RAISE(ARG_ERROR, UPPER_BOUND_ERROR_MSG);
  if (upper_term && lower_term && (strcmp(upper_term, lower_term) < 0))
    RAISE(ARG_ERROR, BOUND_ORDER_ERROR_MSG);

  range = ALLOC(Range);

  range->field = estrdup((char *)field);
  range->lower_term = lower_term ? estrdup(lower_term) : NULL;
  range->upper_term = upper_term ? estrdup(upper_term) : NULL;
  range->include_lower = include_lower;
  range->include_upper = include_upper;
  return range;
}

/***************************************************************************
 *
 * RangeFilter
 *
 ***************************************************************************/

void rfilt_destroy(Filter *self)
{
  range_destroy(self->data);
  filt_destroy(self);
}

char *rfilt_to_s(Filter *self)
{
  Range *range = (Range *)self->data;
  char *rstr = range_to_s(range, "", 1.0);
  char *rfstr = epstrdup("RangeFilter< %s >", strlen(rstr), rstr);
  free(rstr);
  return rfstr;
}

BitVector *rfilt_get_bv(Filter *self, IndexReader *ir)
{
  BitVector *bv = bv_create_size(ir->max_doc(ir));
  Range *range = (Range *)self->data;
  char *field = range->field;
  char *lower_term = range->lower_term ? range->lower_term : (char *)EMPTY_STRING;
  char *upper_term = range->upper_term;
  bool include_upper = range->include_upper;

  Term *term_from = term_create(range->field, lower_term);
  Term term;
  TermBuffer *tb;
  TermEnum* te;
  TermDocEnum *tde;
  bool check_lower;

  te = ir->terms_from(ir, term_from);
  if (te->tb_curr == NULL) {
    return bv;
  }
            
  check_lower = false;
  if (!range->include_lower) // make adjustments to set to exclusive
      check_lower = true;
        
  tde = ir->term_docs(ir);
  tb = te->tb_curr;
  term.text = tb->text;
  do {
    if (tb && strcmp(tb->field, field) == 0) {
      if (!check_lower || lower_term == EMPTY_STRING ||
          strcmp(tb->text, lower_term) > 0) {
        check_lower = false;
        if (upper_term) {
          int compare = strcmp(upper_term, tb->text);
          /* if beyond the upper term, or is exclusive and
           * this is equal to the upper term, break out */
          if ((compare < 0) ||
              (!include_upper && compare==0)) {
              break;
          }
        }
        /* we have a good term, find the docs */
        /* text is already pointing to term buffer text */
        term.field = tb->field;
        tde->seek(tde, &term);
        while (tde->next(tde)) {
          bv_set(bv, tde->doc_num(tde));
          //printf("Setting %d\n", tde->doc_num(tde));
        }
      }
    } else {
      break;
    }
  } while (te->next(te));
          
  tde->close(tde);
  te->close(te);
  term_destroy(term_from);

  return bv;
}

uint rfilt_hash(Filter *self)
{
  return range_hash((Range *)self->data);
}

int rfilt_eq(Filter *self, Filter *o)
{
  return range_eq((Range *)self->data, (Range *)o->data);
}

Filter *rfilt_create(const char *field, char *lower_term, char *upper_term,
    bool include_lower, bool include_upper)
{
  Filter *self;
  Range *range = range_create(field, lower_term, upper_term,
      include_lower, include_upper); 

  self = filt_create("RangeFilter");
  self->data = range;
  self->get_bv = &rfilt_get_bv;
  self->hash = &rfilt_hash;
  self->eq = &rfilt_eq;
  self->to_s = &rfilt_to_s;
  self->destroy = &rfilt_destroy;
  return self;
}

/*****************************************************************************
 *
 * RangeQuery
 *
 *****************************************************************************/

char *rq_to_s(Query *self, char *field)
{
  Range *range = (Range *)self->data;
  return range_to_s(range, field, self->boost);
}

void rq_destroy(Query *self)
{
  range_destroy(self->data);
  q_destroy_i(self);
}

Query *rq_rewrite(Query *self, IndexReader *ir)
{
  Range *r = (Range *)self->data;
  Filter *filter = rfilt_create(r->field, r->lower_term, r->upper_term,
      r->include_lower, r->include_upper);
  return csq_create(filter);
}

static uint rq_hash(Query *self)
{
  return range_hash((Range *)self->data);
}

static int rq_eq(Query *self, Query *o)
{
  return range_eq((Range *)self->data, (Range *)o->data);
}

Query *rq_create_less(const char *field, char *upper_term, bool include_upper)
{
  return rq_create(field, NULL, upper_term, false, include_upper);
}

Query *rq_create_more(const char *field, char *lower_term, bool include_lower)
{
  return rq_create(field, lower_term, NULL, include_lower, false);
}

Query *rq_create(const char *field, char *lower_term, char *upper_term,
    bool include_lower, bool include_upper)
{
  Query *self = q_create();
  Range *range = range_create(field, lower_term, upper_term,
      include_lower, include_upper); 

  self->data = range;

  self->type = RANGE_QUERY;
  self->rewrite = &rq_rewrite;
  self->to_s = &rq_to_s;
  self->hash = &rq_hash;
  self->eq = &rq_eq;
  self->destroy_i = &rq_destroy;
  self->create_weight_i = &q_create_weight_unsup;
  return self;
}
