#include <string.h>
#include "search.h"

/*****************************************************************************
 *
 * NearSpanEnum
 *
 *****************************************************************************/

/*****************************************************************************
 *
 * SpanWeight
 *
 *****************************************************************************/

Explanation *spanw_explain(Weight *self, IndexReader *ir, int target)
{
  Explanation *expl;
  Explanation *idf_expl1;
  Explanation *idf_expl2;
  Explanation *query_expl;
  Explanation *qnorm_expl;
  Explanation *field_expl;
  Explanation *tf_expl;
  Scorer *scorer;
  uchar *field_norms;
  float field_norm;
  Explanation *field_norm_expl;

  char *query_str = self->query->to_s(self->query, "");
  HashSet *terms = (HashSet *)self->data;
  char *field = ((SpanQuery *)self->query->data)->field;
  char *doc_freqs = NULL;
  size_t df_i = 0;
  int i;
  Term *t;


  for (i = 0; i < terms->size; i++) {
    t = (Term *)terms->elems[i];
    REALLOC_N(doc_freqs, char, df_i + strlen(t->text) + 23);
    sprintf(doc_freqs + df_i, "%s=%d, ", t->text, ir->doc_freq(ir, t));
    df_i = strlen(doc_freqs);
  }
  /* remove the ',' at the end of the string if it exists */
  if (terms->size > 0) {
    df_i -= 2;
    doc_freqs[df_i] = '\0';
  } else {
    doc_freqs = "";
  }

  expl = expl_create(0.0,
      strfmt("weight(%s in %d), product of:", query_str, target));

  /* We need two of these as it's included in both the query explanation
   * and the field explanation */
  idf_expl1 = expl_create(self->idf,
      strfmt("idf(%s: %s)", field, doc_freqs));
  idf_expl2 = expl_create(self->idf,
      strfmt("idf(%s: %s)", field, doc_freqs));
  if (terms->size > 0) {
    free(doc_freqs); /* only free if allocated */
  }

  /* explain query weight */
  query_expl = expl_create(0.0,
      strfmt("query_weight(%s), product of:", query_str));

  if (self->query->boost != 1.0) {
    expl_add_detail(query_expl, expl_create(self->query->boost, estrdup("boost")));
  }

  expl_add_detail(query_expl, idf_expl1);

  qnorm_expl = expl_create(self->qnorm, estrdup("query_norm"));
  expl_add_detail(query_expl, qnorm_expl);

  query_expl->value = self->query->boost * idf_expl1->value * qnorm_expl->value;

  expl_add_detail(expl, query_expl);

  /* explain field weight */
  field_expl = expl_create(0.0, 
      strfmt("field_weight(%s:%s in %d), product of:", field, query_str, target));
  free(query_str);

  scorer = self->scorer(self, ir);
  tf_expl = scorer->explain(scorer, target);
  scorer->destroy(scorer);
  expl_add_detail(field_expl, tf_expl);
  expl_add_detail(field_expl, idf_expl2);

  field_norms = ir->get_norms(ir, field);
  field_norm = (field_norms 
    ? sim_decode_norm(self->similarity, field_norms[target]) 
    : (float)0.0);
  field_norm_expl = expl_create(field_norm,
      strfmt("field_norm(field=%s, doc=%d)", field, target));
  expl_add_detail(field_expl, field_norm_expl);

  field_expl->value = tf_expl->value * idf_expl2->value * field_norm_expl->value;

  /* combine them */
  if (query_expl->value == 1.0) {
    expl_destoy(expl);
    return field_expl;
  } else {
    expl->value = (query_expl->value * field_expl->value);
    expl_add_detail(expl, field_expl);
    return expl;
  }
}

char *spanw_to_s(Weight *self)
{
  return strfmt("SpanWeight(%f)", self->value);
}

void spanw_destroy(Weight *self)
{
  hs_destroy_all(self->data);
  w_destroy(self);
}

Weight *spanw_create(Query *query, Searcher *searcher)
{
  Weight *self = w_create(query);
  SpanQuery *spanq = (SpanQuery *)query->data;
  HashSet *terms = spanq->get_terms(query);

  self->data = terms;
  self->scorer    = &spansc_create;
  self->explain   = &spanw_explain;
  self->to_s      = &spanw_to_s;
  self->destroy   = &spanw_destroy;
  self->sum_of_squared_weights = &w_sum_of_squared_weights;

  self->similarity = query->get_similarity(query, searcher);

  self->idf = sim_idf_phrase(self->similarity, (Term **)terms->elems,
      terms->size, searcher);

  return self;
}


/*****************************************************************************
 *
 * SpanTermEnum
 *
 *****************************************************************************/

bool spante_next(SpanEnum *self)
{
  SpanTermEnum *ste = (SpanTermEnum *)self->data;
  TermDocEnum *tde = ste->positions;

  if (ste->count == ste->freq) {
    if (! tde->next(tde)) {
      ste->doc = INT_MAX;
      return false;
    }
    ste->doc = tde->doc_num(tde);
    ste->freq = tde->freq(tde);
    ste->count = 0;
  }
  ste->position = tde->next_position(tde);
  ste->count++;
  return true;
}

bool spante_skip_to(SpanEnum *self, int target)
{
  SpanTermEnum *ste = (SpanTermEnum *)self->data;
  TermDocEnum *tde = ste->positions;

  /* are we already at the correct position? */
  if (ste->doc >= target) return true;

  if (! tde->skip_to(tde, target)) {
    ste->doc = INT_MAX;
    return false;
  }

  ste->doc = tde->doc_num(tde);
  ste->freq = tde->freq(tde);
  ste->count = 0;

  ste->position = tde->next_position(tde);
  ste->count++;
  return true;
}

int  spante_doc(SpanEnum *self)
{
  SpanTermEnum *ste = (SpanTermEnum *)self->data;
  return ste->doc;
}

int  spante_start(SpanEnum *self)
{
  SpanTermEnum *ste = (SpanTermEnum *)self->data;
  return ste->position;
}

int  spante_end(SpanEnum *self)
{
  SpanTermEnum *ste = (SpanTermEnum *)self->data;
  return ste->position + 1;
}

char *spante_to_s(SpanEnum *self) 
{
  char *field = ((SpanQuery *)self->query->data)->field;
  char *query_str = self->query->to_s(self->query, field);
  char pos_str[20];
  size_t len = strlen(query_str);
  int pos;
  char *str = ALLOC_N(char, len + 40);

  if (self->doc(self) < 0) {
      sprintf(pos_str, "START");
  } else {
    if (self->doc(self) == INT_MAX) {
      sprintf(pos_str, "END");
    } else {
      pos = ((SpanTermEnum *)self->data)->position;
      sprintf(pos_str, "%d", self->doc(self) - pos);
    }
  }
  sprintf("SpanTermEnum(%s)@%s", query_str, pos_str);
  free(query_str);
  return str;
}

void spante_destroy(SpanEnum *self)
{
  SpanTermEnum *ste = (SpanTermEnum *)self->data;
  TermDocEnum *tde = ste->positions;
  tde->close(tde);
  free(ste);
  free(self);
}

SpanEnum *spante_create(Query *query, IndexReader *ir)
{
  Term *term = (Term *)((SpanQuery *)query->data)->data;
  SpanEnum *self = ALLOC(SpanEnum);

  SpanTermEnum *ste = ALLOC(SpanTermEnum);
  ste->positions = ir_term_positions_for(ir, term);
  ste->position = -1;
  ste->doc = -1;
  ste->count = 0;
  ste->freq = 0;

  self->data = ste;

  self->query = query;
  self->next = &spante_next;
  self->skip_to = &spante_skip_to;
  self->doc = &spante_doc;
  self->start = &spante_start;
  self->end = &spante_end;
  self->destroy = &spante_destroy;
  self->to_s = &spante_to_s;

  return self;
}


/*****************************************************************************
 *
 * SpanFirstEnum
 *
 *****************************************************************************/

bool spanfe_next(SpanEnum *self)
{
  SpanEnum *se = (SpanEnum *)(self->data);
  int end = ((SpanFirstQuery *)((SpanQuery *)self->query->data)->data)->end;
  while (se->next(se)) { /* scan to next match */
    if (se->end(se) <= end) return true;
  }
  return false;
}

bool spanfe_skip_to(SpanEnum *self, int target)
{
  SpanEnum *se = (SpanEnum *)(self->data);
  int end = ((SpanFirstQuery *)((SpanQuery *)self->query->data)->data)->end;

  if (! se->skip_to(se, target)) return false;

  if (se->end(se) <= end) /* there is a match */
    return true;

  return se->next(se);    /* scan to next match */
}

int  spanfe_doc(SpanEnum *self)
{
  SpanEnum *se = (SpanEnum *)(self->data);
  return se->doc(se);
}

int  spanfe_start(SpanEnum *self)
{
  SpanEnum *se = (SpanEnum *)(self->data);
  return se->start(se);
}

int  spanfe_end(SpanEnum *self)
{
  SpanEnum *se = (SpanEnum *)(self->data);
  return se->end(se);
}

char *spanfe_to_s(SpanEnum *self) 
{
  char *field = ((SpanQuery *)self->query->data)->field;
  char *query_str = self->query->to_s(self->query, field);
  char *res = strfmt("SpanFirstEnum(%s)", query_str);
  free(query_str);
  return res;
}

void spanfe_destroy(SpanEnum *self)
{
  SpanEnum *se = (SpanEnum *)self->data;
  se->destroy(se);
  free(self);
}

SpanEnum *spanfe_create(Query *query, IndexReader *ir)
{
  SpanEnum *self = ALLOC(SpanEnum);
  SpanFirstQuery *sfq = (SpanFirstQuery *)((SpanQuery *)query->data)->data;

  self->data = ((SpanQuery *)sfq->match->data)->get_spans(sfq->match, ir);

  self->query = query;
  self->next = &spanfe_next;
  self->skip_to = &spanfe_skip_to;
  self->doc = &spanfe_doc;
  self->start = &spanfe_start;
  self->end = &spanfe_end;
  self->destroy = &spanfe_destroy;
  self->to_s = &spanfe_to_s;

  return self;
}


/*****************************************************************************
 *
 * SpanOrEnum
 *
 *****************************************************************************/

bool span_less_than(void *p1, void *p2)
{
  SpanEnum *s1 = (SpanEnum *)p1;
  SpanEnum *s2 = (SpanEnum *)p2;
  int doc_diff, start_diff;
  doc_diff = s1->doc(s1) - s2->doc(s2);
  if (doc_diff == 0) {
    start_diff = s1->start(s1) - s2->start(s2);
    if (start_diff == 0) {
      return s1->end(s1) < s2->end(s2);
    } else {
      return start_diff < 0;
    }
  } else {
    return doc_diff < 0;
  }
}

bool spanoe_next(SpanEnum *self)
{
  SpanOrEnum *soe = (SpanOrEnum *)self->data;
  SpanEnum *se;
  int i;

  if (soe->first_time) { /* first time -- initialize */
    for (i = 0; i < soe->s_cnt; i++) {
      se = soe->span_enums[i];
      if (se->next(se)) /* move to first entry */
        pq_push(soe->queue, se);
    }
    soe->first_time = false;
    return soe->queue->count != 0;
  }

  if (soe->queue->count == 0) return false; /* all done */

  se = (SpanEnum *)pq_top(soe->queue);
  if (se->next(se)) { /* move to next */
    pq_down(soe->queue);
    return true;
  }

  pq_pop(soe->queue); /* exhausted a clause */

  return soe->queue->count != 0;
}

bool spanoe_skip_to(SpanEnum *self, int target)
{
  SpanOrEnum *soe = (SpanOrEnum *)self->data;
  SpanEnum *se;
  int i;

  if (soe->first_time) { /* first time -- initialize */
    for (i = 0; i < soe->s_cnt; i++) {
      se = soe->span_enums[i];
      if (se->skip_to(se, target)) /* move to target */
        pq_push(soe->queue, se);
    }
    soe->first_time = false;
  } else {
    while ((soe->queue->count != 0) &&
        ((se=(SpanEnum *)pq_top(soe->queue))->doc(se) < target)) {
      if (se->skip_to(se, target)) {
        pq_down(soe->queue);
      } else {
        pq_pop(soe->queue);
      }
    }
  }

  return soe->queue->count != 0;
}

#define GET_TOP_SOE SpanOrEnum *soe = (SpanOrEnum *)self->data;\
                    SpanEnum *se = (SpanEnum *)pq_top(soe->queue)
int  spanoe_doc(SpanEnum *self)
{
  GET_TOP_SOE;
  return se->doc(se);
}

int  spanoe_start(SpanEnum *self)
{
  GET_TOP_SOE;
  return se->start(se);
}

int  spanoe_end(SpanEnum *self)
{
  GET_TOP_SOE;
  return se->end(se);
}

char *spanoe_to_s(SpanEnum *self) 
{
  SpanOrEnum *soe = (SpanOrEnum *)self->data;
  char *field = ((SpanQuery *)self->query->data)->field;
  char *query_str = self->query->to_s(self->query, field);
  char doc_str[62];
  size_t len = strlen(query_str);
  char *str = ALLOC_N(char, len + 80);

  if (soe->first_time) {
      sprintf(doc_str, "START");
  } else {
    if (soe->queue->count == 0) {
      sprintf(doc_str, "END");
    } else {
      sprintf(doc_str, "%d:%d-%d", self->doc(self),
          self->start(self), self->end(self));
    }
  }
  sprintf("SpanOrEnum(%s)@%s", query_str, doc_str);
  free(query_str);
  return str;
}

void spanoe_destroy(SpanEnum *self)
{
  SpanEnum *se;
  SpanOrEnum *soe = (SpanOrEnum *)self->data;
  int i;
  pq_destroy(soe->queue);
  for (i = 0; i < soe->s_cnt; i++) {
    se = soe->span_enums[i];
    se->destroy(se);
  }
  free(soe->span_enums);
  free(soe);
  free(self);
}

SpanEnum *spanoe_create(Query *query, IndexReader *ir)
{
  Query *clause;
  SpanEnum *self = ALLOC(SpanEnum);
  SpanOrQuery *soq = (SpanOrQuery *)((SpanQuery *)query->data)->data;
  SpanOrEnum *soe = ALLOC(SpanOrEnum);
  int i;
  soe->first_time = true;
  soe->s_cnt = soq->c_cnt;
  soe->span_enums = ALLOC_N(SpanEnum *, soe->s_cnt);
  for (i = 0; i < soe->s_cnt; i++) {
    clause = soq->clauses[i];
    soe->span_enums[i] = ((SpanQuery *)clause->data)->get_spans(clause, ir);
  }
  
  soe->queue = pq_create(soe->s_cnt, &span_less_than);

  self->data = soe;

  self->query = query;
  self->next = &spanoe_next;
  self->skip_to = &spanoe_skip_to;
  self->doc = &spanoe_doc;
  self->start = &spanoe_start;
  self->end = &spanoe_end;
  self->destroy = &spanoe_destroy;
  self->to_s = &spanoe_to_s;

  return self;
}

/*****************************************************************************
 *
 * SpanNearEnum
 *
 *****************************************************************************/

#define SNE_NEXT() do {\
  sne->current = (sne->current+1) % sne->s_cnt;\
  se = sne->span_enums[sne->current];\
} while (0);

bool sne_init(SpanNearEnum *sne)
{
  SpanEnum *se = sne->span_enums[sne->current];
  int prev_doc = se->doc(se);
  int i;

  for (i = 1; i < sne->s_cnt; i++) {
    SNE_NEXT();
    if (!se->skip_to(se, prev_doc)) return false;
    prev_doc = se->doc(se);
  }
  return true;
}

bool sne_goto_next_doc(SpanNearEnum *sne)
{
  SpanEnum *se = sne->span_enums[sne->current];
  int prev_doc = se->doc(se);

  SNE_NEXT();

  while (se->doc(se) < prev_doc) {
    if (! se->skip_to(se, prev_doc)) return false;
    prev_doc = se->doc(se);
    SNE_NEXT();
  }
  return true;
}

bool sne_next_unordered_match(SpanEnum *self)
{
  SpanNearEnum *sne = (SpanNearEnum *)self->data;
  SpanEnum *se, *min_se = NULL;
  int i;
  int max_end, end, min_start, start, doc;
  int lengths_sum;

  while (true) {

    max_end = 0;
    min_start = INT_MAX;
    lengths_sum = 0;

    for (i = 0; i < sne->s_cnt; i++) {
      se = sne->span_enums[i];
      if ((end=se->end(se)) > max_end) max_end = end;
      if ((start=se->start(se)) < min_start) {
        min_start = start;
        min_se = se;
        sne->current = i; /* current should point to the minimum span */
      }
      lengths_sum += end - start;
    }

    if ((max_end - min_start - lengths_sum) <= sne->slop) {
      /* we have a match */
      sne->start = min_start;
      sne->end = max_end;
      sne->doc = min_se->doc(min_se);
      return true;
    }

    /* increment the minimum span_enum and try again */
    doc = min_se->doc(min_se);
    if (!min_se->next(min_se)) return false;
    if (doc < min_se->doc(min_se)) {
      if (!sne_goto_next_doc(sne)) return false;
    }
  }
}

bool sne_next_ordered_match(SpanEnum *self)
{
  SpanNearEnum *sne = (SpanNearEnum *)self->data;
  SpanEnum *se;
  int i;
  int prev_doc, prev_start, prev_end;
  int doc=0, start=0, end=0;
  int lengths_sum;

  while (true) {
    se = sne->span_enums[0];

    prev_doc = se->doc(se);
    sne->start = prev_start = se->start(se);
    prev_end = se->end(se);

    i = 1;
    lengths_sum = prev_end - prev_start;

    while (i < sne->s_cnt) {
      se = sne->span_enums[i];
      doc = se->doc(se);
      start = se->start(se);
      end = se->end(se);
      while ((doc == prev_doc) && ((start < prev_start) ||
           ((start == prev_start) && (end < prev_end)))) {
        if (!se->next(se)) return false;
        doc = se->doc(se);
        start = se->start(se);
        end = se->end(se);
      }
      if (doc != prev_doc) {
        sne->current = i;
        if (!sne_goto_next_doc(sne)) return false;
        break;
      }
      i++;
      lengths_sum += end - start;
      prev_doc = doc;
      prev_start = start;
      prev_end = end;
    }
    if (i == sne->s_cnt) {
      if ((end - sne->start - lengths_sum) <= sne->slop) {
        /* we have a match */
        sne->end = end;
        sne->doc = doc;

        /* the minimum span is always the first span so it needs to be
         * incremented next time around */
        sne->current = 0;
        return true;

      } else {
        se = sne->span_enums[0];
        if (!se->next(se)) return false;
        if (se->doc(se) != prev_doc) {
          sne->current = 0;
          if (!sne_goto_next_doc(sne)) return false;
        }
      }
    }
  }
}
  
bool sne_next_match(SpanEnum *self)
{
  SpanNearEnum *sne = (SpanNearEnum *)self->data;
  SpanEnum *se_curr, *se_next;

  if (!sne->first_time) {
    if (!sne_init(sne)) return false;
    sne->first_time = false;
  }
  se_curr = sne->span_enums[sne->current];
  se_next = sne->span_enums[(sne->current+1)%sne->s_cnt];
  if (se_curr->doc(se_curr) > se_next->doc(se_next)) {
    if (!sne_goto_next_doc(sne)) return false;
  }

  if (sne->in_order) {
    return sne_next_ordered_match(self);
  } else {
    return sne_next_unordered_match(self);
  }
}

bool spanne_next(SpanEnum *self)
{
  SpanNearEnum *sne = (SpanNearEnum *)self->data;
  SpanEnum *se;

  se = sne->span_enums[sne->current];
  if (!se->next(se)) return false;

  return sne_next_match(self);
}

bool spanne_skip_to(SpanEnum *self, int target)
{
  SpanNearEnum *sne = (SpanNearEnum *)self->data;
  SpanEnum *se;

  se = sne->span_enums[sne->current];
  if (!se->skip_to(se, target)) return false;

  return sne_next_match(self);
}

#define GET_TOP_SNE SpanNearEnum *sne = (SpanNearEnum *)self->data;

int  spanne_doc(SpanEnum *self)
{
  GET_TOP_SNE;
  return sne->doc;
}

int  spanne_start(SpanEnum *self)
{
  GET_TOP_SNE;
  return sne->start;
}

int  spanne_end(SpanEnum *self)
{
  GET_TOP_SNE;
  return sne->end;
}

char *spanne_to_s(SpanEnum *self) 
{
  SpanNearEnum *sne = (SpanNearEnum *)self->data;
  char *field = ((SpanQuery *)self->query->data)->field;
  char *query_str = self->query->to_s(self->query, field);
  char doc_str[62];
  size_t len = strlen(query_str);
  char *str = ALLOC_N(char, len + 80);

  if (sne->first_time) {
    sprintf(doc_str, "START");
  } else {
    sprintf(doc_str, "%d:%d-%d", self->doc(self),
          self->start(self), self->end(self));
  }
  sprintf("SpanNearEnum(%s)@%s", query_str, doc_str);
  free(query_str);
  return str;
}

void spanne_destroy(SpanEnum *self)
{
  SpanEnum *se;
  SpanNearEnum *sne = (SpanNearEnum *)self->data;
  int i;
  for (i = 0; i < sne->s_cnt; i++) {
    se = sne->span_enums[i];
    se->destroy(se);
  }
  free(sne->span_enums);
  free(sne);
  free(self);
}

SpanEnum *spanne_create(Query *query, IndexReader *ir)
{
  Query *clause;
  SpanEnum *self = ALLOC(SpanEnum);
  SpanNearQuery *snq = (SpanNearQuery *)((SpanQuery *)query->data)->data;
  SpanNearEnum *sne = ALLOC(SpanNearEnum);
  int i;
  sne->first_time = true;
  sne->in_order = snq->in_order;
  sne->slop = snq->slop;
  sne->s_cnt = snq->c_cnt;
  sne->span_enums = ALLOC_N(SpanEnum *, sne->s_cnt);
  for (i = 0; i < sne->s_cnt; i++) {
    clause = snq->clauses[i];
    sne->span_enums[i] = ((SpanQuery *)clause->data)->get_spans(clause, ir);
  }
  sne->current = 0;
  
  sne->doc = -1;
  sne->start = -1;
  sne->end = -1;

  self->data = sne;

  self->query = query;
  self->next = &spanne_next;
  self->skip_to = &spanne_skip_to;
  self->doc = &spanne_doc;
  self->start = &spanne_start;
  self->end = &spanne_end;
  self->destroy = &spanne_destroy;
  self->to_s = &spanne_to_s;

  return self;
}

/*****************************************************************************
 *
 * SpanNotEnum
 *
 *****************************************************************************/

bool spanxe_next(SpanEnum *self)
{
  SpanNotEnum *sxe = (SpanNotEnum *)(self->data);
  SpanEnum *inc = sxe->inc, *exc = sxe->exc;
  if (sxe->more_inc) {                       // move to next incl
    sxe->more_inc = inc->next(inc);
  }

  while (sxe->more_inc && sxe->more_exc) {
    if (inc->doc(inc) > exc->doc(exc)) {     // skip excl
      sxe->more_exc = exc->skip_to(exc, inc->doc(inc));
    }

    while (sxe->more_exc &&                  // while excl is before
           (inc->doc(inc) == exc->doc(exc)) &&
           (exc->end(exc) <= inc->start(inc))) {
      sxe->more_exc = exc->next(exc);        // increment excl
    }

    if (! sxe->more_exc ||                   // if no intersection
        (inc->doc(inc) != exc->doc(exc)) ||
        inc->end(inc) <= exc->start(exc)) {
      break;                                 // we found a match
    }

    sxe->more_inc = inc->next(inc);          // intersected: keep scanning
  }
  return sxe->more_inc;
}

bool spanxe_skip_to(SpanEnum *self, int target)
{
  SpanNotEnum *sxe = (SpanNotEnum *)(self->data);
  SpanEnum *inc = sxe->inc, *exc = sxe->exc;
  int doc;

  if (sxe->more_inc) {                       // move to next incl
    if (!(sxe->more_inc=sxe->inc->skip_to(sxe->inc, target))) return false;
  }

  if (sxe->more_inc && ((doc=inc->doc(inc)) > exc->doc(exc))) {
    sxe->more_exc = exc->skip_to(exc, doc);
  }

  while (sxe->more_exc &&                    // while excl is before
         inc->doc(inc) == exc->doc(exc) &&
         exc->end(exc) <= inc->start(inc)) {
    sxe->more_exc = exc->next(exc);          // increment excl
  }

  if (!sxe->more_exc ||                      // if no intersection
        inc->doc(inc) != exc->doc(exc) ||
        inc->end(inc) <= exc->start(exc)) {
    return true;                             // we found a match
  }

  return spanxe_next(self);                  // scan to next match
}

int  spanxe_doc(SpanEnum *self)
{
  SpanEnum *inc = ((SpanNotEnum *)(self->data))->inc;
  return inc->doc(inc);
}

int  spanxe_start(SpanEnum *self)
{
  SpanEnum *inc = ((SpanNotEnum *)(self->data))->inc;
  return inc->start(inc);
}

int  spanxe_end(SpanEnum *self)
{
  SpanEnum *inc = ((SpanNotEnum *)(self->data))->inc;
  return inc->end(inc);
}

char *spanxe_to_s(SpanEnum *self) 
{
  char *field = ((SpanQuery *)self->query->data)->field;
  char *query_str = self->query->to_s(self->query, field);
  char *res = strfmt("SpanNotEnum(%s)", query_str);
  free(query_str);
  return res;
}

void spanxe_destroy(SpanEnum *self)
{
  SpanNotEnum *sxe = (SpanNotEnum *)self->data;
  sxe->inc->destroy(sxe->inc);
  sxe->exc->destroy(sxe->exc);
  free(sxe);
  free(self);
}

SpanEnum *spanxe_create(Query *query, IndexReader *ir)
{
  SpanEnum *self = ALLOC(SpanEnum);
  SpanNotQuery *sxq = (SpanNotQuery *)((SpanQuery *)query->data)->data;

  SpanNotEnum *sxe = self->data = ALLOC(SpanNotEnum);
  sxe->inc = ((SpanQuery *)sxq->inc->data)->get_spans(sxq->inc, ir);
  sxe->exc = ((SpanQuery *)sxq->exc->data)->get_spans(sxq->exc, ir);
  sxe->more_inc = true;
  sxe->more_exc = sxe->exc->next(sxe->exc);

  self->query = query;
  self->next = &spanxe_next;
  self->skip_to = &spanxe_skip_to;
  self->doc = &spanxe_doc;
  self->start = &spanxe_start;
  self->end = &spanxe_end;
  self->destroy = &spanxe_destroy;
  self->to_s = &spanxe_to_s;

  return self;
}
/*****************************************************************************
 *
 * SpanQuery
 *
 *****************************************************************************/

void spanq_destroy(Query *self)
{
  SpanQuery *sq = (SpanQuery *)self->data;
  free(sq);
  q_destroy_i(self);
}

/*****************************************************************************
 *
 * SpanTermQuery
 *
 *****************************************************************************/

char *spantq_to_s(Query *self, char *field)
{
  Term *term = (Term *)((SpanQuery *)self->data)->data;
  char *term_str, *res;
  if (field == term->field) {
    term_str = estrdup(term->text);
  } else {
    term_str = term_to_s(term);
  }
  res = strfmt("span_term(%s)", term_str);
  free(term_str);
  return res;
}

static void spantq_destroy(Query *self)
{
  SpanQuery *sq = (SpanQuery *)self->data;
  if (self->destroy_all) {
    Term *term = (Term *)sq->data;
    term_destroy(term);
  }
  free(sq);
  q_destroy_i(self);
}

static void spantq_extract_terms(Query *self, HashSet *terms)
{
  Term *term = (Term *)((SpanQuery *)self->data)->data;
  hs_add(terms, term_clone(term));
}

static HashSet *spantq_get_terms(Query *self)
{
  Term *term = (Term *)((SpanQuery *)self->data)->data;
  HashSet *terms = term_set_create();
  hs_add(terms, term_clone(term));
  return terms;
}

static uint spantq_hash(Query *self)
{
  return term_hash((Term *)((SpanQuery *)self->data)->data);
}

static int spantq_eq(Query *self, Query *o)
{
  return term_eq((Term *)((SpanQuery *)self->data)->data,
                 (Term *)((SpanQuery *)o->data)->data);
}

Query *spantq_create(Term *term)
{
  Query *self = q_create();

  SpanQuery *sq = ALLOC(SpanQuery);
  sq->data = term;
  sq->get_spans = &spante_create;
  sq->get_terms = &spantq_get_terms;
  sq->field = term->field;
  self->data = sq;

  self->type = SPAN_TERM_QUERY;
  self->extract_terms = &spantq_extract_terms;
  self->to_s = &spantq_to_s;
  self->hash = &spantq_hash;
  self->eq = &spantq_eq;
  self->destroy_i = &spantq_destroy;
  self->create_weight_i = &spanw_create;
  return self;
}

/*****************************************************************************
 *
 * SpanFirstQuery
 *
 *****************************************************************************/

char *spanfq_to_s(Query *self, char *field)
{
  SpanFirstQuery *sfq = (SpanFirstQuery *)((SpanQuery *)self->data)->data;
  Query *match = sfq->match;
  char *q_str = match->to_s(match, field);
  char *res = strfmt("span_first(%s, %d)", q_str, sfq->end);
  free(q_str);
  return res;
}

void spanfq_extract_terms(Query *self, HashSet *terms)
{
  SpanFirstQuery *sfq = (SpanFirstQuery *)((SpanQuery *)self->data)->data;
  sfq->match->extract_terms(sfq->match, terms);
}

HashSet *spanfq_get_terms(Query *self)
{
  SpanFirstQuery *sfq = (SpanFirstQuery *)((SpanQuery *)self->data)->data;
  SpanQuery *match_sq = (SpanQuery *)sfq->match->data;
  return match_sq->get_terms(sfq->match);
}

Query *spanfq_rewrite(Query *self, IndexReader *ir)
{
  SpanFirstQuery *sfq = (SpanFirstQuery *)((SpanQuery *)self->data)->data;
  Query *q, *rq;

  q = sfq->match;
  rq = q->rewrite(q, ir);
  if (rq == q || self->destroy_all) q_deref(q);
  sfq->match = rq;

  self->ref_cnt++;
  return self;                    /* no clauses rewrote */
}

void spanfq_destroy(Query *self)
{
  SpanQuery *sq = (SpanQuery *)self->data;
  SpanFirstQuery *sfq = (SpanFirstQuery *)sq->data;
  if (self->destroy_all) q_deref(sfq->match);
  free(sfq);
  free(sq);
  q_destroy_i(self);
}

static uint spanfq_hash(Query *self)
{
  SpanFirstQuery *sfq = (SpanFirstQuery *)((SpanQuery *)self->data)->data;
  return sfq->match->hash(sfq->match) ^ sfq->end;
}

static int spanfq_eq(Query *self, Query *o)
{
  SpanFirstQuery *sfq1 = (SpanFirstQuery *)((SpanQuery *)self->data)->data;
  SpanFirstQuery *sfq2 = (SpanFirstQuery *)((SpanQuery *)o->data)->data;
  return sfq1->match->eq(sfq1->match, sfq2->match) && (sfq1->end == sfq2->end);
}

Query *spanfq_create(Query *match, int end)
{
  Query *self = q_create();

  SpanQuery *sq = ALLOC(SpanQuery);

  SpanFirstQuery *sfq = ALLOC(SpanFirstQuery);
  sfq->match = match;
  sfq->end = end;
  sq->data = sfq;

  sq->get_spans = &spanfe_create;
  sq->get_terms = &spanfq_get_terms;
  sq->field = ((SpanQuery *)match->data)->field;
  self->data = sq;

  self->type = SPAN_FIRST_QUERY;
  self->rewrite = &spanfq_rewrite;
  self->extract_terms = &spanfq_extract_terms;
  self->to_s = &spanfq_to_s;
  self->hash = &spanfq_hash;
  self->eq = &spanfq_eq;
  self->destroy_i = &spanfq_destroy;
  self->create_weight_i = &spanw_create;
  return self;
}

/*****************************************************************************
 *
 * SpanOrQuery
 *
 *****************************************************************************/

char *spanoq_to_s(Query *self, char *field)
{
  SpanOrQuery *soq = (SpanOrQuery *)((SpanQuery *)self->data)->data;
  char *res = estrdup("span_or["), *q_str;
  Query *clause;
  int i;
  for (i = 0; i < soq->c_cnt; i++) {
    clause = soq->clauses[i];
    q_str = clause->to_s(clause, field);
    REALLOC_N(res, char, strlen(res) + strlen(q_str) + 10);
    if (i > 0) strcat(res, ", ");
    strcat(res, q_str);
    free(q_str);
  }
  strcat(res, "]");

  return res;
}

void spanoq_extract_terms(Query *self, HashSet *terms)
{
  SpanOrQuery *soq = (SpanOrQuery *)((SpanQuery *)self->data)->data;
  Query *clause;
  int i;
  for (i = 0; i < soq->c_cnt; i++) {
    clause = soq->clauses[i];
    clause->extract_terms(clause, terms);
  }
}

HashSet *spanoq_get_terms(Query *self)
{
  SpanOrQuery *soq = (SpanOrQuery *)((SpanQuery *)self->data)->data;
  HashSet *terms = term_set_create();
  Query *clause;
  int i;
  for (i = 0; i < soq->c_cnt; i++) {
    clause = soq->clauses[i];
    clause->extract_terms(clause, terms);
  }

  return terms;
}

SpanEnum *spanoq_get_spans(Query *self, IndexReader *ir)
{
  SpanOrQuery *soq = (SpanOrQuery *)((SpanQuery *)self->data)->data;
  Query *q;
  if (soq->c_cnt == 1) {
    q = soq->clauses[0];
    return ((SpanQuery *)q->data)->get_spans(q, ir);
  }

  return spanoe_create(self, ir);
}

Query *spanoq_rewrite(Query *self, IndexReader *ir)
{
  SpanOrQuery *soq = (SpanOrQuery *)((SpanQuery *)self->data)->data;
  Query *clause, *rewritten;
  int i;
  /* replace clauses with their rewritten queries */
  for (i = 0; i < soq->c_cnt; i++) {
    clause = soq->clauses[i];
    rewritten = clause->rewrite(clause, ir);
    if ((rewritten == clause) || self->destroy_all) q_deref(clause);
    soq->clauses[i] = rewritten;
  }

  self->ref_cnt++;
  return self;
}

void spanoq_destroy(Query *self)
{
  SpanQuery *sq = (SpanQuery *)self->data;
  SpanOrQuery *soq = (SpanOrQuery *)sq->data;

  if (self->destroy_all) {
    Query *clause;
    int i;
    for (i = 0; i < soq->c_cnt; i++) {
      clause = soq->clauses[i];
      q_deref(clause);
    }
    free(soq->clauses);
  }


  free(soq);
  free(sq);
  q_destroy_i(self);
}

static uint spanoq_hash(Query *self)
{
  int i;
  uint hash = 0;
  Query *q;
  SpanOrQuery *soq = (SpanOrQuery *)((SpanQuery *)self->data)->data;

  for (i = 0; i < soq->c_cnt; i++) {
    q = soq->clauses[i];
    hash ^= q->hash(q);
  }
  return hash;
}

static int spanoq_eq(Query *self, Query *o)
{
  int i;
  Query *q1, *q2;
  SpanOrQuery *soq1 = (SpanOrQuery *)((SpanQuery *)self->data)->data;
  SpanOrQuery *soq2 = (SpanOrQuery *)((SpanQuery *)o->data)->data;
  if (soq1->c_cnt != soq2->c_cnt) return false;
  for (i = 0; i < soq1->c_cnt; i++) {
    q1 = soq1->clauses[i];
    q2 = soq2->clauses[i];
    if (!q1->eq(q1, q2)) return false;
  }
  return true;
}

Query *spanoq_create(Query **clauses, int c_cnt)
{
  Query *self = q_create();

  SpanQuery *sq = ALLOC(SpanQuery);

  SpanOrQuery *soq = ALLOC(SpanOrQuery);
  soq->clauses = clauses;
  soq->c_cnt = c_cnt;
  sq->data = soq;

  sq->get_spans = &spanoq_get_spans;
  sq->get_terms = &spanoq_get_terms;
  sq->field = ((SpanQuery *)clauses[0]->data)->field;
  self->data = sq;

  self->type = SPAN_OR_QUERY;
  self->rewrite = &spanoq_rewrite;
  self->extract_terms = &spanoq_extract_terms;
  self->to_s = &spanoq_to_s;
  self->hash = &spanoq_hash;
  self->eq = &spanoq_eq;
  self->destroy_i = &spanoq_destroy;
  self->create_weight_i = &spanw_create;
  return self;
}

/*****************************************************************************
 *
 * SpanNearQuery
 *
 *****************************************************************************/

char *spannq_to_s(Query *self, char *field)
{
  SpanNearQuery *snq = (SpanNearQuery *)((SpanQuery *)self->data)->data;
  char *res = estrdup("span_near(["), *q_str;
  Query *clause;
  int i;
  for (i = 0; i < snq->c_cnt; i++) {
    clause = snq->clauses[i];
    q_str = clause->to_s(clause, field);
    REALLOC_N(res, char, strlen(res) + strlen(q_str) + 10);
    if (i > 0) strcat(res, ", ");
    strcat(res, q_str);
    free(q_str);
  }
  REALLOC_N(res, char, strlen(res) + 40);
  sprintf(res + strlen(res), "], %d, %s)", snq->slop,
      snq->in_order ? "Ordered" : "Unordered");

  return res;
}

void spannq_extract_terms(Query *self, HashSet *terms)
{
  SpanNearQuery *snq = (SpanNearQuery *)((SpanQuery *)self->data)->data;
  Query *clause;
  int i;
  for (i = 0; i < snq->c_cnt; i++) {
    clause = snq->clauses[i];
    clause->extract_terms(clause, terms);
  }
}

HashSet *spannq_get_terms(Query *self)
{
  SpanNearQuery *snq = (SpanNearQuery *)((SpanQuery *)self->data)->data;
  HashSet *terms = term_set_create();
  Query *clause;
  int i;
  for (i = 0; i < snq->c_cnt; i++) {
    clause = snq->clauses[i];
    clause->extract_terms(clause, terms);
  }

  return terms;
}

SpanEnum *spannq_get_spans(Query *self, IndexReader *ir)
{
  SpanNearQuery *snq = (SpanNearQuery *)((SpanQuery *)self->data)->data;
  Query *q;

  if (snq->c_cnt == 1) {
    q = snq->clauses[0];
    return ((SpanQuery *)q->data)->get_spans(q, ir);
  }

  return spanne_create(self, ir);
}

Query *spannq_rewrite(Query *self, IndexReader *ir)
{
  SpanNearQuery *snq = (SpanNearQuery *)((SpanQuery *)self->data)->data;
  Query *clause, *rewritten;
  int i;
  for (i = 0; i < snq->c_cnt; i++) {
    clause = snq->clauses[i];
    rewritten = clause->rewrite(clause, ir);
    if ((rewritten == clause) || self->destroy_all) q_deref(clause);
    snq->clauses[i] = rewritten;
  }

  self->ref_cnt++;
  return self;
}

void spannq_destroy(Query *self)
{
  SpanQuery *sq = (SpanQuery *)self->data;
  SpanNearQuery *snq = (SpanNearQuery *)sq->data;

  if (self->destroy_all) {
    Query *clause;
    int i;
    for (i = 0; i < snq->c_cnt; i++) {
      clause = snq->clauses[i];
      q_deref(clause);
    }
    free(snq->clauses);
  }

  free(snq);
  free(sq);
  q_destroy_i(self);
}

static uint spannq_hash(Query *self)
{
  int i;
  uint hash = 0;
  Query *q;
  SpanNearQuery *snq = (SpanNearQuery *)((SpanQuery *)self->data)->data;

  for (i = 0; i < snq->c_cnt; i++) {
    q = snq->clauses[i];
    hash ^= q->hash(q);
  }
  return ((hash ^ snq->slop) << 1) | snq->in_order;
}

static int spannq_eq(Query *self, Query *o)
{
  int i;
  Query *q1, *q2;
  SpanNearQuery *snq1 = (SpanNearQuery *)((SpanQuery *)self->data)->data;
  SpanNearQuery *snq2 = (SpanNearQuery *)((SpanQuery *)o->data)->data;
  if (snq1->c_cnt != snq2->c_cnt ||
      snq1->slop != snq2->slop ||
      snq1->in_order != snq2->in_order) {
    return false;
  }

  for (i = 0; i < snq1->c_cnt; i++) {
    q1 = snq1->clauses[i];
    q2 = snq2->clauses[i];
    if (!q1->eq(q1, q2)) return false;
  }

  return true;
}

Query *spannq_create(Query **clauses, int c_cnt, int slop, bool in_order)
{
  Query *self = q_create();

  SpanQuery *sq = ALLOC(SpanQuery);

  SpanNearQuery *snq = ALLOC(SpanNearQuery);
  snq->clauses = clauses;
  snq->c_cnt = c_cnt;
  snq->slop = slop;
  snq->in_order = in_order;
  sq->data = snq;

  sq->get_spans = &spannq_get_spans;
  sq->get_terms = &spannq_get_terms;
  sq->field = ((SpanQuery *)clauses[0]->data)->field;
  self->data = sq;

  self->type = SPAN_NEAR_QUERY;
  self->rewrite = &spannq_rewrite;
  self->extract_terms = &spannq_extract_terms;
  self->to_s = &spannq_to_s;
  self->hash = &spannq_hash;
  self->eq = &spannq_eq;
  self->destroy_i = &spannq_destroy;
  self->create_weight_i = &spanw_create;
  return self;
}

/*****************************************************************************
 *
 * SpanNotQuery
 *
 *****************************************************************************/

char *spanxq_to_s(Query *self, char *field)
{
  SpanNotQuery *sxq = (SpanNotQuery *)((SpanQuery *)self->data)->data;
  char *inc_s = sxq->inc->to_s(sxq->inc, field);
  char *exc_s = sxq->exc->to_s(sxq->exc, field);
  char *res = strfmt("span_not(inc:<%s>, exc:<%s>)", inc_s, exc_s);
  
  free(inc_s);
  free(exc_s);
  return res;
}

void spanxq_extract_terms(Query *self, HashSet *terms)
{
  SpanNotQuery *sxq = (SpanNotQuery *)((SpanQuery *)self->data)->data;
  sxq->inc->extract_terms(sxq->inc, terms);
}

HashSet *spanxq_get_terms(Query *self)
{
  SpanNotQuery *sxq = (SpanNotQuery *)((SpanQuery *)self->data)->data;
  HashSet *terms = term_set_create();
  sxq->inc->extract_terms(sxq->inc, terms);
  return terms;
}

Query *spanxq_rewrite(Query *self, IndexReader *ir)
{
  SpanNotQuery *sxq = (SpanNotQuery *)((SpanQuery *)self->data)->data;
  Query *q, *rq;

  /* rewrite inclusive query */
  q = sxq->inc;
  rq = q->rewrite(q, ir);
  if (rq == q || self->destroy_all) q_deref(q);
  sxq->inc = rq;

  /* rewrite exclusive query */
  q = sxq->exc;
  rq = q->rewrite(q, ir);
  if (rq == q || self->destroy_all) q_deref(q);
  sxq->exc = rq;

  self->ref_cnt++;
  return self;
}

void spanxq_destroy(Query *self)
{
  SpanQuery *sq = (SpanQuery *)self->data;
  SpanNotQuery *sxq = (SpanNotQuery *)sq->data;

  if (self->destroy_all) {
    q_deref(sxq->inc);
    q_deref(sxq->exc);
  }

  free(sxq);
  free(sq);
  q_destroy_i(self);
}

static uint spanxq_hash(Query *self)
{
  SpanNotQuery *sxq = (SpanNotQuery *)((SpanQuery *)self->data)->data;
  return sxq->inc->hash(sxq->inc) ^ sxq->exc->hash(sxq->exc);
}

static int spanxq_eq(Query *self, Query *o)
{
  SpanNotQuery *sxq1 = (SpanNotQuery *)((SpanQuery *)self->data)->data;
  SpanNotQuery *sxq2 = (SpanNotQuery *)((SpanQuery *)o->data)->data;
  return sxq1->inc->eq(sxq1->inc, sxq2->inc) && 
         sxq1->exc->eq(sxq1->exc, sxq2->exc);
}


Query *spanxq_create(Query *inc, Query *exc)
{
  Query *self = q_create();

  SpanQuery *sq = ALLOC(SpanQuery);

  SpanNotQuery *sxq = ALLOC(SpanNotQuery);
  sxq->inc = inc;
  sxq->exc = exc;
  sq->data = sxq;

  sq->get_spans = &spanxe_create;
  sq->get_terms = &spanxq_get_terms;
  sq->field = ((SpanQuery *)inc->data)->field;
  self->data = sq;

  self->type = SPAN_NOT_QUERY;
  self->rewrite = &spanxq_rewrite;
  self->extract_terms = &spanxq_extract_terms;
  self->to_s = &spanxq_to_s;
  self->hash = &spanxq_hash;
  self->eq = &spanxq_eq;
  self->destroy_i = &spanxq_destroy;
  self->create_weight_i = &spanw_create;

  return self;
}

/***************************************************************************
 *
 * SpanScorer
 *
 ***************************************************************************/

float spansc_score(Scorer *self)
{
  SpanScorer *spansc = (SpanScorer *)self->data;
  float raw = sim_tf(spansc->sim, spansc->freq) * spansc->value;

  /* normalize */
  return raw * sim_decode_norm(self->similarity, spansc->norms[self->doc]);
}

bool spansc_next(Scorer *self)
{
  SpanScorer *spansc = (SpanScorer *)self->data;
  SpanEnum *se = spansc->spans;
  int match_length;

  if (spansc->first_time) {
    spansc->more = se->next(se);
    spansc->first_time = false;
  }

  if (!spansc->more) return false;

  spansc->freq = 0.0;
  self->doc = se->doc(se);

  while (spansc->more && (self->doc == se->doc(se))) {
    match_length = se->end(se) - se->start(se);
    spansc->freq += sim_sloppy_freq(spansc->sim, match_length);
    spansc->more = se->next(se);
  }

  return (spansc->more || (spansc->freq != 0.0));
}

bool spansc_skip_to(Scorer *self, int target)
{
  SpanScorer *spansc = (SpanScorer *)self->data;
  SpanEnum *se = spansc->spans;

  spansc->more = se->skip_to(se, target);

  if (!spansc->more) return false;

  spansc->freq = 0.0;
  self->doc = se->doc(se);

  while (spansc->more && (se->doc(se) == target)) {
    spansc->freq += sim_sloppy_freq(spansc->sim, se->end(se) - se->start(se));
    spansc->more = se->next(se);
  }

  return (spansc->more || (spansc->freq != 0.0));
}

Explanation *spansc_explain(Scorer *self, int target)
{
  Explanation *tf_explanation;
  SpanScorer *spansc = (SpanScorer *)self->data;
  float phrase_freq;
  self->skip_to(self, target);
  phrase_freq = (self->doc == target) ? spansc->freq : (float)0.0;

  tf_explanation = expl_create(sim_tf(self->similarity, phrase_freq),
      strfmt("tf(phrase_freq(%f)", phrase_freq));

  return tf_explanation;
}

void spansc_destroy(Scorer *self)
{
  SpanScorer *spansc = (SpanScorer *)self->data;
  if (spansc->spans) spansc->spans->destroy(spansc->spans);
  scorer_destroy_i(self);
}

Scorer *spansc_create(Weight *weight, IndexReader *ir)
{
  Scorer *self = scorer_create(weight->similarity);
  SpanScorer *spansc = ALLOC(SpanScorer);
  SpanQuery *spanq = (SpanQuery *)weight->query->data;
  ZEROSET(spansc, SpanScorer, 1);
  spansc->first_time = true;
  spansc->more = true;
  spansc->spans = spanq->get_spans(weight->query, ir);
  spansc->sim = weight->similarity;
  spansc->norms = ir->get_norms(ir, spanq->field);
  spansc->weight = weight;
  spansc->value = weight->value;
  spansc->freq = 0.0;

  self->data = spansc;

  self->score = &spansc_score;
  self->next = &spansc_next;
  self->skip_to = &spansc_skip_to;
  self->explain = &spansc_explain;
  self->destroy = &spansc_destroy;
  return self;
}
