#include <string.h>
#include "search.h"

/***************************************************************************
 *
 * TermWeight
 *
 ***************************************************************************/

Scorer *tw_scorer(Weight *self, IndexReader *ir)
{
  Term *term = ((TermQuery *)self->query->data)->term;
  TermDocEnum *tde = ir_term_docs_for(ir,term);
  if (!tde) return NULL;

  return tsc_create(self, tde, ir->get_norms_always(ir, term->field));
}

Explanation *tw_explain(Weight *self, IndexReader *ir, int doc_num)
{
  Explanation *qnorm_expl;
  Explanation *field_expl;
  Scorer *scorer;
  Explanation *tf_expl;
  uchar *field_norms;
  float field_norm;
  Explanation *field_norm_expl;

  char *query_str = self->query->to_s(self->query, "");
  TermQuery *tq = (TermQuery *)self->query->data;
  Term *term = tq->term;
  char *field_name = term->field;

  Explanation *expl = expl_create(0.0,
      strfmt("weight(%s in %d), product of:", query_str, doc_num));

  /* We need two of these as it's included in both the query explanation
   * and the field explanation */
  Explanation *idf_expl1 = expl_create(self->idf,
      strfmt("idf(doc_freq=%d)", ir->doc_freq(ir, tq->term)));
  Explanation *idf_expl2 = expl_create(self->idf,
      strfmt("idf(doc_freq=%d)", ir->doc_freq(ir, tq->term)));

  /* explain query weight */
  Explanation *query_expl = expl_create(0.0,
      strfmt("query_weight(%s), product of:", query_str));
  free(query_str);

  if (self->query->boost != 1.0) {
    expl_add_detail(query_expl, expl_create(self->query->boost, estrdup("boost")));
  }

  expl_add_detail(query_expl, idf_expl1);

  qnorm_expl = expl_create(self->qnorm, estrdup("query_norm"));
  expl_add_detail(query_expl, qnorm_expl);

  query_expl->value = self->query->boost * idf_expl1->value * qnorm_expl->value;

  expl_add_detail(expl, query_expl);

  /* explain field weight */
  field_expl = expl_create(0.0, 
      strfmt("field_weight(%s:%s in %d), product of:",
        field_name, term->text, doc_num));

  scorer = self->scorer(self, ir);
  tf_expl = scorer->explain(scorer, doc_num);
  scorer->destroy(scorer);
  expl_add_detail(field_expl, tf_expl);
  expl_add_detail(field_expl, idf_expl2);

  field_norms = ir->get_norms(ir, field_name);
  field_norm = (field_norms 
    ? sim_decode_norm(self->similarity, field_norms[doc_num]) 
    : (float)0.0);
  field_norm_expl = expl_create(field_norm,
      strfmt("field_norm(field=%s, doc=%d)", field_name, doc_num));
  expl_add_detail(field_expl, field_norm_expl);

  field_expl->value = tf_expl->value * idf_expl2->value * field_norm_expl->value;

  /* combine them */
  if (query_expl->value == 1.0) {
    expl_destoy(expl);
    return field_expl;
  } else {
    expl->value = (query_expl->value * field_expl->value);
    expl_add_detail(expl, field_expl);
    return expl;
  }
}

char *tw_to_s(Weight *self)
{
  return strfmt("TermWeight(%f)", self->value);
}

Weight *tw_create(Query *query, Searcher *searcher)
{
  Weight *self = w_create(query);
  self->scorer    = &tw_scorer;
  self->explain   = &tw_explain;
  self->to_s      = &tw_to_s;
  self->sum_of_squared_weights = &w_sum_of_squared_weights;

  self->similarity = query->get_similarity(query, searcher);
  self->idf = sim_idf(self->similarity,
      searcher->doc_freq(searcher, ((TermQuery *)query->data)->term),
      searcher->max_doc(searcher)); // compute idf

  return self;
}

/***************************************************************************
 *
 * TermQuery
 *
 ***************************************************************************/

void tq_destroy(Query *self)
{
  TermQuery *tq = self->data;
  term_destroy(tq->term);
  free(tq);
  q_destroy_i(self);
}

char *tq_to_s(Query *self, char *field)
{
  Term *term = ((TermQuery *)self->data)->term;
  size_t flen = strlen(term->field);
  size_t tlen = strlen(term->text);
  char *buffer = ALLOC_N(char, 34 + flen + tlen);
  char *b = buffer;
  if (strcmp(field, term->field) != 0) {
    memcpy(b, term->field, sizeof(char) * flen);
    b[flen] = ':';
    b += flen + 1;
  }
  memcpy(b, term->text, tlen);
  b += tlen;
  *b = 0;
  if (self->boost != 1.0) {
    *b = '^';
    dbl_to_s(b+1, self->boost);
  }
  return buffer;
}

static void tq_extract_terms(Query *self, HashSet *terms)
{
  Term *term = ((TermQuery *)self->data)->term;
  hs_add(terms, term_clone(term));
}

static uint tq_hash(Query *self)
{
  return term_hash(((TermQuery *)self->data)->term);
}

static int tq_eq(Query *self, Query *o)
{
  return term_eq(((TermQuery *)self->data)->term,
              ((TermQuery *)o->data)->term);
}

Query *tq_create(Term *term)
{
  Query *self = q_create();
  TermQuery *tq = ALLOC(TermQuery);
  tq->term = term;
  self->type = TERM_QUERY;
  self->data = tq;
  self->extract_terms = &tq_extract_terms;
  self->to_s = &tq_to_s;
  self->hash = &tq_hash;
  self->eq = &tq_eq;

  self->destroy_i = &tq_destroy;
  self->create_weight_i = &tw_create;

  return self;
}


/***************************************************************************
 *
 * TermScorer
 *
 ***************************************************************************/

float tsc_score(Scorer *self)
{
  TermScorer *ts = (TermScorer *)self->data;
  int freq = ts->freqs[ts->pointer];
  float score;
  /* compute tf(f)*weight */
  if (freq < SCORE_CACHE_SIZE) {    /* check cache */
    score = ts->score_cache[freq];  /* cache hit */
  } else {
    score = sim_tf(self->similarity, (float)freq) * ts->weight_value; /* cache miss */
  }
  /* normalize for field */
  score *= sim_decode_norm(self->similarity, ts->norms[self->doc]);
  return score;
}

bool tsc_next(Scorer *self)
{
  TermScorer *ts = (TermScorer *)self->data;

  ts->pointer++;
  if (ts->pointer >= ts->pointer_max) {
    /* refill buffer */
    ts->pointer_max = ts->tde->read(ts->tde, ts->docs, ts->freqs, TDE_READ_SIZE);
    if (ts->pointer_max != 0) {
      ts->pointer = 0;
    } else {
      ts->tde->close(ts->tde); /* close stream */
      ts->tde = NULL;
      return false;
    }
  }
  self->doc = ts->docs[ts->pointer];
  return true;
}

bool tsc_skip_to(Scorer *self, int doc_num)
{
  TermScorer *ts = (TermScorer *)self->data;
  TermDocEnum *tde = ts->tde;
  
  /* first scan in cache */
  while (++(ts->pointer) < ts->pointer_max) {
    if (ts->docs[ts->pointer] >= doc_num) {
      self->doc = ts->docs[ts->pointer];
      return true;
    }
  }

  /* not found in cache, seek underlying stream */
  if (tde->skip_to(tde, doc_num)) {
    ts->pointer_max = 1;
    ts->pointer = 0;
    ts->docs[0] = self->doc = tde->doc_num(tde);
    ts->freqs[0] = tde->freq(tde);
    return true;
  } else {
    return false;
  }
}

Explanation *tsc_explain(Scorer *self, int doc_num)
{
  Explanation *tf_explanation;
  TermScorer *ts = (TermScorer *)self->data;
  Query *query = ts->weight->get_query(ts->weight);
  Term *term = ((TermQuery *)query->data)->term;
  int tf = 0;
  TermDocEnum *tde = ts->tde;
  while (ts->pointer < ts->pointer_max) {
    if (ts->docs[ts->pointer] == doc_num)
      tf = ts->freqs[ts->pointer];
    ts->pointer++;
  }
  if (tf == 0) {
    while (tde->next(tde)) {
      if (tde->doc_num(tde) == doc_num) 
        tf = tde->freq(tde);
    }
  }
  tde->close(tde);
  ts->tde = NULL;
  tf_explanation = expl_create(sim_tf(self->similarity, (float)tf),
      strfmt("tf(term_freq(%s:%s)=%d)", term->field, term->text, tf));

  return tf_explanation;
}

void tsc_destroy(Scorer *self)
{
  TermScorer *ts = (TermScorer *)self->data;
  if (ts->tde) ts->tde->close(ts->tde);
  scorer_destroy_i(self);
}

Scorer *tsc_create(Weight *weight, TermDocEnum *tde, uchar *norms)
{
  int i;
  Scorer *self = scorer_create(weight->similarity);
  TermScorer *ts = ALLOC(TermScorer);
  ZEROSET(ts, TermScorer, 1);
  self->data = ts;
  ts->weight = weight;
  ts->tde = tde;
  ts->norms = norms;
  ts->weight_value = weight->value;

  for (i = 0; i < SCORE_CACHE_SIZE; i++) {
    ts->score_cache[i] = sim_tf(self->similarity, (float)i) * ts->weight_value;
  }

  self->score = &tsc_score;
  self->next = &tsc_next;
  self->skip_to = &tsc_skip_to;
  self->explain = &tsc_explain;
  self->destroy = &tsc_destroy;
  return self;
}
