#include <string.h>
#include "search.h"

static char * const NUM_DOCS_ARG_ERROR_MSG = "num_docs must be > 0 to run a search";
static char * const FIRST_DOC_ARG_ERROR_MSG = "first_doc must be >= 0 to run a search";

/***************************************************************************
 *
 * Explanation
 *
 ***************************************************************************/

Explanation *expl_create(float value, char *description)
{
  Explanation *self = ALLOC(Explanation);
  self->value = value;
  self->description = description;
  self->dcnt = 0;
  self->dcapa = EXPLANATION_DETAILS_START_SIZE;
  self->details = ALLOC_N(Explanation *, EXPLANATION_DETAILS_START_SIZE);
  return self;
}

void expl_destoy(void *p)
{
  Explanation *expl = (Explanation *)p;
  int i;
  for (i = 0; i < expl->dcnt; i++) {
    expl_destoy(expl->details[i]);
  }
  free(expl->details);
  free(expl->description);
  free(expl);
}

Explanation *expl_add_detail(Explanation *self, Explanation *detail)
{
  if (self->dcnt >= self->dcapa) {
    self->dcapa *= 2;
    REALLOC_N(self->details, Explanation *, self->dcapa);
  }
  self->details[self->dcnt] = detail;
  self->dcnt++;
  return self;
}

char *expl_to_s(Explanation *self, int depth)
{
  int i;
  char *buffer = ALLOC_N(char, depth * 2 + 1);
  memset(buffer, ' ', sizeof(char) * depth * 2);
  buffer[depth*2] = 0;

  buffer = estrcat(buffer, strfmt("%f = %s\n", self->value, self->description));
  for (i = 0; i < self->dcnt; i++) {
    buffer = estrcat(buffer, expl_to_s(self->details[i], depth + 1));
  }

  return buffer;
}

char *expl_to_html(Explanation *self)
{
  int i;
  char *buffer;
  buffer = strfmt("<ul>\n<li>%f = %s</li>\n", self->value, self->description);

  for (i = 0; i < self->dcnt; i++) {
    estrcat(buffer, expl_to_html(self->details[i]));
  }

  REALLOC_N(buffer, char, strlen(buffer) + 10);
  return strcat(buffer, "</ul>\n");
}

/***************************************************************************
 *
 * Hit
 *
 ***************************************************************************/

bool hit_less_than(void *hit1, void *hit2)
{
  if (((Hit *)hit1)->score == ((Hit *)hit2)->score) {
    return ((Hit *)hit1)->doc > ((Hit *)hit2)->doc;
  } else {
    return ((Hit *)hit1)->score < ((Hit *)hit2)->score;
  }
}

inline bool hit_lt(Hit *hit1, Hit *hit2)
{
  if (hit1->score == hit2->score) {
    return hit1->doc > hit2->doc;
  } else {
    return hit1->score < hit2->score;
  }
}

void hit_pq_down(PriorityQueue *pq)
{
  register int i = 1;
  register int j = 2; //i << 1;
  register int k = 3; //j + 1;
  Hit **heap = (Hit **)pq->heap;
  Hit *node = heap[i]; // save top node

  if ((k <= pq->count) && hit_lt(heap[k], heap[j]))
    j = k;

  while ((j <= pq->count) && hit_lt(heap[j], node)) {
    heap[i] = heap[j];      // shift up child
    i = j;
    j = i << 1;
    k = j + 1;
    if ((k <= pq->count) && hit_lt(heap[k], heap[j]))
      j = k;
  }
  heap[i] = node;
}

Hit *hit_pq_pop(PriorityQueue *pq)
{
  if (pq->count > 0) {
    Hit *result = (Hit *)pq->heap[1];  // save first value
    pq->heap[1] = pq->heap[pq->count]; // move last to first
    pq->heap[pq->count] = NULL;
    pq->count--;
    hit_pq_down(pq);                   // adjust heap
    return result;
  } else {
    return NULL;
  }
}

inline void hit_pq_up(PriorityQueue *pq)
{
  Hit **heap = (Hit **)pq->heap;
  Hit *node;
  int i = pq->count;
  int j = i >> 1;
  node = heap[i];

  while ((j > 0) && hit_lt(node, heap[j])) {
    heap[i] = heap[j];
    i = j;
    j = j >> 1;
  }
  heap[i] = node;
}

void hit_pq_insert(PriorityQueue *pq, Hit *hit) 
{
  if (pq->count < pq->size) {
    Hit *new_hit = ALLOC(Hit);
    memcpy(new_hit, hit, sizeof(Hit));
    pq->count++;
    pq->heap[pq->count] = new_hit;
    hit_pq_up(pq);
  } else if (pq->count > 0 && hit_lt((Hit *)pq->heap[1], hit)) {
    memcpy(pq->heap[1], hit, sizeof(Hit));
    hit_pq_down(pq);
  }
}

/***************************************************************************
 *
 * TopDocs
 *
 ***************************************************************************/

TopDocs *td_create(int total_hits, int size, Hit **hits)
{
  TopDocs *td = ALLOC(TopDocs);
  td->total_hits = total_hits;
  td->size = size;
  td->hits = hits;
  return td;
}

void td_destroy(TopDocs *td)
{
  int i;
  for (i = 0; i < td->size; i++) {
    free(td->hits[i]);
  }
  free(td->hits);
  free(td);
}

char *td_to_s(TopDocs *td)
{
  int i;
  Hit *hit;
  char *buffer = strfmt("%d hits sorted by <score, doc_num>\n", td->total_hits);
  for (i = 0; i < td->size; i++) {
    hit = td->hits[i];
    estrcat(buffer, strfmt("\t%d:%f\n", hit->doc, hit->score));
  }
  return buffer;
}

/***************************************************************************
 *
 * Weight
 *
 ***************************************************************************/

Query *w_get_query(Weight *self)
{
  return self->query;
}

float w_get_value(Weight *self)
{
  return self->value;
}

float w_sum_of_squared_weights(Weight *self)
{
  self->qweight = self->idf * self->query->boost;
  return self->qweight * self->qweight; // square it
}

void w_normalize(Weight *self, float normalization_factor)
{
  self->qnorm = normalization_factor;
  self->qweight *= normalization_factor; // normalize query weight
  self->value = self->qweight * self->idf; // idf for document
}

void w_destroy(Weight *self)
{
  q_deref(self->query);
  free(self);
}

Weight *w_create(Query *query)
{
  Weight *self = ALLOC_AND_ZERO_N(Weight, 1);
  ref(query);
  self->query = query;

  self->get_query = &w_get_query;
  self->get_value = &w_get_value;
  self->normalize = &w_normalize;
  self->destroy   = &w_destroy;
  return self;
}

/***************************************************************************
 *
 * Query
 *
 ***************************************************************************/

Similarity *q_get_similarity_i(Query *self, Searcher *searcher)
{
  return searcher->get_similarity(searcher);
}

Query *q_rewrite(Query *self, IndexReader *ir)
{
  self->ref_cnt++;
  return self;
}

Weight *q_weight(Query *self, Searcher *searcher)
{
  Query  *query   = searcher->rewrite(searcher, self);
  Weight *weight  = query->create_weight_i(query, searcher);
  float   sum     = weight->sum_of_squared_weights(weight);
  Similarity *sim = query->get_similarity(query, searcher);
  float   norm    = sim_query_norm(sim, sum);
  q_deref(query);

  weight->normalize(weight, norm);
  return self->weight = weight;
}

Weight *q_create_weight_unsup(Query *self, Searcher *searcher)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return NULL;
}

void q_destroy_i(Query *self)
{
  free(self);
}

void q_extract_terms(Query *self, HashSet *terms)
{
  /* do nothing by default */
}

void q_deref(Query *self)
{  
  if (--self->ref_cnt == 0) {
    self->destroy_i(self);
  }
}

Query *q_create()
{
  Query *self = ALLOC(Query);
  ZEROSET(self, Query, 1);
  self->destroy_all = true;
  self->boost = 1.0;
  self->rewrite = &q_rewrite;
  self->get_similarity = &q_get_similarity_i;
  self->extract_terms = &q_extract_terms;
  self->weight = NULL;
  self->ref_cnt = 1;
  return self;
}

uint q_hash(Query *self)
{
  return (self->hash(self) << 4) | self->type;
}

int q_eq(Query *self, Query *o)
{
  return (self == o) || ((self->type == o->type) &&
                         (self->boost == o->boost) &&
                         self->eq(self, o));
}

Query *q_combine(Query **queries, int q_cnt)
{
  int i;
  Query *q, *ret_q;
  HashSet *uniques =
    hs_create((hash_ft)&q_hash, (eq_ft)&q_eq, NULL);

  for (i = 0; i < q_cnt; i++) {
    q = queries[i];
    if (q->type == BOOLEAN_QUERY) {
      int j;
      bool splittable = true;
      BooleanQuery *bq = (BooleanQuery *)q->data;
      if (bq->coord_disabled == false) {
        splittable = false;
      } else {
        for (j = 0; j < bq->clause_cnt; j++) {
          if (bq->clauses[j]->occur != BC_SHOULD) {
            splittable = false;
            break;
          }
        }
      }
      if (splittable) {
        for (j = 0; j < bq->clause_cnt; j++) {
          q = bq->clauses[j]->query;
          hs_add(uniques, q);
        }
      } else {
        hs_add(uniques, q);
      }
    } else {
      hs_add(uniques, q);
    }
  }
  if (uniques->size == 1) {
    ret_q = (Query *)uniques->elems[0]; 
    ref(ret_q);
  } else {
    ret_q = bq_create(true);
    for (i = 0; i < uniques->size; i++) {
      q = (Query *)uniques->elems[i];
      ref(q);
      bq_add_query(ret_q, q, BC_SHOULD);
    }
  }
  hs_destroy(uniques);

  return ret_q;
}

/***************************************************************************
 *
 * Scorer
 *
 ***************************************************************************/

void scorer_destroy_i(Scorer *self)
{
  free(self->data);
  free(self);
}

Scorer *scorer_create(Similarity *similarity)
{
  Scorer *self = ALLOC(Scorer);
  self->destroy = &scorer_destroy_i;
  self->data = NULL;
  self->similarity = similarity;
  return self;
}

bool scorer_less_than(void *p1, void *p2)
{
  Scorer *s1 = (Scorer *)p1;
  Scorer *s2 = (Scorer *)p2;
  return s1->score(s1) < s2->score(s2);
}

bool scorer_doc_less_than(void *p1, void *p2)
{
  return ((Scorer *)p1)->doc < ((Scorer *)p2)->doc;
}

int scorer_doc_cmp(const void *p1, const void *p2)
{
  return (*(Scorer **)p1)->doc - (*(Scorer **)p2)->doc;
}

/***************************************************************************
 *
 * Searcher
 *
 ***************************************************************************/

static int s_doc_freq(Searcher *self, Term *term)
{
  return self->ir->doc_freq(self->ir, term);
}

static int *s_doc_freqs(Searcher *self, Term **terms, int tcnt)
{
  int i;
  int *freqs = ALLOC_N(int, tcnt);

  for (i = 0; i < tcnt; i++) {
    freqs[i] = self->ir->doc_freq(self->ir, terms[i]);
  }
  return freqs;
}

static int *ss_doc_freqs(Searcher *self, Term **terms, int tcnt)
{
  int i;
  int *freqs = ALLOC_N(int, tcnt);

  for (i = 0; i < tcnt; i++) {
    freqs[i] = self->doc_freq(self, terms[i]);
  }

  return freqs;
}


static Document *s_get_doc(Searcher *self, int doc_num)
{
  return self->ir->get_doc(self->ir, doc_num);
}

static int s_max_doc(Searcher *self)
{
  return self->ir->max_doc(self->ir);
}

static Weight *s_create_weight(Searcher *self, Query *query)
{
  return q_weight(query, self);
}

static TopDocs *s_search(Searcher *self, Query *query, int first_doc,
    int num_docs, Filter *filter, Sort *sort)
{
  int max_size = first_doc + num_docs;
  int i;
  Weight *weight;
  Scorer *scorer;
  Hit **score_docs = NULL;
  Hit hit;
  int total_hits = 0;
  float score;
  BitVector *bits = (filter ? filter->get_bv(filter, self->ir) : NULL);
  Hit *(*hq_pop)(PriorityQueue *pq);
  void (*hq_insert)(PriorityQueue *pq, Hit *hit);
  void (*hq_destroy)(PriorityQueue *self);
  PriorityQueue *hq;


  if (num_docs <= 0)
    RAISE(ARG_ERROR, NUM_DOCS_ARG_ERROR_MSG);

  if (first_doc < 0)
    RAISE(ARG_ERROR, FIRST_DOC_ARG_ERROR_MSG);

  weight = q_weight(query, self);
  scorer = weight->scorer(weight, self->ir);
  if (!scorer) {
    if (bits) bv_destroy(bits);
    weight->destroy(weight);
    return td_create(0, 0, NULL);
  }

  if (sort) {
    hq = fshq_pq_create(max_size, sort, self->ir);
    hq_pop = &fshq_pq_pop;
    hq_insert = &fshq_pq_insert;
    hq_destroy = &fshq_pq_destroy;
  } else {
    hq = pq_create(max_size, &hit_less_than);
    hq_pop = &hit_pq_pop;
    hq_insert = &hit_pq_insert;
    hq_destroy = &pq_destroy;
  }

  while (scorer->next(scorer)) {
    if (bits && !bv_get(bits, scorer->doc)) continue;
    total_hits++;
    score = scorer->score(scorer);
    hit.doc = scorer->doc; hit.score = score;
    hq_insert(hq, &hit);
  }
  scorer->destroy(scorer);
  weight->destroy(weight);

  if (hq->count > first_doc) {
    if ((hq->count - first_doc) < num_docs) {
      num_docs = hq->count - first_doc;
    }
    score_docs = ALLOC_N(Hit *, num_docs);
    for (i = num_docs - 1; i >= 0; i--) {
      score_docs[i] = hq_pop(hq);
      //hit = score_docs[i] = pq_pop(hq);
      //printf("hit = %d-->%f\n", hit->doc, hit->score);
    }
  } else {
    num_docs = 0;
  }
  pq_clear(hq);
  hq_destroy(hq);

  if (bits) bv_destroy(bits);
  return td_create(total_hits, num_docs, score_docs);
}

static void s_search_each_w(Searcher *self, Weight *weight, Filter *filter,
    void (*fn)(Searcher *, int, float, void *), void *arg)
{
  Scorer *scorer;
  BitVector *bits = (filter ? filter->get_bv(filter, self->ir) : NULL);

  scorer = weight->scorer(weight, self->ir);
  if (!scorer) {
    if (bits) bv_destroy(bits);
    return;
  }

  while (scorer->next(scorer)) {
    if (bits && !bv_get(bits, scorer->doc)) continue;
    fn(self, scorer->doc, scorer->score(scorer), arg);
  }
  scorer->destroy(scorer);
}

static void s_search_each(Searcher *self, Query *query, Filter *filter,
    void (*fn)(Searcher *, int, float, void *), void *arg)
{
  Weight *weight = q_weight(query, self);
  s_search_each_w(self, weight, filter, fn, arg);
  weight->destroy(weight);
}

static Query *s_rewrite(Searcher *self, Query *original)
{
  int q_is_destroyed = false;
  Query *query = original;
  Query *rewritten_query = query->rewrite(query, self->ir);
  while (q_is_destroyed || (query != rewritten_query)) {
    query = rewritten_query;
    rewritten_query = query->rewrite(query, self->ir);
    q_is_destroyed = (query->ref_cnt <= 1);
    q_deref(query); /* destroy intermediate queries */
  }
  return query;
}

static Explanation *s_explain(Searcher *self, Query *query, int doc_num)
{
  Weight *weight = q_weight(query, self);
  Explanation *e =  weight->explain(weight, self->ir, doc_num);
  weight->destroy(weight);
  return e;
}

static Explanation *s_explain_w(Searcher *self, Weight *w, int doc_num)
{
  return w->explain(w, self->ir, doc_num);
}

static Similarity *s_get_similarity(Searcher *self)
{
  return self->similarity;
}

static void s_close(Searcher *self)
{
  if (self->ir && self->close_ir) {
    ir_close(self->ir);
  }
  free(self);
}

Searcher *sea_create(IndexReader *ir)
{
  Searcher *self = ALLOC(Searcher);
  self->ir = ir;
  self->close_ir = true;
  self->similarity = sim_create_default();
  self->doc_freq = &s_doc_freq;
  self->doc_freqs = &s_doc_freqs;
  self->get_doc = &s_get_doc;
  self->max_doc = &s_max_doc;
  self->create_weight = &s_create_weight;
  self->search = &s_search;
  self->search_each = &s_search_each;
  self->search_each_w = &s_search_each_w;
  self->rewrite = &s_rewrite;
  self->explain = &s_explain;
  self->explain_w = &s_explain_w;
  self->get_similarity = &s_get_similarity;
  self->close = &s_close;
  return self;
}

/***************************************************************************
 *
 * CachedDFSearcher
 *
 ***************************************************************************/

typedef struct CachedDFSearcher {
  HshTable *df_map;
  int max_doc;
} CachedDFSearcher;

static int cdfsea_doc_freq(Searcher *self, Term *term)
{
  CachedDFSearcher *cdfsea = (CachedDFSearcher *)self->data;
  return (int)h_get(cdfsea->df_map, term);
}

static Document *cdfsea_get_doc(Searcher *self, int doc_num)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return NULL;
}

static int cdfsea_max_doc(Searcher *self)
{
  return ((CachedDFSearcher *)self->data)->max_doc;
}

static Weight *cdfsea_create_weight(Searcher *self, Query *query)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return NULL;
}

static TopDocs *cdfsea_search(Searcher *self, Query *query, int first_doc,
    int num_docs, Filter *filter, Sort *sort)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return NULL;
}

static void cdfsea_search_each(Searcher *self, Query *query, Filter *filter,
    void (*fn)(Searcher *, int, float, void *), void *arg)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
}

static void cdfsea_search_each_w(Searcher *self, Weight *w, Filter *filter,
    void (*fn)(Searcher *, int, float, void *), void *arg)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
}

static Query *cdfsea_rewrite(Searcher *self, Query *original)
{
  original->ref_cnt++;
  return original;
}

static Explanation *cdfsea_explain(Searcher *self, Query *query, int doc_num)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return NULL;
}

static Explanation *cdfsea_explain_w(Searcher *self, Weight *w, int doc_num)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return NULL;
}

static Similarity *cdfsea_get_similarity(Searcher *self)
{
  RAISE(UNSUPPORTED_ERROR, UNSUPPORTED_ERROR_MSG);
  return NULL;
}

static void cdfsea_close(Searcher *self)
{
  CachedDFSearcher *cdfsea = (CachedDFSearcher *)self->data;
  h_destroy(cdfsea->df_map);
  free(cdfsea);
  free(self);
}

Searcher *cdfsea_create(HshTable *df_map, int max_doc)
{
  Searcher *self = ALLOC(Searcher);

  CachedDFSearcher *cdfsea = ALLOC(CachedDFSearcher);

  cdfsea->df_map = df_map;
  cdfsea->max_doc = max_doc;
  self->data = cdfsea;

  self->doc_freq = &cdfsea_doc_freq;
  self->doc_freqs = &ss_doc_freqs;
  self->get_doc = &cdfsea_get_doc;
  self->max_doc = &cdfsea_max_doc;
  self->create_weight = &cdfsea_create_weight;
  self->search = &cdfsea_search;
  self->search_each = &cdfsea_search_each;
  self->search_each_w = &cdfsea_search_each_w;
  self->rewrite = &cdfsea_rewrite;
  self->explain = &cdfsea_explain;
  self->explain_w = &cdfsea_explain_w;
  self->get_similarity = &cdfsea_get_similarity;
  self->close = &cdfsea_close;
  return self;
}

/***************************************************************************
 *
 * MultiSearcher
 *
 ***************************************************************************/

static inline int msea_get_searcher_index(Searcher *self, int n)
{
  MultiSearcher *msea = (MultiSearcher *)self->data;
  int lo = 0;               /* search starts array */
  int hi = msea->s_cnt - 1; /* for 1st element < n, return its index */
  int mid, mid_val;

  while (hi >= lo) {
    mid = (lo + hi) >> 1;
    mid_val = msea->starts[mid];
    if (n < mid_val) {
      hi = mid - 1;
    } else if (n > mid_val) {
      lo = mid + 1;
    } else {                  /* found a match */
      while (((mid+1) < msea->s_cnt) && (msea->starts[mid+1] == mid_val)) {
        mid++;                /* scan to last match */
      }
      return mid;
    }
  }
  return hi;
}

static int msea_doc_freq(Searcher *self, Term *term)
{
  int i;
  int doc_freq = 0;
  Searcher *s;
  MultiSearcher *msea = (MultiSearcher *)self->data;
  for (i = 0; i < msea->s_cnt; i++) {
    s = msea->searchers[i];
    doc_freq += s->doc_freq(s, term);
  }

  return doc_freq;
}

static Document *msea_get_doc(Searcher *self, int doc_num)
{
  MultiSearcher *msea = (MultiSearcher *)self->data;
  int i = msea_get_searcher_index(self, doc_num);
  Searcher *s = msea->searchers[i];
  return s->get_doc(s, doc_num - msea->starts[i]);
}

static int msea_max_doc(Searcher *self)
{
  return ((MultiSearcher *)self->data)->max_doc;
}

static Weight *msea_create_weight(Searcher *self, Query *query)
{
  int i, *dfs;
  Searcher *cdfsea;
  Weight *w;
  HshTable *df_map = h_new((hash_ft)&term_hash, (eq_ft)&term_eq,
                          (free_ft)NULL, (free_ft)NULL);
  Query *rq = self->rewrite(self, query);
  HashSet *terms = term_set_create();
  rq->extract_terms(rq, terms);
  dfs = self->doc_freqs(self, (Term **)terms->elems, terms->size);

  for (i = 0; i < terms->size; i++) {
    h_set(df_map, terms->elems[i], (void *)dfs[i]); 
  }
  /* don't destroy the individual terms, only the HashSet */
  hs_destroy(terms);
  free(dfs);

  cdfsea = cdfsea_create(df_map, ((MultiSearcher *)self->data)->max_doc);

  w = q_weight(rq, cdfsea);
  q_deref(rq);
  cdfsea->close(cdfsea);

  return w;
}

struct MultiSearchEachArg {
  int start;
  void *arg;
  void (*fn)(Searcher *, int, float, void *);
};

void msea_search_each_i(Searcher *self, int doc_num, float score, void *arg)
{
  struct MultiSearchEachArg *mse_arg = (struct MultiSearchEachArg *)arg;

  mse_arg->fn(self, doc_num + mse_arg->start, score, mse_arg->arg);
}

static void msea_search_each_w(Searcher *self, Weight *w, Filter *filter,
    void (*fn)(Searcher *, int, float, void *), void *arg)
{
  int i;
  struct MultiSearchEachArg mse_arg;
  MultiSearcher *msea = (MultiSearcher *)self->data;
  Searcher *s;

  mse_arg.fn = fn;
  mse_arg.arg = arg;
  for (i = 0; i < msea->s_cnt; i++) {
    s = msea->searchers[i];
    mse_arg.start = msea->starts[i];
    s->search_each_w(s, w, filter, &msea_search_each_i, &mse_arg);
  }
}

static void msea_search_each(Searcher *self, Query *query, Filter *filter,
    void (*fn)(Searcher *, int, float, void *), void *arg)
{
  Weight *w = q_weight(query, self);
  msea_search_each_w(self, w, filter, fn, arg);
  w->destroy(w);
}

struct MultiSearchArg {
  int total_hits, max_size;
  PriorityQueue *hq;
  void (*hq_insert)(PriorityQueue *pq, Hit *hit);
};

void msea_search_i(Searcher *self, int doc_num, float score, void *arg)
{
  struct MultiSearchArg *ms_arg = (struct MultiSearchArg *)arg;
  Hit hit;

  ms_arg->total_hits++;
  hit.doc = doc_num;
  hit.score = score;
  ms_arg->hq_insert(ms_arg->hq, &hit);
}

static TopDocs *msea_search(Searcher *self, Query *query, int first_doc,
    int num_docs, Filter *filter, Sort *sort)
{
  int max_size = first_doc + num_docs;
  int i;
  Weight *weight;
  Hit **score_docs = NULL;
  BitVector *bits = (filter ? filter->get_bv(filter, self->ir) : NULL);
  Hit *(*hq_pop)(PriorityQueue *pq);
  void (*hq_insert)(PriorityQueue *pq, Hit *hit);
  void (*hq_destroy)(PriorityQueue *self);
  PriorityQueue *hq;
  struct MultiSearchArg ms_arg;


  if (num_docs <= 0)
    RAISE(ARG_ERROR, NUM_DOCS_ARG_ERROR_MSG);

  if (first_doc < 0)
    RAISE(ARG_ERROR, FIRST_DOC_ARG_ERROR_MSG);

  weight = q_weight(query, self);
  if (sort) {
    hq = fshq_pq_create(max_size, sort, self->ir);
    hq_pop = &fshq_pq_pop;
    hq_insert = &fshq_pq_insert;
    hq_destroy = &fshq_pq_destroy;
  } else {
    hq = pq_create(max_size, &hit_less_than);
    hq_pop = &hit_pq_pop;
    hq_insert = &hit_pq_insert;
    hq_destroy = &pq_destroy;
  }


  ms_arg.hq = hq;
  ms_arg.total_hits = 0;
  ms_arg.max_size = max_size;
  ms_arg.hq_insert = hq_insert;

  msea_search_each_w(self, weight, filter, msea_search_i, &ms_arg);

  weight->destroy(weight);

  if (hq->count > first_doc) {
    if ((hq->count - first_doc) < num_docs) {
      num_docs = hq->count - first_doc;
    }
    score_docs = ALLOC_N(Hit *, num_docs);
    for (i = num_docs - 1; i >= 0; i--) {
      score_docs[i] = hq_pop(hq);
      //hit = score_docs[i] = pq_pop(hq);
      //printf("hit = %d-->%f\n", hit->doc, hit->score);
    }
  } else {
    num_docs = 0;
  }
  pq_clear(hq);
  hq_destroy(hq);

  if (bits) bv_destroy(bits);
  return td_create(ms_arg.total_hits, num_docs, score_docs);
}

static Query *msea_rewrite(Searcher *self, Query *original)
{
  int i;
  Searcher *s;
  MultiSearcher *msea = (MultiSearcher *)self->data;
  Query **queries = ALLOC_N(Query *, msea->s_cnt), *rewritten;

  for (i = 0; i < msea->s_cnt; i++) {
    s = msea->searchers[i];
    queries[i] = s->rewrite(s, original);
  }
  rewritten = q_combine(queries, msea->s_cnt);

  for (i = 0; i < msea->s_cnt; i++) {
    q_deref(queries[i]);
  }
  free(queries);
  return rewritten;
}

static Explanation *msea_explain(Searcher *self, Query *query, int doc_num)
{
  MultiSearcher *msea = (MultiSearcher *)self->data;
  int i = msea_get_searcher_index(self, doc_num);
  Weight *w = q_weight(query, self);
  Searcher *s = msea->searchers[i];
  Explanation *e = s->explain_w(s, w, doc_num - msea->starts[i]);
  w->destroy(w);
  return e;
}

static Explanation *msea_explain_w(Searcher *self, Weight *w, int doc_num)
{
  MultiSearcher *msea = (MultiSearcher *)self->data;
  int i = msea_get_searcher_index(self, doc_num);
  Searcher *s = msea->searchers[i];
  Explanation *e = s->explain_w(s, w, doc_num - msea->starts[i]);
  return e;
}

static Similarity *msea_get_similarity(Searcher *self)
{
  return self->similarity;
}

static void msea_close(Searcher *self)
{
  int i;
  Searcher *s;
  MultiSearcher *msea = (MultiSearcher *)self->data;
  if (msea->close_subs) {
    for (i = 0; i < msea->s_cnt; i++) {
      s = msea->searchers[i];
      s->close(s);
    }
    free(msea->searchers);
  }
  free(msea->starts);
  free(msea);
  free(self);
}

Searcher *msea_create(Searcher **searchers, int s_cnt, bool close_subs)
{
  int i, max_doc = 0, *starts;
  Searcher *self = ALLOC(Searcher);

  MultiSearcher *msea = ALLOC(MultiSearcher);

  starts = ALLOC_N(int, s_cnt + 1);
  for (i = 0; i < s_cnt; i++) {
    starts[i] = max_doc;
    max_doc += searchers[i]->max_doc(searchers[i]);
  }
  starts[i] = max_doc;

  msea->s_cnt = s_cnt;
  msea->searchers = searchers;
  msea->starts = starts;
  msea->max_doc = max_doc;
  msea->close_subs = close_subs;
  self->data = msea;

  self->ir = (IndexReader *)NULL;
  self->similarity = sim_create_default();
  self->doc_freq = &msea_doc_freq;
  self->doc_freqs = &ss_doc_freqs;
  self->get_doc = &msea_get_doc;
  self->max_doc = &msea_max_doc;
  self->create_weight = &msea_create_weight;
  self->search = &msea_search;
  self->search_each = &msea_search_each;
  self->search_each_w = &msea_search_each_w;
  self->rewrite = &msea_rewrite;
  self->explain = &msea_explain;
  self->explain_w = &msea_explain_w;
  self->get_similarity = &msea_get_similarity;
  self->close = &msea_close;
  return self;
}
