#include <search.h>
#include <global.h>
#include <math.h>

static int low_bit = 0, low_mid_bit = 0, high_mid_bit = 0, high_bit = 0;
static void
setup_endian()
{
    static int init = 0;
    char *p;

    if (init) return;
    init = 1;
    p = (char*)&init;

    if (p[0]) {
      low_bit = 0;
      low_mid_bit = 1;
      high_mid_bit = 2;
      high_bit = 3;
    } else {
      low_bit = 3;
      low_mid_bit = 2;
      high_mid_bit = 1;
      high_bit = 0;
    }
}

float byte_to_float(uchar b)
{
  char flt[4];
  if (b == 0) {
    return 0.0;
  } else {
    int mantissa = b & 0x07;           // 0x07 =  7 = 0b00000111
    int exponent = (b >> 3) & 0x1F;    // 0x1f = 31 = 0b00011111

    if (!low_mid_bit) setup_endian();
    flt[low_bit] = flt[low_mid_bit] = 0;
    flt[high_mid_bit] = mantissa << 5;
    flt[high_bit] = exponent + 48;
    return *((float *)flt);
  }
}

uchar float_to_byte(float f)
{
  if (f <= 0.0) {
    return 0;
  } else {
    char *bits = (char *)&f;
    int mantissa = (bits[high_mid_bit] & 0xEf) >> 5;
    int exponent = (bits[high_bit] - 48);

    if (exponent > 0x1f) {
      exponent = 0x1f;   // 0x1f = 31 = 0b00011111
      mantissa = 0x07;   // 0x07 =  7 = 0b00000111
    }

    if (exponent < 0) {
      exponent = 0;
      mantissa = 1;
    }

    return ((exponent<<3) | mantissa);
  }
}

float simdef_length_norm(Similarity *s, char *field, int num_terms)
{
  return (float)(1.0 / sqrt(num_terms));
}

float simdef_query_norm(struct Similarity *s, float sum_of_squared_weights)
{
  return (float)(1.0 / sqrt(sum_of_squared_weights));
}

float simdef_tf(struct Similarity *s, float freq)
{
  return (float)sqrt(freq);
}

float simdef_sloppy_freq(struct Similarity *s, int distance)
{
  return (float)(1.0 / (double)(distance + 1));
}

float simdef_idf_term(struct Similarity *s, Term *term, Searcher *searcher)
{
  return s->idf(s, searcher->doc_freq(searcher, term), searcher->max_doc(searcher));
}

float simdef_idf_phrase(struct Similarity *s, Term **terms, int tcnt, Searcher *searcher)
{
  float idf = 0.0;
  int i;
  for (i = 0; i < tcnt; i++) {
    idf += s->idf_term(s, terms[i], searcher);
  }
  return idf;
}

float simdef_idf(struct Similarity *s, int doc_freq, int num_docs)
{
  return (float)(log((float)num_docs/(float)(doc_freq+1)) + 1.0);
}

float simdef_coord(struct Similarity *s, int overlap, int max_overlap)
{
  return (float)((double)overlap / (double)max_overlap);
}

float simdef_decode_norm(struct Similarity *s, uchar b)
{
  return s->norm_table[b];
}

uchar simdef_encode_norm(struct Similarity *s, float f)
{
  return float_to_byte(f);
}

void simdef_destroy(Similarity *s)
{
  /* nothing to do here */
}

#ifdef WIN32
static Similarity default_similarity = {
  NULL,
  {0},
  &simdef_length_norm,
  &simdef_query_norm,
  &simdef_tf,
  &simdef_sloppy_freq,
  &simdef_idf_term,
  &simdef_idf_phrase,
  &simdef_idf,
  &simdef_coord,
  &simdef_decode_norm,
  &simdef_encode_norm,
  &simdef_destroy
};
#else
static Similarity default_similarity = {
  data:NULL,
  length_norm:&simdef_length_norm,
  query_norm:&simdef_query_norm,
  tf:&simdef_tf,
  sloppy_freq:&simdef_sloppy_freq,
  idf_term:&simdef_idf_term,
  idf_phrase:&simdef_idf_phrase,
  idf:&simdef_idf,
  coord:&simdef_coord,
  decode_norm:&simdef_decode_norm,
  encode_norm:&simdef_encode_norm,
  destroy:&simdef_destroy
};
#endif

Similarity *sim_create_default()
{
  int i;
  if (!default_similarity.data) {
    for (i = 0; i < 256; i++)
      default_similarity.norm_table[i] = byte_to_float(i);

    default_similarity.data = &default_similarity;
  }
  return &default_similarity;
}
