#include "store.h"

static char * const COULD_NOT_OBTAIN_LOCK = "Could not obtain lock";

/**
 * Call the function +func+ with the +lock+ locked. The argument +arg+ will be
 * passed to +func+. If you need to pass more than one argument you should use
 * a struct. When the function is finished, release the lock.
 * 
 * @param lock lock to be locked while func is called
 * @param func function to call with the lock locked
 * @param arg  argument to pass to the function
 * @throws IO_ERROR if the lock is already locked
 * @see with_lock_name
 */
void with_lock(Lock *lock, void (*func)(void *arg), void *arg)
{
  if (!lock->obtain(lock)) {
    RAISE(IO_ERROR, COULD_NOT_OBTAIN_LOCK);
  }
  func(arg);
  lock->release(lock);
}

/**
 * Create a lock in the +store+ with the name +lock_name+. Call the function
 * +func+ with the lock locked. The argument +arg+ will be passed to +func+.
 * If you need to pass more than one argument you should use a struct. When
 * the function is finished, release and destroy the lock.
 * 
 * @param store     store to open the lock in
 * @param lock_name name of the lock to open
 * @param func function to call with the lock locked
 * @param arg       argument to pass to the function
 * @throws IO_ERROR if the lock is already locked
 * @see with_lock
 */
void with_lock_name(Store *store, char *lock_name,
    void (*func)(void *arg), void *arg)
{
  Lock *lock = store->open_lock(store, lock_name);
  if (!lock->obtain(lock)) {
    RAISE(IO_ERROR, COULD_NOT_OBTAIN_LOCK);
  }
  func(arg);
  lock->release(lock);
  store->close_lock(lock);
}

/**
 * Remove a reference to the store. If the reference count gets to zero free
 * all resources used by the store.
 *
 * @param store the store to be dereferenced
 */
void store_deref(Store *store)
{
  mutex_lock(&store->mutex);
  if (--store->ref_cnt == 0) {
    store->close_i(store);
  } else {
    mutex_unlock(&store->mutex);
  }
}

/**
 * Create a store struct initializing the mutex.
 */
Store *store_create()
{
  Store *store = ALLOC(Store);
  store->ref_cnt = 1;
  mutex_init(&store->mutex, NULL);
  mutex_init(&store->ext_mutex, NULL);
  return store;
}

/**
 * Destroy the store, destroying the mutex.
 */
void store_destroy(Store *store)
{
  mutex_destroy(&store->mutex);
  mutex_destroy(&store->ext_mutex);
  free(store);
}
