#include "global.h"
#include "hash.h"

#ifndef FRT_STORE_H
#define FRT_STORE_H

#define BUFFER_SIZE 1024
#define LOCK_PREFIX "ferret-"

#define VINT_MAX_LEN 10
#define VINT_END BUFFER_SIZE - VINT_MAX_LEN

typedef struct Buffer {
  uchar buf[BUFFER_SIZE];
  int start;
  int pos;
  int len;
} Buffer;

typedef struct OutStream {
  Buffer buf;
  void   *file;
  int   pointer; /* only used by RAMOut */
  void   (*flush_internal)(struct OutStream *os, uchar *buf, int len);
  void   (*seek_internal)(struct OutStream *os, int pos);
  void   (*close_internal)(struct OutStream *os);
} OutStream;

typedef struct CompoundInStream CompoundInStream;

typedef struct InStream {
  int    is_clone;
  Buffer buf;
  void   *file;
  union {
    int   pointer; /* only used by RAMIn */
    char *path;    /* only used by FSIn */
    CompoundInStream *cis;
  } d;
  void   (*read_internal)(struct InStream *is, uchar *buf, int offset, int len);
  void   (*seek_internal)(struct InStream *is, int pos);
  void   (*close_internal)(struct InStream *is);
  void   (*clone_internal)(struct InStream *is, struct InStream *new_index_i);
  int    (*length_internal)(struct InStream *is);
} InStream;

struct CompoundInStream {
  InStream *sub;
  int offset;
  int length;
};

#define is_length(mis) mis->length_internal(mis)

typedef struct Store Store;
typedef struct Lock Lock;
struct Lock {
  char *name;
  Store *store;
  int (*obtain)(Lock *lock);
  int (*is_locked)(Lock *lock);
  void (*release)(Lock *lock);
};

typedef struct CompoundStore {
  Store *store;
  const char *name;
  HshTable *entries;
  InStream *stream;
} CompoundStore;

struct Store {
  int ref_cnt; /* for fs_store only */
  mutex_t mutex;
  mutex_t ext_mutex;
  union {
    char *path;          /* for fs_store only */
    HshTable *ht;        /* for ram_store only */
    CompoundStore *cmpd; /* for compound_store only */
  } dir;
  void (*touch)(Store *store, char *filename);
  int  (*exists)(Store *store, char *filename);
  int  (*remove)(Store *store, char *filename);
  int  (*rename)(Store *store, char *from, char *to);
  int  (*count)(Store *store);
  void (*clear)(Store *store);
  void (*clear_all)(Store *store);
  void (*clear_locks)(Store *store);
  int (*length)(Store *store, char *filename);
  void (*each)(Store *store, void (*func)(char *fname, void *arg), void *arg);
  OutStream *(*create_output)(Store *store, const char *filename);
  InStream *(*open_input)(Store *store, const char *filename);
  Lock *(*open_lock)(Store *store, char *lockname);
  void (*close_lock)(Lock *lock);
  void (*close_i)(Store *store);
};

void store_deref(Store *store);
Store *store_create();
char *store_to_s(Store *store);
Store *open_fs_store(const char *pathname);
Store *open_ram_store();
Store *open_ram_store_and_copy(Store *store, bool close_dir);
Store *open_cmpd_store(Store *sub, const char *filename);
void ram_close(Store *store);
Buffer *buf_create();
void os_flush(OutStream *os);
void os_close(OutStream *os);
int os_pos(OutStream *os);
void os_seek(OutStream *os, int new_pos);
void os_write_byte(OutStream *os, uchar b);
void os_write_bytes(OutStream *os, uchar *b, int len);
uchar is_read_byte(InStream *is);
int is_pos(InStream *is);
uchar *is_read_bytes(InStream *is, uchar *b, int offset, int len);
void is_seek(InStream *is, int pos);
InStream *is_clone(InStream *is);
void is_close(InStream *is);
int is_read_int(InStream *is);
llong is_read_long(InStream *is);
unsigned int is_read_uint(InStream *is);
ullong is_read_ulong(InStream *is);
ullong is_read_vint(InStream *is);
void is_skip_vints(InStream *is, register int cnt);
void is_read_chars(InStream *is, char* buffer, int off, int len) ;
char *is_read_string(InStream *is);
void os_write_int(OutStream *os, int l);
void os_write_long(OutStream *os, llong l);
void os_write_uint(OutStream *os, unsigned int l);
void os_write_ulong(OutStream *os, ullong l);
void os_write_vint(OutStream *os, register ullong i);
void os_write_chars(OutStream *os, char *buf, int start, int length);
void os_write_string(OutStream *os, char *str);
void os_copy_bytes(OutStream *os, InStream *is, int len);
OutStream *os_create();
InStream *is_create();
void buf_destroy(Buffer *buf);

/* RamStore functions */
int ramo_length(OutStream *os);
void ramo_reset(OutStream *os);
int rami_length(InStream *is);
void ramo_write_to(OutStream *os, OutStream *other_o);
OutStream *ram_create_buffer();
void ram_destroy_buffer(OutStream *os);

int file_is_lock(char *filename);

void with_lock(Lock *lock, void (*func)(void *arg), void *arg);
void with_lock_name(Store *store, char *lock_name,
    void (*func)(void *arg), void *arg);

#endif
