#ifdef WIN32

#include "global.h"
#include "store.h"
#include <stdio.h>
#include <io.h>
#include <errno.h>
#include <string.h>

/**
 * Create a filepath for a file in the store using the operating systems
 * default file seperator.
 */
char *join_path(char *buf, const char *base, const char *filename)
{
  sprintf(buf, "%s\\%s", base, filename);
  return buf;
}

bool exists(char *path)
{
  int fd = _open(path, 0);
  if (fd < 0) {
    if (errno != ENOENT) {
      RAISE(IO_ERROR, strerror(errno));
    }
    return false;
  }
  _close(fd);
  return true;
}

int fcount(char *path)
{
  char buf[MAX_FILE_PATH];
  struct _finddata_t fd;
  intptr_t d;
  int cnt = 0;

  join_path(buf, path, "*");

  if ((d = _findfirst(buf, &fd)) < 0) {
    RAISE(IO_ERROR, strerror(errno));
  }

  do {
    if (fd.name[0] != '.') {
      cnt++;
    }
  } while (_findnext(d, &fd) == 0);
  _findclose(d);

  return cnt;
}

void dir_each(char *path, void (*func)(char *fname, void *arg), void *arg)
{
  char buf[MAX_FILE_PATH];
  struct _finddata_t fd;
  intptr_t d;
  join_path(buf, path, "*");

  if ((d = _findfirst(buf, &fd)) < 0) {
    RAISE(IO_ERROR, strerror(errno));
  }

  while (_findnext(d, &fd) == 0) {
    if (fd.name[0] != '.' && !file_is_lock(fd.name)) {
      func(fd.name, arg);
    }
  }
  _findclose(d);
}

/**
 * Clear all the locks in the store.
 *
 * @param store the store to clear the locks from
 * @throws IO_ERROR if there is an error opening the directory
 */
void fs_clear_locks(Store *store)
{
  char buf[MAX_FILE_PATH];
  struct _finddata_t fd;
  intptr_t d;
  join_path(buf, store->dir.path, "*");

  if ((d = _findfirst(buf, &fd)) < 0) {
    RAISE(IO_ERROR, strerror(errno));
  }

  while (_findnext(d, &fd) == 0) {
    if (file_is_lock(fd.name)) {
      remove(join_path(buf, store->dir.path, fd.name));
    }
  }
  _findclose(d);
}

/**
 * Clear all files from the store except the lock files.
 *
 * @param store the store to clear all the files from
 * @throws IO_ERROR if there is an error deleting the files
 */
void fs_clear(Store *store)
{
  char buf[MAX_FILE_PATH];
  struct _finddata_t fd;
  intptr_t d;
  join_path(buf, store->dir.path, "*");

  if ((d = _findfirst(buf, &fd)) < 0) {
    RAISE(IO_ERROR, strerror(errno));
  }

  while (_findnext(d, &fd) == 0) {
    if (fd.name[0] != '.' && !file_is_lock(fd.name)) {
      remove(join_path(buf, store->dir.path, fd.name));
    }
  }
  _findclose(d);
}

/**
 * Clear all files from the store including the lock files.
 *
 * @param store the store to clear all the files from
 * @throws IO_ERROR if there is an error deleting the files
 */
void fs_clear_all(Store *store)
{
  char buf[MAX_FILE_PATH];
  struct _finddata_t fd;
  intptr_t d;
  join_path(buf, store->dir.path, "*");

  if ((d = _findfirst(buf, &fd)) < 0) {
    RAISE(IO_ERROR, strerror(errno));
  }

  while (_findnext(d, &fd) == 0) {
    if (fd.name[0] != '.') {
      remove(join_path(buf, store->dir.path, fd.name));
    }
  }
  _findclose(d);
}

#endif
