#include <string.h>
#include "search.h"

/****************************************************************************
 *
 * FuzzyStuff
 *
 * The main method here is the fuzq_score method which scores a term against
 * another term. The other methods all act in support.
 *
 ****************************************************************************/


int fuzq_calculate_max_distance(FuzzyQuery *fuzq, int m) 
{
  return (int)((1.0 - fuzq->min_sim) * (MIN(fuzq->text_len, m) + fuzq->pre_len));
}

void fuzq_initialize_max_distances(FuzzyQuery *fuzq)
{
  int i;
  for (i = 0; i < TYPICAL_LONGEST_WORD; i++) {
    fuzq->max_distances[i] = fuzq_calculate_max_distance(fuzq, i);
  }
}

int fuzq_get_max_distance(FuzzyQuery *fuzq, int m)
{
  return (m < TYPICAL_LONGEST_WORD) ? fuzq->max_distances[m]
                                    : fuzq_calculate_max_distance(fuzq, m);
}

float fuzq_score(FuzzyQuery *fuzq, char *target)
{
  int i, j;
  int max_distance;
  int m = (int)strlen(target);
  int n = fuzq->text_len;
  int *d = fuzq->da;
  char *text = fuzq->text;
  if (n == 0)  {
    /* we don't have anything to compare.  That means if we just add
     * the letters for m we get the new word */
    return fuzq->pre_len == 0 ? 0.0f : 1.0f - ((float) m / fuzq->pre_len);
  }
  if (m == 0) {
    return fuzq->pre_len == 0 ? 0.0f : 1.0f - ((float) n / fuzq->pre_len);
  }

  max_distance = fuzq_get_max_distance(fuzq, m);

  //printf("n%dm%dmd%ddiff%d<%s><%s>\n", n, m, max_distance, m-n, fuzq->text, target);
  if (max_distance < ((m > n) ? (m-n) : (n-m))) { /* abs */
    /* Just adding the characters of m to n or vice-versa results in too many
     * edits for example "pre" length is 3 and "prefixes" length is 8. We can
     * see that given this optimal circumstance, the edit distance cannot be
     * less than 5 which is 8-3 or more precisesly Math.abs(3-8). If our
     * maximum edit distance is 4, then we can discard this word without
     * looking at it. */
    return 0.0f;
  }

  /* Let's make sure we have enough room in our array to do the distance
   * calculations. */
  if (((m+1) * (n+1)) >= fuzq->da_capa) {
    fuzq->da_capa = ((m+1) * (n+1)) * 2;
    REALLOC_N(fuzq->da, int, fuzq->da_capa);
    printf("making capa %d -> %d\n", fuzq->da_capa, (int)fuzq->da);
    d = fuzq->da;
  }

  /* init matrix d */
  for (i = 0; i <= n; i++) d[i + n * 0] = i;
  for (j = 0; j <= m; j++) d[0 + n * j] = j;
  
  /* start computing edit distance */
  for (i = 1; i <= n; i++) {
    int best_pos_ed_dist = m;
    char s_i = text[i - 1];
    for (j = 1; j <= m; j++) {
      if (s_i != target[j-1]) {
        d[i + n*j] = min3(d[i-1 + n*j], d[i + n*(j-1)], d[i-1 + n*(j-1)])+1;
      } else {
        d[i + n*j] = min3(d[i-1 + n*j]+1, d[i + n*(j-1)]+1, d[i-1 + n*(j-1)]);
      }
      best_pos_ed_dist = min2(best_pos_ed_dist, d[i + n*j]);
    }
    //printf("(bped = %d, i = %d, md = %d)", best_pos_ed_dist, i, max_distance);

    /* After calculating row i, the best possible edit distance can be found
     * by found by finding the smallest value in a given column.  If the
     * best_pos_ed_dist is greater than the max distance, abort.
     */

    if ((i > max_distance) && (best_pos_ed_dist > max_distance)) {
      /* equal is okay, but not greater
       * the closest the target can be to the text is just too far away.
       * this target is leaving the party early. */
      return 0.0f;
    }
  }
  //printf("<%f, d[n + m*m] = %d min_len = %d>", 1.0f - ((float)d[n + m*m] / (float) (fuzq->pre_len + min(n, m))), d[n + m*m], fuzq->pre_len + min(n, m));

  /* this will return less than 0.0 when the edit distance is greater than the
   * number of characters in the shorter word.  but this was the formula that
   * was previously used in FuzzyTermEnum, so it has not been changed (even
   * though min_sim must be greater than 0.0) */
  return 1.0f - ((float)d[n + n*m] / (float) (fuzq->pre_len + min2(n, m)));
}

/****************************************************************************
 *
 * FuzzyQuery
 *
 ****************************************************************************/

char *fuzq_to_s(Query *self, char *field) 
{
  char *buffer, *bptr;
  FuzzyQuery *fuzq = (FuzzyQuery *)self->data;
  Term *term = fuzq->term;
  int tlen = (int)strlen(term->text);
  int flen = (int)strlen(term->field);
  bptr = buffer = ALLOC_N(char, tlen + flen + 35);

  if (strcmp(term->field, field) != 0) {
    sprintf(bptr, "%s:", term->field);
    bptr += strlen(term->field) + 1;
  }
  sprintf(bptr, "%s~", term->text);
  bptr += strlen(bptr);
  if (fuzq->min_sim != 0.5) {
    dbl_to_s(bptr, fuzq->min_sim);
    bptr += strlen(bptr);
  }
  if (self->boost != 1.0) {
    *bptr = '^';
    dbl_to_s(++bptr, self->boost);
  }
  return buffer;
}

typedef struct ScoredTerm {
  Term *term;
  float score;
} ScoredTerm;

bool scored_term_less_than(void *p1, void *p2)
{
  ScoredTerm *st1 = (ScoredTerm *)p1;
  ScoredTerm *st2 = (ScoredTerm *)p2;

  if (st1->score == st2->score)
    return (strcmp(st1->term->text, st2->term->text) < 0);

  return (st1->score < st2->score);
}

void scored_term_destroy(ScoredTerm *self)
{
  term_destroy(self->term);
  free(self);
}

ScoredTerm *scored_term_create(Term *term, float score)
{
  ScoredTerm *self = ALLOC(ScoredTerm);
  self->term = term;
  self->score = score;
  return self;
}

Query *fuzq_rewrite(Query *self, IndexReader *ir)
{
  Query *q;
  Query *tq;
  FuzzyQuery *fuzq = (FuzzyQuery *)self->data;

  Term *term = fuzq->term;
  char *text = term->text;
  char *field = term->field;
  Term prefix_term;
  prefix_term.field = field;
  if (fuzq->pre_len >= (int)strlen(text)) {
    q = tq_create(term_clone(term));
  } else {
    PriorityQueue *term_pq;
    TermEnum *te;
    Term prefix_term;
    char *prefix = NULL;
    int pre_len = fuzq->pre_len;
    ScoredTerm *scored_term;

    q = bq_create(true);

    term_pq = pq_create(((BooleanQuery *)q->data)->max_clause_cnt,
        &scored_term_less_than);
    term_pq->free_elem = (free_ft)&scored_term_destroy;

    prefix_term.field = field;
    prefix_term.text = (char *)EMPTY_STRING;
    if (pre_len >= 0) {
      prefix = ALLOC_N(char, pre_len + 1);
      strncpy(prefix, text, pre_len);
      prefix_term.text = prefix;
      prefix_term.text[pre_len] = '\0';
    }
    te = ir->terms_from(ir, &prefix_term);

    fuzq->scale_factor = (float)(1.0 / (1.0 - fuzq->min_sim));
    fuzq->text = fuzq->term->text + pre_len;
    fuzq->text_len = (int)strlen(fuzq->text);
    fuzq_initialize_max_distances(fuzq);

    if (te) {
      TermBuffer *tb = te->tb_curr;
      float score = 0.0, min_score = fuzq->min_sim;

      TRY
        do { 
          if (strcmp(tb->field, field) != 0 ||
              (prefix && strncmp(tb->text, prefix, pre_len) != 0))
            break;

          score = fuzq_score(fuzq, tb->text + pre_len);
          //printf("%s:%s:%f\n", tb->text, fuzq->text, score);

          if (score > min_score) {
            pq_insert(term_pq, scored_term_create(tb_get_term(tb), score));
            if (pq_full(term_pq))
              min_score = ((ScoredTerm *)pq_top(term_pq))->score;
          }
        } while ((tb = te->next(te)) != NULL);
      XFINALLY
        te->close(te);
      XENDTRY
    }
    free(prefix);

    while ((scored_term = pq_pop(term_pq)) != NULL) {
      tq = tq_create(scored_term->term); /* found match */
      tq->boost = self->boost;           /* set boost */
      bq_add_query(q, tq, BC_SHOULD);    /* add query */
      free(scored_term); /* no need to free the term as it's in the query */
    }
    pq_destroy(term_pq);
  }

  return q;
}

void fuzq_destroy(Query *self)
{
  FuzzyQuery *fuzq = (FuzzyQuery *)self->data;
  if (self->destroy_all) term_destroy((Term *)fuzq->term);
printf("freeing %d -> %d\n", fuzq->da_capa, (int)fuzq->da);
  free(fuzq->da);
printf("success\n");
  free(fuzq);
  q_destroy_i(self);
}

static uint fuzq_hash(Query *self)
{
  FuzzyQuery *fq = (FuzzyQuery *)self->data;
  return term_hash(fq->term) ^ *((int *)&fq->min_sim) ^ fq->pre_len;
}

static int fuzq_eq(Query *self, Query *o)
{
  FuzzyQuery *fq1 = (FuzzyQuery *)self->data;
  FuzzyQuery *fq2 = (FuzzyQuery *)o->data;
  return term_eq(fq1->term, fq2->term) &&
    (fq1->pre_len == fq2->pre_len) && 
    (fq1->min_sim == fq2->min_sim);
}

Query *fuzq_create(Term *term)
{
  Query *self = q_create();

  FuzzyQuery *fq = ALLOC(FuzzyQuery);
  ZEROSET(fq, FuzzyQuery, 1);
  fq->term = term;
  fq->pre_len = DEF_PRE_LEN;
  fq->min_sim = DEF_MIN_SIM;
  self->data = fq;

  self->type = FUZZY_QUERY;
  self->to_s = &fuzq_to_s;
  self->hash = &fuzq_hash;
  self->eq = &fuzq_eq;
  self->rewrite = &fuzq_rewrite;
  self->destroy_i = &fuzq_destroy;
  self->create_weight_i = &q_create_weight_unsup;

  return self;
}

Query *fuzq_create_mp(Term *term, float min_sim, int pre_len)
{
  Query *self = fuzq_create(term);
  FuzzyQuery *fuzq = (FuzzyQuery *)self->data;
  if (pre_len) fuzq->pre_len = pre_len;
  if (min_sim) fuzq->min_sim = min_sim;
  return self;
}
